/*****************************************************************************
 *
 * $RCSfile: Main.c,v $
 *
 * Main event loop for FetchMap
 *   - OS version check
 *   - branch to Application Event Handlers
 *   - preferences load/store
 *
 * created: 2003-01-04
 *
 * History:
 * 2003-01-04:	first version
 *
 * This program is Copyright (C) 01/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005/04/10 15:33:56 $
 *
 * $Revision: 1.12 $
 *
 * $Log: Main.c,v $
 * Revision 1.12  2005/04/10 15:33:56  mp
 * preferences are now loaded even if started from another application
 *
 * Revision 1.11  2005/03/05 11:53:14  mp
 * added VFS support
 *
 * Revision 1.10  2004/04/06 20:50:21  mp
 * fixed Application Preferences version check in AppStart()
 *
 * Revision 1.9  2004/04/04 13:45:02  mp
 * added invocation of Set_UTM_Parameters()
 *
 * Revision 1.8  2004/03/24 19:24:07  mp
 * removed unnecessary #include
 *
 * Revision 1.7  2004/03/21 19:18:22  mp
 * added MathLib, added comments
 *
 * Revision 1.5  2003/11/16 14:56:35  mp
 * replaced printDouble by format_number
 *
 * Revision 1.4  2003/10/27 21:18:25  mp
 * Added launching with code sysAppLaunchCmdLookup
 * (for testing) and with code sysAppLaunchCmdNormalLaunch
 * and map parameters from other application.
 *
 * Revision 1.3  2003/10/25 19:55:27  mp
 * changed map parameter block type from char* to MapParamType*
 *
 * Revision 1.2  2003/10/24 17:46:50  mp
 * moved MapParamType declaration to common.h
 *
 * Revision 1.1.1.1  2003/10/24 15:56:42  mp
 * Imported FetchMap to CVS revision control.
 *
 *
 *
 ****************************************************************************/

#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
//#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <Extensions/ExpansionMgr/VFSMgr.h>
#include <Unix/sys_types.h>
#include "stringil.h"
#include <sys_socket.h>
#include "MathLib.h"
#include "ResourceDefines.h"
#include "Utils.h"		/* RomVersionCompatible() */
#include "common.h"
#include "DownloadMap.h"
#include "Constants.h"
#include "utm.h"
#include "geo.h"
#include "fp.h"

#define NUM_DIGITS 7

extern Boolean MainFormHandleEvent(EventPtr eventP);
extern Boolean AboutFormHandleEvent(EventPtr eventP);

/** Application Preferences */
PrefsType Preferences;

/** VFS Feature Set available */
Boolean gVfsManagerFtr;


/*****************************************************************************
 * FUNCTION:	ApplicationHandleEvent
 *
 * DESCRIPTION: Set Application Event Handler to Form specific Handler.
 *
 * PARAMETERS:	eventP	-- pointer to event (only frmLoadEvent handled)
 *
 * RETURNED:    Event handled as boolean.
 ****************************************************************************/
static Boolean AppHandleEvent( EventPtr eventP)
{
  UInt16 formId;
  FormPtr frmP;

  if (eventP->eType == frmLoadEvent) {
    /* Load the form resource. */
    formId = eventP->data.frmLoad.formID;
    frmP = FrmInitForm(formId);
    ErrFatalDisplayIf(!frmP, "Can't initialize form");
    FrmSetActiveForm(frmP);

    /*  
     * Set the event handler for the form.  The handler of the currently
     * active form is called by FrmHandleEvent each time is receives an
     * event.
     */
    switch (formId) {
      case MainForm:
	FrmSetEventHandler(frmP, MainFormHandleEvent);
	break;

      case AboutForm:
        FrmSetEventHandler(frmP, AboutFormHandleEvent);
        break;

      default:
	ErrFatalDisplay("Invalid Form Load Event");
	break;
    }
    return true;
  } else
    return false;
}


/*****************************************************************************
 * FUNCTION:	EventLoop
 *
 * DESCRIPTION: Main Event Loop
 *
 ****************************************************************************/
static void AppEventLoop(void)
{
  Err error;
  EventType event;

  do {
    EvtGetEvent(&event, evtWaitForever);
    
    if (! SysHandleEvent(&event))
      if (! MenuHandleEvent(0, &event, &error))
        if (! AppHandleEvent(&event))
          FrmDispatchEvent(&event);

  } while (event.eType != appStopEvent);
}


/*****************************************************************************
 * FUNCTION:	AppStart
 *
 * DESCRIPTION: Opens Net Library
 *		Loads Application Preferences.
 *		Checks if Math Library is available and
 *		shows appropriate error message on failure.
 *
 * GLOBALS:	Preferences		-- Application Preferences
 ****************************************************************************/
static Err AppStart(void)
{
   UInt32	vfsMgrVersion;			/* VFS Manager Version */
   UInt16	prefsSize;
   Int16	prefsVersion = noPreferenceFound;
   Err		err;
   
   /* Get the refNum of the Net Library */
   err = SysLibFind("Net.lib", &AppNetRefnum);
   if (err) {
     ErrDisplay("\nNet Library not found, exiting...");
     return err;
   }

   prefsSize = sizeof(PrefsType);
   
   /* Get Application Preferences (unsaved) */
   prefsVersion = PrefGetAppPreferences(kCreatorId, kPrefsId, &Preferences,
      &prefsSize, false);
   
   if ( (prefsVersion != kAppVer) || (prefsVersion == noPreferenceFound) ) {
     prefsVersion = kPrefsId;
     /* no preferences or wrong version: set defaults */ 
     Preferences.lat = 52.269166; /* 5216.1500' N */
     Preferences.lon = 10.521110; /* 1031.2666' E */
     Preferences.scale   = 2000;
     Preferences.width   = 160;
     Preferences.height  = 160;
     Preferences.units   = POS_DM;
     Preferences.use_vfs = false;
   }

   /* check if VFS Manager Feature Set is available */
   err = FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &vfsMgrVersion);
   gVfsManagerFtr = (err == 0) ? true : false; 
   err = 0;

   err = SysLibFind(MathLibName, &MathLibRef);
   if (err)
     err = SysLibLoad(LibType, MathLibCreator, &MathLibRef);

   if (err) {
     FrmAlert(NoMathLibAlert);
   } else {
     err = MathLibOpen(MathLibRef, MathLibVersion);

     if (err) {
       FrmAlert(NoMathLibAlert);
     }
   }

  err = Set_UTM_Parameters(	SEMIMAJOR_AXIS,
      			    	FLATTENING,
			    	0 /* Override_Zone */);
  
  return err;
}


/*****************************************************************************
 * FUNCTION:	AppStop
 *
 * DESCRIPTION: Closes NetLib immediately.
 *		Saves Application Preferences.
 *		Closes MathLib.
 *
 ****************************************************************************/
static void AppStop(void)
{
  Int	usecount;	/* MathLib use count */
  Err	err;		/* error code */

  /*
   * Close network library, do not tear down immediately if still in use by
   *   another application. Possible errors:
   *   <Success | netErrNotOpen | netErrStillOpen>
   *   No error handling required. 
   */
  NetLibClose(AppNetRefnum, true /* immediate */);

  /* Save Data stored in Form */
  FrmSaveAllForms();
  
  /* Set Application Preferences (unsaved) */
  PrefSetAppPreferences(kCreatorId, kPrefsId, kAppVer, &Preferences,
     sizeof(PrefsType), false);
  FrmCloseAllForms();

  err = MathLibClose(MathLibRef, &usecount);
  ErrFatalDisplayIf(err, "Can't close MathLib");
  if (usecount == 0)
    SysLibRemove(MathLibRef);
}


/*****************************************************************************
 * FUNCTION:	PilotMain
 *
 * DESCRIPTION: Application Entry Function
 *
 * PARAMETERS:	launchCode		- Launch Code
 *		launchParameters	- Command Parameter Block Pointer
 *		launchFlags		- Launch Flags (set by OS)
 *
 * RETURNED:    always 0
 *
 ****************************************************************************/
UInt32 PilotMain(UInt16 launchCode, MemPtr launchParameters,
  UInt16 launchFlags)
{
  Err		error;		/* Error Code */
  ProgressPtr	prgP;		/* Progress Manager Data Block Ptr */
  Char		buf[40];	/* String Buffer */
  MapParamType*	map;		/* Map Parameter Ptr */
  PrefsType     prefs;		
  UInt16	prefsSize;
  Int16		prefsVersion = noPreferenceFound;
  UInt16	vfs_vol;

  switch (launchCode) {
  case sysAppLaunchCmdNormalLaunch:
    error = RomVersionCompatible (kOurMinVersion, launchFlags);
    if (error)
      return error;
        
    error = AppStart();
    if (error) 
      return error;
    
    /* check if launched with parameters (from other application) */
    if (launchParameters != NULL) {
      
      /* check if globals are available */ 
      if (launchFlags & sysAppLaunchFlagNewGlobals) {
	
	/* Start Progress Dialog */
        prgP = PrgStartDialogV31("Fetching Map...", callbackFunction);
        
        if (!gVfsManagerFtr || (check_dir(&vfs_vol) != 0)) {
           Preferences.use_vfs = false;	 
	}
	
	/* Download Map */
	error = downloadMap((MapParamType *) launchParameters, prgP,
	                    vfs_vol, Preferences.use_vfs);
        
	/* End Progress Dialog */
	PrgStopDialog(prgP, false);
        
	/* Close the Net Library */
        NetLibClose(AppNetRefnum, false /* immediate */);
      }
      
      return error;
    }
      
    
    /* launched from Application Launcher */
    FrmGotoForm(MainForm);
    AppEventLoop();
    AppStop();
    break;
  
  case sysAppLaunchCmdLookup:
    /*
     * Request from other application to fetch a map.
     * (That's probably not exactly what this launch code is intended for.)
     * Unfortunately the PalmOS Socket Library uses some globals!
     * Therefore it is not possible to use this launch code.
     * Instead of fetching the map, we just show an alert with the
     * requested parameters.
     */
    error = RomVersionCompatible (kOurMinVersion, launchFlags);
    if (error)
      return error;
    
    if (launchParameters == NULL)
      return sysErrParamErr;
    
    map = (MapParamType *)launchParameters;

    format_number(map->lat, NUM_DIGITS, buf);
    StrCat(buf, "\n");
    format_number(map->lon, NUM_DIGITS, &buf[ strlen(buf) ] );
    FrmCustomAlert(LaunchAlert, buf, NULL, NULL);

    return 0;
    break;
  
  default:
    break;
  }
  
  return 0;
}
