/*****************************************************************************
 *
 * $RCSfile: MainForm.c,v $
 *
 * Main Form event loop for FetchMap
 *   - selection of format for position entry
 *   - entry of coordinates, map scale and dimensions
 *   - sanity checks on entries and conversion to internal representation
 *     (decimal degrees)
 *   - start of map download
 *   - network connection closing
 *
 * created: 2003-01-04
 *
 * This program is Copyright (C) 01/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005/04/10 15:30:58 $
 *
 * $Revision: 1.11 $
 *
 * $Log: MainForm.c,v $
 * Revision 1.11  2005/04/10 15:30:58  mp
 * removed static declaration of check_dir()
 *
 * Revision 1.10  2005/03/05 11:51:56  mp
 * moved static function declarations from MainForm.h
 * added Virtual File System (VFS) support
 *
 * Revision 1.9  2004/04/11 07:51:08  mp
 * fixed saving of map size upon application closing
 *
 * Revision 1.8  2004/04/10 12:28:19  mp
 * fixed map size setting
 *
 * Revision 1.7  2004/04/10 10:39:36  mp
 * fixed unit conversion and input processing
 *
 * Revision 1.6  2004/04/06 20:53:04  mp
 * fixed UTM to/from geodedic conversion, removed obsolete code
 *
 * Revision 1.5  2004/04/04 14:02:30  mp
 * modified user interface, work in progress
 *
 * Revision 1.4  2004/03/23 20:14:12  mp
 * switched to MathLib, modified cvt_atof() in fp.c
 *
 * Revision 1.3  2003/11/16 14:59:17  mp
 * replaced printDouble by format_number
 *
 * Revision 1.2  2003/10/27 21:15:31  mp
 * Moved downloadMap() to DownloadMap.c,
 * map parameters passed with as function parameter
 * instead of globals.
 *
 * Revision 1.1.1.1  2003/10/24 15:56:42  mp
 * Imported FetchMap to CVS revision control.
 *
 ****************************************************************************/

#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include <Extensions/ExpansionMgr/VFSMgr.h>
#include <Unix/sys_types.h>
#include "stringil.h"
#include <PalmCompatibility.h>
#include <sys_socket.h>
#include "MathLib.h"
#include "ResourceDefines.h"
#include "common.h"
#include "DownloadMap.h"
#include "MainForm.h"
#include "SetFieldText.h"
#include "Constants.h"
#include "err_alert.h"
#include "fp.h"
#include "geo.h"
#include "utm.h"

#define	SCALE_ITEMS	30
#define NUM_DIGITS	7
#define DEFAULT_WIDTH	160
#define DEFAULT_HEIGHT	160
//#define VERBOSE

/* Globals */
extern PrefsType	Preferences;
extern Boolean		gVfsManagerFtr;       /* VFS Feature Set available */
const double		radius_km = 6378.135; /* earth's equatorial radius */
const double		lat_dist_km = 110.87; /* distance per degree of lat. */

/* Locals */
static double		lat;			/* latitude */
static double		lon;			/* longitude */
static UInt16		utm_zone = 32;		/* UTM zone */
static char		utm_belt = 'U';		/* UTM belt */

/* Caution: Make sure the entries match the list ScaleList in Resources.rcp ! */
static UInt32	ScaleMappings[SCALE_ITEMS] = {
		  1000, 1500, 2000, 3000, 5000, 7500,
		  10000, 15000, 20000, 30000, 50000, 75000,
		  100000, 150000, 200000, 300000, 500000, 750000,
		  1000000, 1500000, 2000000, 3000000, 5000000, 7500000,
		  10000000, 15000000, 20000000, 30000000, 50000000, 75000000
		};

static Char InputErrMsg[] = "Invalid Input! ";
static Char InputOkMsg[]  = "Input o.k.     ";
static Char DoneMsg[]     = "Done.          ";
static Char ClrMsg[]      = "               ";

static void MainFormInit(FormPtr frmP)			MAINFORM_SECTION;
static void MainFormDeinit(FormPtr frmP)		MAINFORM_SECTION;
static UInt16 mapToPosition(UInt32 *mappingArray,
  UInt32 value, UInt16 mappings, UInt32 defaultItem)	MAINFORM_SECTION;
static Boolean field_active(UInt16 fieldID)		MAINFORM_SECTION;
static Boolean scan_position(UInt16 fieldID,
  double *lat, double *lon)				MAINFORM_SECTION;
static Boolean scan_zone(UInt16 *utm_zone,
  Char *utm_belt)					MAINFORM_SECTION;
static Boolean insert_position(double lat, double lon,
  double *eastP, double *northP)			MAINFORM_SECTION;
static Boolean scan_field(UInt16 field, Char *strP,
  Boolean *lat_valid, Boolean *lon_valid)		MAINFORM_SECTION;
static void scan_mapsize(UInt16 *width, UInt16 *height)	MAINFORM_SECTION;

static void MainFormInit(FormPtr frmP)
{
#pragma unused(frmP)
}

static void MainFormDeinit(FormPtr frmP)
{
#pragma unused(frmP)
}


/*****************************************************************************
 * FUNCTION:	mapToPosition
 *
 * DESCRIPTION: Get list position from list item (reverse array lookup).
 *
 * PARAMETERS:	mappingArray	-- Pointer to array
 *		value		-- Lookup value
 *		mappings	-- Total number of items in array
 *		defaultItem	-- Default Item
 *
 * RETURNED:    Position of 'value' in array or default item if not found.
 ****************************************************************************/
static UInt16 mapToPosition(UInt32 *mappingArray, UInt32 value,
					UInt16 mappings, UInt32 defaultItem)
{
  UInt16 i;

  i = 0;
  while (mappingArray[i] != value && i < mappings)
	i++;
  if (i >= mappings)
	return defaultItem;

  return i;
}


/*****************************************************************************
 * FUNCTION:	field_active
 *
 * DESCRIPTION: Determines if the specified field is active, i.e. has the
 *		input focus.
 *
 * PARAMETERS:	fieldID		-- Field ID
 *
 * RETURNED:    Status flag: specified field has focus
 ****************************************************************************/
static
Boolean field_active(UInt16 fieldID)
{
  FormPtr		frmP = FrmGetActiveForm();
  FieldAttrType		fld_attr;
  
  FldGetAttributes(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, fieldID)),
    &fld_attr);
  
  return (fld_attr.hasFocus == 1);
}


/*****************************************************************************
 * FUNCTION:	scan_position
 *
 * DESCRIPTION: Scan position fields to latitude/longitude
 *		depending on selevted position units.
 *
 *		Note: The code assumes to run on a device with
 *		      latin character set.
 *		Note: The degree sign '' is named chrDegreeSign in
 *                    System/CharLatin.h (but we are sloppy...)
 *
 * GLOBALS:	Preferences	-- Application Preferences
 *
 * PARAMETERS:	fieldID		-- ID of selected field
 *		lat		-- pointer to latitude
 *		lon		-- pointer to longitude
 *
 * RETURNED:    Status flag: valid entry
 ****************************************************************************/
static
Boolean scan_position(UInt16 fieldID, double *lat, double *lon)
{
  FormPtr	frmP;		/* form ptr */
  Char		*strP;		/* string ptr */
  Char		*trailP;	/* trailing string ptr */
  double	tmp;		/* temp. double */
  short		frac;		/* number of fractional digits */
  Int16		sign;		/* sign */
  
  frmP = FrmGetActiveForm();
  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
	    FrmGetObjectIndex(frmP, fieldID)));
  
  if (strP[0] == '-') {
    sign = -1;
    strP++;
  } else {
    sign = 1;
  }
  
  /* first value is either angle in degrees or cartesian coordinate in meters */
  tmp = cvt_atof(strP, &trailP, &frac);

  /* handle degrees part or esting/northing */
  switch (Preferences.units) {
    case POS_D:
      /*
       * Field's contents considered as a single floating point number, 
       * which may or may not end with a degree sign (but nothing else).
       */
      if (*trailP == '\0' || *trailP == '') {
        /* valid entry */
	if (fieldID == LatitudeField) {
	  /* Latitude */
	  if ((tmp > 90.0) || (tmp < -90.0)) {
	    /* range error */
	    return false;
	  }
	  /* o.k., update position */
	  *lat = tmp * sign;

	} else {
	  /* Longitude */
	  if ((tmp > 180.0) || (tmp < -180.0)) {
	    /* range error */
	    return false;
	  }
	  /* o.k., update position */
	  *lon = tmp * sign;
	
	}
	return true;
	
      } else {
        /* invalid entry */
	return false;
      }
      break;
      
    case POS_DM:
    case POS_DMS:
      /*
       * Field's contents must start with an integer number, 
       * which must end with a degree sign.
       */
      if (*trailP != '' || frac > 0) {
        /* invalid entry */
	return false;
      } else {
        /* valid entry */
	if (fieldID == LatitudeField) {
	  /* Latitude */
	  if ((tmp > 90.0) || (tmp < -90.0)) {
	    /* range error */
	    return false;
	  }
	  /* o.k., update position */
	  *lat = tmp;

	} else {
	  /* Longitude */
	  if ((tmp > 180.0) || (tmp < -180.0)) {
	    /* range error */
	    return false;
	  }
	  /* o.k., update position */
	  *lon = tmp;
	}
      }
      break;
      
    case POS_UTM:
      /*
       * Field's contents must consist of a single floating point number only 
       */
      if (*trailP != '\0') {
        /* invalid entry */
	return false;
      }
      if (fieldID == LatitudeField) {
        if ((tmp >= 100000.0) && (tmp <= 900000.0)) {
	  *lat = tmp;
          return true;
	} else {
	  return false;
	}
      } else {
        if ((tmp >= 0.0) && (tmp <= 10000000.0)) {
          *lon = tmp;
	  return true;
        } else {
	  return false;
	}
      }
      break;
  } /* switch (Preferences.units.pos_unit) */

  /* skip degree sign */
  strP = trailP + 1;
  
  /* convert string to float */
  tmp = cvt_atof(strP, &trailP, &frac);  

  /*
   * Handle minutes part
   *   The following input is valid:
   *   POS_DM:  <mm.m>'
   *   POS_DMS: <mm>'
   */
  if (*trailP != '\'') {
    /* invalid entry */
    return false;
  }

  if (Preferences.units == POS_DMS && frac > 0) {
    /* invalid entry */
    return false;
  }
  if ((tmp >= 60.0) || (tmp < 0.0)) {
    /* range error */
    return false;
  }

  if (fieldID == LatitudeField) {
    /* o.k., update Latitude  */
    *lat += tmp/60.0;
  } else {
    /* o.k., update Longitude  */
    *lon += tmp/60.0;
  }
    
  if (Preferences.units == POS_DMS) {

    /* skip minutes sign */
    strP = trailP + 1;

    /* convert string to float */
    tmp = cvt_atof(strP, &trailP, NULL);  

    /*
     * Handle seconds part
     *   The following input is valid:
     *   <ss.s>" or <ss>"
     */
    if (*trailP != '\"') {
      /* invalid entry */
      return false;
    }
    if ((tmp >= 60.0) || (tmp < 0.0)) {
      /* range error */
      return false;
    }
    if (fieldID == LatitudeField) {
      /* o.k., update Latitude  */
      *lat += tmp/3600.0;
    } else {
      /* o.k., update Longitude  */
      *lon += tmp/3600.0;
    }
  } /* if (Preferences.units.pos_unit == POS_DMS) */

  if (fieldID == LatitudeField) {
    /* update sign of Latitude  */
    *lat = *lat * sign;
  } else {
    /* update sign of Longitude  */
    *lon = *lon * sign;
  }

  return true;     
}


/*****************************************************************************
 * FUNCTION:	scan_zone
 *
 * DESCRIPTION: Scan zone field to UTM zone and belt,
 *		update latitude and longitude
 *
 * PARAMETERS:	eventP		-- pointer to event structure
 *
 * RETURNED:    Status flag: valid entry
 ****************************************************************************/
static
Boolean scan_zone(UInt16 *utm_zone, Char *utm_belt)
{
  FormPtr	frmP;
  Char		*strP;
  UInt16	zone;
  Char		belt;
  
  frmP = FrmGetActiveForm();
  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
	    FrmGetObjectIndex(frmP, ZoneField)));

  if ((*strP >= '0') && (*strP <= '9')) {
    zone = *strP - '0';
    strP++;
  } else {
    return false;
  }
  
  if ((*strP >= '0') && (*strP <= '9')) {
    zone = zone * 10 + (*strP - '0');
    strP++;
  }
  
  if ((zone < 1) || (zone > 60)) {
    return false;
  }
  
  belt = toupper(*strP);
  
  if ((belt >= 'C') && (belt <= 'X') &&
      (belt != 'I') && (belt != 'O')) {
    *utm_zone = zone;
    *utm_belt = belt;
    return true;
  } else {
    return false;
  }
}


/*****************************************************************************
 * FUNCTION:	insert_position
 *
 * DESCRIPTION: Insert position to fields
 *
 * GLOBALS:	Preferences	-- Application Preferences
 *
 * PARAMETERS:	lat		-- Latitude
 *		lon		-- Longitude
 *		eastP		-- UTM Easting (ptr)
 *		northP		-- UTM Northing (ptr)
 *
 * RETURNED:    flag	-- Conversion to UTM successful. Always true, if any
 *		position format other than UTM is set in the preferences.
 *		If UTM position format is selected and the conversion is
 *		successful, easting and northing are updated by eastP/northP.
 ****************************************************************************/
static
Boolean insert_position(double lat, double lon, double *eastP, double *northP)
{
  Char		lat_str[20];	/* string buffer */
  Char		lon_str[20];	/* string buffer */
  Char		field_str[4];	/* string buffer */
  Int32		err;		/* error code */
  Int32		zone;		/* UTM Zone */
  char		belt;		/* UTM Belt */		
  char		hemisphere;	/* 'N' / 'S' */
  double	easting;	/* Easting in m */
  double	northing;	/* Northing in m */
  
  switch (Preferences.units) {
    case POS_D:
      /* Degrees */
      format_number(lat, 7, lat_str);
      StrCat(lat_str, "");
      SetFieldText(LatitudeField, lat_str, false, true);
      
      format_number(lon, 7, lon_str);
      StrCat(lon_str, "");
      SetFieldText(LongitudeField, lon_str, false, true);
      break;
    
    case POS_DM:
      deg_to_dm_str(lat, lon, lat_str, lon_str);
      SetFieldText(LatitudeField, lat_str, false, true);
      SetFieldText(LongitudeField, lon_str, false, true);
      break;

    case POS_DMS:
      deg_to_dms_str(lat, lon, lat_str, lon_str);
      SetFieldText(LatitudeField, lat_str, false, true);
      SetFieldText(LongitudeField, lon_str, false, true);
      break;
      
   case POS_UTM:   
      /* Universal Transverse Mercator - convert from lat/lon double values */ 
      err = Convert_Geodetic_To_UTM (
      				deg2rad(lat) /* Latitude (rad) */,
                                deg2rad(lon) /* Longitude (rad) */,
                                &zone /* long  *Zone */,
                                &hemisphere /* Hemisphere */,
                                &easting /* *Easting */,
                                &northing /* *Northing */);
      
      if (err == UTM_NO_ERROR) {
        /* print zone and belt to string */
	StrPrintF(field_str, "%ld%c ", zone, Lat_To_UTM_Belt(lat));
	SetFieldText(ZoneField, field_str, false, true);
	
	/* print easting to string */
	format_number(easting, 1, lat_str);
        SetFieldText(LatitudeField, lat_str, false, true);
		
	/* print northing to string */
	format_number(northing, 1, lon_str);
        SetFieldText(LongitudeField, lon_str, false, true);
        
	/* return easting and northing */
	if (eastP) *eastP = easting;
	if (northP) *northP = northing;
      } else {
        return false;
      }
      break;   
  } /* switch (gPrefs.units.pos_unit) */           
  
  return true;
}


/*****************************************************************************
 * FUNCTION:	scan_field
 *
 * DESCRIPTION: Check field and update variables
 *
 * LOCALS:	lat		-- latitude
 *		lon		-- longitude
 *		utm_zone	-- UTM zone
 *		utm_belt	-- UTM belt
 *
 * PARAMETERS:	field		-- field ID
 *		strP		-- pointer to string containing previous
 *				   field contents
 *		lat_valid	-- latitude valid (ref.)  (or NULL)
 *		lon_valid	-- longitude valid (ref.) (or NULL)
 *
 * RETURNED:    status flag: field valid
 ****************************************************************************/
static
Boolean scan_field(UInt16 field, Char *strP, Boolean *lat_valid, 
  Boolean *lon_valid)
{
  Boolean	valid = true;		/* input valid */

  switch (field) {
    case LatitudeField:
      valid = scan_position(LatitudeField, &lat, &lon);
      if (lat_valid && valid) {
	*lat_valid = true; 
      }
      break;

    case LongitudeField:
      valid = scan_position(LongitudeField, &lat, &lon);
      if (lon_valid && valid) {
	*lon_valid = true;
      }
      break;

    case ZoneField:
      valid = scan_zone(&utm_zone, &utm_belt);
      break;

  } /* switch (field) */

  switch (field) {
    case LatitudeField:
    case LongitudeField:
    case ZoneField:
      if (!valid) {
	/* restore old text */
	SetFieldText(field, strP, false /* append */,
	  true /* redraw */);
      }
  }
  
  return valid;
}


/*****************************************************************************
 * FUNCTION:	scan_mapsize
 *
 * DESCRIPTION: Scan map size fields and update variables
 *
 * PARAMETERS:  lat_valid	-- width (ref.)
 *		lon_valid	-- height (ref.)
 *
 * RETURNED:    %
 ****************************************************************************/
static
void scan_mapsize(UInt16 *width, UInt16 *height)
{
  FormPtr	frmP;
  UInt16	w;		/* width */
  UInt16	h;		/* height */
  Char		*strP;		/* string */
  Char		tmp[8];		/* tmp. string buffer */

  frmP = FrmGetActiveForm();

  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
    FrmGetObjectIndex(frmP, WidthField)));
  w = cvt_atof(strP, NULL, NULL);
  

  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
    FrmGetObjectIndex(frmP, HeightField)));
  h = cvt_atof(strP, NULL, NULL);

  /* set values to default if input is invalid */
  if ((w < 10) || (w > 2000)) { 
    w = DEFAULT_WIDTH;
    StrPrintF(tmp, "%u", DEFAULT_WIDTH);
    SetFieldText(WidthField, tmp, false /* append */, true /* redraw */);
  }
  if ((h < 10) || (h > 2000)) { 
    h = DEFAULT_WIDTH;
    StrPrintF(tmp, "%u", DEFAULT_HEIGHT);
    SetFieldText(HeightField, tmp, false /* append */, true /* redraw */);
  }
  
  *width  = w;
  *height = h;
}


/****************************************************************************/
/**
 * \brief	check_dir
 *
 *		Check/create VFS Map Directory
 *
 * \param	vol_ref		VFS Volume Reference Number
 *				returned by reference
 *		
 *
 * \return	0 if directory exists
 ****************************************************************************/
Err check_dir(UInt16 *vol_ref)
{
  FileRef	dirRef;			/* directory reference */
  UInt32	volIterator;		/* VFS volume iterator */
  UInt16	volRefNum;		/* VFS volume reference number */
  Err		err = 0;		/* error code */
  
  volIterator = vfsIteratorStart;

  if (VFSVolumeEnumerate(&volRefNum, &volIterator) == expErrEnumerationEmpty)
    return expErrEnumerationEmpty;
  
    volIterator = vfsIteratorStart;

  /* Iterate through all VFS volumes */
  while (volIterator != vfsIteratorStop) {
    err = VFSVolumeEnumerate(&volRefNum, &volIterator);
    
    if (err == errNone) {

      /* Open the directory and iterate through the files in it. */
      err = VFSFileOpen(volRefNum, FILEDIR, vfsModeRead, &dirRef);
    
      if (err == errNone) {
        /* destination directory exists */
	*vol_ref = volRefNum;
	VFSFileClose(dirRef);
	return 0;

      } else {
        err = VFSFileOpen(volRefNum, BASEDIR, vfsModeRead, &dirRef);

	if (err == errNone) {
          /* parent directory exists */
	  *vol_ref = volRefNum;
	  VFSFileClose(dirRef);
          
	  /* create destination directory */
	  err = VFSDirCreate(volRefNum, FILEDIR);
	
	  if (err == errNone) {
	    return 0;
	  }
	} /* parent directory exists */
        
      }
      
    } /* if VFSVolumeEnumerate o.k. */
  } /* while (volIterator != vfsIteratorStop) */
  
  return vfsErrVolumeBadRef;
}


/*****************************************************************************
 * FUNCTION:	MainFormHandleEvent
 *
 * DESCRIPTION: Main Form Event Handler
 *
 * GLOBALS:	Preferences	-- Application Preferences
 *
 * PARAMETERS:	eventP		-- pointer to event structure
 *
 * RETURNED:    Status flag: event handled
 ****************************************************************************/
Boolean MainFormHandleEvent(EventPtr eventP)
{
  Boolean		handled = false;	/* event handled flag */
  UInt16		volRefNum;		/* VFS volume reference number */
  FormPtr		frmP;			/* form ptr */
  FieldPtr		fldP;			/* field ptr */
  ListPtr		lstP;			/* list ptr */
  ProgressPtr		prgP;			/* progress ptr */
  UInt16		sel;			/* list selection */
  MapParamType  	map;			/* map parameters */
  static Char		tmp[20];		/* temp. string buffer */
  Boolean		valid;			/* input valid */
  UInt16		fld_focus;		/* ID of active field */
  static UInt16		fld_focus_old;		/* ID of previously act field */
  Err			err;			/* UTM error */
  
  frmP = FrmGetActiveForm();
  
  /*
   * The fields' contents have to be checked whenever a field looses focus. 
   * Unfortunately, there is no appropriate event (ctlExitEvent is not triggered
   * by fields). So we have to track the active field manually.
   */
  if ( field_active(LatitudeField) ) {
    fld_focus = LatitudeField;
  } else if ( field_active(LongitudeField) ) {
    fld_focus = LongitudeField;
  } else if ( field_active(ZoneField) ) {
    fld_focus = ZoneField;
  } else {
    /* none */
    fld_focus = 0;
  }
  
  if ((fld_focus_old != 0) && (fld_focus != fld_focus_old)) {
    /* Pseudo-fldExitEvent */
    valid = scan_field(fld_focus_old, tmp, NULL, NULL);
  }


  switch (eventP->eType) {
    case frmOpenEvent:
      MainFormInit(frmP);
      FrmDrawForm(frmP);
      
      /* Init the scale pop-up from Preferences */
      lstP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, ScaleList));
      
      /*  Reverse array lookup (value -> index) */
      sel = mapToPosition(	ScaleMappings,		/* array */
				Preferences.scale,	/* Prefs. value */
				SCALE_ITEMS,		/* array size */
				2000);			/* default value */
      
      LstSetSelection (lstP, sel);
      CtlSetLabel(FrmGetObjectPtr (frmP, FrmGetObjectIndex(frmP, ScalePopup)),
        LstGetSelectionText(lstP, sel));      
     
      /* Init the width/height fields from Preferences */
      StrIToA(tmp, Preferences.width);
      SetFieldText(WidthField, tmp, false /* append */, true /* redraw */);
      StrIToA(tmp, Preferences.height);
      SetFieldText(HeightField, tmp, false /* append */, true /* redraw */);
 
      /* show objects depending on selected position units */
      if (Preferences.units == POS_UTM) {
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, ZoneLabel));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, ZoneField));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, EastLabel));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, NorthLabel));
      } else {
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, LatLabel));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, LonLabel));       
      }

      /* Insert position from preferences to fields */
      if (insert_position(Preferences.lat, Preferences.lon, &lat, &lon)) {
	if (Preferences.units == POS_UTM) {
	  /* easting/northing returned in lat/lon ! */
	  /* get contents of UTM field */
	  scan_zone(&utm_zone, &utm_belt);
	} else {
	  lat = Preferences.lat;
	  lon = Preferences.lon;	    
	}  
      }

      if (!gVfsManagerFtr || (check_dir(&volRefNum) != 0)) {
        FrmHideObject(frmP, FrmGetObjectIndex(frmP, DestLabel));
        FrmHideObject(frmP, FrmGetObjectIndex(frmP, DestPopup));
        Preferences.use_vfs = false;
	
      } else {
        lstP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DestList));
	LstSetSelection(lstP, (Preferences.use_vfs) ? 1 : 0);
        CtlSetLabel(FrmGetObjectPtr (frmP, FrmGetObjectIndex(frmP, DestPopup)),
          LstGetSelectionText(lstP, (Preferences.use_vfs) ? 1 : 0));      
      }
	
      handled = true;
      break;

    case fldEnterEvent:
      switch (eventP->data.ctlSelect.controlID) {
        case LatitudeField:
	  /*
	   * save contents, we might have to restore it
	   * if the new input is invalid
	   */
	  if (fld_focus_old != LatitudeField) {
	    StrCopy(tmp, FldGetTextPtr(FrmGetObjectPtr(frmP,
	      FrmGetObjectIndex(frmP, LatitudeField))));
	  }
	  break;

        case LongitudeField:
	  /*
	   * save contents, we might have to restore it
	   * if the new input is invalid
	   */
	  if (fld_focus_old != LongitudeField) {
	    StrCopy(tmp, FldGetTextPtr(FrmGetObjectPtr(frmP,
	      FrmGetObjectIndex(frmP, LongitudeField))));
	  }
	  break;

        case ZoneField:
	  /*
	   * save contents, we might have to restore it
	   * if the new input is invalid
	   */
	  if (fld_focus_old != ZoneField) {
	    StrCopy(tmp, FldGetTextPtr(FrmGetObjectPtr(frmP,
	      FrmGetObjectIndex(frmP, ZoneField))));
	  }
	  break;
	  
      }
      /* case fldEnterEvent */
      break;
      
    case menuEvent:
      /* Clear status output */
      SetFieldText(StatusField, ClrMsg, false, true);

      if (eventP->data.menu.itemID == DisconnectMenu) {
	/*
	 * Close network library, tear down connection immediately
	 *   Possible errors:
	 *   <Success | netErrNotOpen>
	 *   No error handling required. 
	 */
	NetLibClose(AppNetRefnum, true /* immediate */);
	break;
      }
      
      if (eventP->data.menu.itemID == AboutMenu) {
	/* Show AboutForm */
	MenuEraseStatus(0);
	FrmGotoForm(AboutForm);
	handled = true;
	break;
      }

      switch (eventP->data.menu.itemID) {
        case Units0Menu:
        case Units1Menu:
        case Units2Menu:
        case Units3Menu:
	  valid = scan_field(fld_focus, tmp, NULL, NULL);

	  if (valid && (Preferences.units == POS_UTM)) {
	    /* lat and lon actually contain easting and northing! */
	    double	easting	 = lat;
	    double	northing = lon;

	    err = Convert_UTM_To_Geodetic(utm_zone,
                        	(utm_belt >= 'N') ? 'N' : 'S' /* Hemisphere */,
                        	easting,
                        	northing,
                        	&lat,
                        	&lon);	    

	    if (err != UTM_NO_ERROR) {
	      /* 
	       * Conversion from UTM not possible!
	       */
	      SetFieldText(StatusField, InputErrMsg, false, true);
	      break;
	    }

	    lat = rad2deg(lat);
	    lon = rad2deg(lon);
	  } /* if (valid && (Preferences.units == POS_UTM)) */
      } /* switch (eventP->data.menu.itemID) */
  
      switch (eventP->data.menu.itemID) {

	case Units0Menu:
	  /* New Format: DDD.DDDDDDD */
	  if (valid) {
	    Preferences.units = POS_D;
	    insert_position(lat, lon, NULL, NULL);
	  }
	  handled = false;
	  break;

	case Units1Menu:
	  /* New Format: DDDMM.MMMM' */
	  if (valid) {
	    Preferences.units = POS_DM;
	    insert_position(lat, lon, NULL, NULL);
	  }
	  handled = false;
	  break;

	case Units2Menu:
	  /* New Format: DDDMM'SS.SSS" */
	  if (valid) {
	    Preferences.units = POS_DMS;
	    insert_position(lat, lon, NULL, NULL);
	  }
	  handled = false;
	  break;

	case Units3Menu:
	  /* New Format: UTM */
	  if (valid) {
	    /* save current unit setting */
	    UInt16	unit_old = Preferences.units;
	  
	    /* new unit: UTM */
	    Preferences.units = POS_UTM;	    
	    
	    /* Note: easting/northing are returned in lat/lon. */
	    valid = insert_position(lat, lon, &lat, &lon);
            
	    if (!valid) {
	      /* 
	       * Conversion to UTM not possible,
	       *   revert to current unit setting!
	       */
	      Preferences.units = unit_old;
	      SetFieldText(StatusField, InputErrMsg, false, true);
	    } else {            
	      /* get UTM field from UI */
              scan_zone(&utm_zone, &utm_belt);
	    }
	  }
	  handled = false;
	  break;
      } /* switch (eventP->data.menu.itemID) */

      /* show objects depending on selected position units */
      if (Preferences.units == POS_UTM) {
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, ZoneLabel));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, ZoneField));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, EastLabel));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, NorthLabel));
	FrmHideObject(frmP, FrmGetObjectIndex(frmP, LatLabel));
	FrmHideObject(frmP, FrmGetObjectIndex(frmP, LonLabel));       
      } else {
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, LatLabel));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, LonLabel));       
	FrmHideObject(frmP, FrmGetObjectIndex(frmP, ZoneLabel));
	FrmHideObject(frmP, FrmGetObjectIndex(frmP, ZoneField));
	FrmHideObject(frmP, FrmGetObjectIndex(frmP, EastLabel));
	FrmHideObject(frmP, FrmGetObjectIndex(frmP, NorthLabel));
      } /* Preferences.units == POS_UTM */

      if (valid) {
        SetFieldText(StatusField, InputOkMsg, false, true);
      }
      
      /* update form after show/hide objects */
      FrmUpdateForm(MainForm, frmRedrawUpdateCode);
    
    /* case menuEvent */
    break;

   case ctlSelectEvent:
     switch (eventP->data.ctlSelect.controlID) {
       case GetButton:
         /* "Download Map" Button */
	 FrmSetFocus(frmP, noFocus);
 
 	 /* Update status output */
	 SetFieldText(StatusField, ClrMsg, false, true);
	 valid = scan_field(fld_focus, tmp, NULL, NULL);

	 if (valid && Preferences.units == POS_UTM) {
	   /*
	    * Convert UTM to geodetic coordinates:
	    *   lat and lon actually contain easting and northing!
	    */
	   double	easting	 = lat;
	   double	northing = lon;

	   err = Convert_UTM_To_Geodetic(utm_zone,
                        	(utm_belt >= 'N') ? 'N' : 'S' /* Hemisphere */,
                        	easting,
                        	northing,
                        	&lat,
                        	&lon);	    

	   if (err != UTM_NO_ERROR) {
	     /* 
	      * Conversion from UTM failed,
	      *   invalid input! Download will not be started.
	      */
	     SetFieldText(StatusField, InputErrMsg, false, true);
	     break;
	   }

	   lat = rad2deg(lat);
	   lon = rad2deg(lon);
	 }
	 
	 /* save current position to preferences */
	 Preferences.lat = lat;
	 Preferences.lon = lon;
  
         /* save selected map size to preferences */
	 scan_mapsize(&Preferences.width, &Preferences.height);

	 /* Update status output */
	 SetFieldText(StatusField, InputOkMsg, false, true);

#ifdef VERBOSE	 
	 /* Display results (debugging only) */
	 {
	   Char lat_str[40];
	   Char lon_str[40];
	   Char tmp[40];
	   Char *pt_pos;		/* decimal point pos. */
	   RectangleType	rect;	/* rectangle structure */
	   
	   RctSetRectangle(&rect, 0, 110, 160 /* w */, 30 /* h */);
           WinEraseRectangle(&rect, 0);
	   
	   format_number(lat, NUM_DIGITS, lat_str);
	   format_number(lon, NUM_DIGITS, lon_str);
	   	   
	   /* Location */
	   StrPrintF(tmp, "Lat: %s Lon: %s", lat_str, lon_str);
	   WinDrawChars(tmp, StrLen(tmp), 0, 110);
	 
	   /* Scale and Size */
	   StrPrintF(tmp, "Scale: %lu Size: %u x %u",
	     Preferences.scale,
	     Preferences.width,
	     Preferences.height);
	   WinDrawChars(tmp, StrLen(tmp), 0, 125);
         }
#endif
	 /* read current parameters from preferences */
         map.lat    = Preferences.lat;
         map.lon    = Preferences.lon;
         map.scale  = Preferences.scale;
         map.width  = Preferences.width;
         map.height = Preferences.height;
	
	 /* Start Progress Dialog */
         prgP = PrgStartDialogV31("Fetching Map...", callbackFunction);
         
	 if (!gVfsManagerFtr || (check_dir(&volRefNum) != 0)) {
           FrmHideObject(frmP, FrmGetObjectIndex(frmP, DestLabel));
           FrmHideObject(frmP, FrmGetObjectIndex(frmP, DestPopup));
           Preferences.use_vfs = false;	 
	 }
	 
	 /* now fetch the map! */
	 downloadMap(&map, prgP, volRefNum, Preferences.use_vfs);
	 
	 /* End Progress Dialog */
	 PrgStopDialog(prgP, false);

	 SetFieldText(StatusField, DoneMsg, false, true);
         handled = true;
	 break;
  
       default:
	 break;
     } /* switch (eventP->data.ctlSelect.controlID) */
     break;
     
   case popSelectEvent:
     sel = eventP->data.popSelect.selection;
     frmP = FrmGetActiveForm();
     FrmSetFocus(frmP, noFocus);
     
     switch (eventP->data.popSelect.listID) {
       /* Map scale pop-up list */
       case ScaleList:
	 Preferences.scale = ScaleMappings[sel];
	 handled = false;
	 break;

       /* Map storage location pop-up list */
       case DestList:
         Preferences.use_vfs = (sel == 1) ? true : false;
	 handled = false;
	 break;
	 
     }
     /* case popSelectEvent */
     break;
  
  case frmSaveEvent:
    valid = scan_field(fld_focus, tmp, NULL, NULL);

    if (valid && Preferences.units == POS_UTM) {
      /* lat and lon actually contain easting and northing! */
      double	easting	 = lat;
      double	northing = lon;

      err = Convert_UTM_To_Geodetic(utm_zone,
                           (utm_belt >= 'N') ? 'N' : 'S' /* Hemisphere */,
                           easting,
                           northing,
                           &lat,
                           &lon);	    

      if (err != UTM_NO_ERROR) {
	/* 
	 * Coordinates will not be saved
	 *   in Application Preferences.
	 */
	break;
      }

      lat = rad2deg(lat);
      lon = rad2deg(lon);
    }
    
    /* save current position to preferences */
    Preferences.lat = lat;
    Preferences.lon = lon;
    
    /* save selected map size to preferences */
    scan_mapsize(&Preferences.width, &Preferences.height);
    
    handled = false;
    break;


  case frmCloseEvent:
    /* deselect current field */
    fld_focus = fld_focus_old = 0;

    /* Deallocate Fields' Text Memory */
    if (Preferences.units == POS_UTM) {
      SetFieldText(ZoneField, NULL, false, true);
    }
    SetFieldText(LatitudeField, NULL, false, true);
    SetFieldText(LongitudeField, NULL, false, true);
    SetFieldText(WidthField, NULL, false, true);
    SetFieldText(HeightField, NULL, false, true);
    SetFieldText(StatusField, NULL, false, true);
    
    MainFormDeinit(frmP);
    handled = false;
    break;

  default:
    break;
  } /* switch (eventP->eType) */
  
  /* save current focus */  
  fld_focus_old = fld_focus;

  return handled;
}
