/*****************************************************************************
 *
 * $RCSfile: geo.c,v $
 *
 * GPS4Palm Geodesic/Geometric Utility Functions
 *
 * This program is Copyright (C) 07/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2004/04/10 10:43:09 $
 *
 * $Revision: 1.1 $
 *
 ****************************************************************************/

/*****************************************************************************
 * 1 Latitude : 111120m
 * 1' Latitude : 1852m = 1nmi
 *
 * 1 Longitude: 111120m * cos(lat)
 * 1' Longitude: 1852m * cos(lat)
 *
 * d_map : distance on map (in m)
 * d_act : distance in real world (actual distance) (in m)
 * r_map : map resolution (in dpi)
 *
 * d_map = d_act / scale
 * 
 * d_screen = d_map * r_map * (1/0.0254) in/m
 *
 ****************************************************************************/

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "MathLib.h"
#include "geo.h"
#include "fp.h"

/* SQR: square, i.e. x^2 */
#define SQR(x)	((x) * (x))

/* EPS: epsilon, i.e. a really small number greater than zero */
#define EPS	0.00000000000000000000005

/*
 * PIXELFACT = r_map [dpi] * 1/0.0254 in/m
 *           = 72 dpi * 1/0.0254 in/m = 2834.65 pixel/m
 */


/*****************************************************************************
 * FUNCTION:	Lat_To_UTM_Belt
 *
 * DESCRIPTION:	Calculate UTM Belt from latitude
 *
 * PARAMETERS:	lat	-- latitude in degrees
 *
 * RETURNED:	UTM Belt Designator
 ****************************************************************************/
char Lat_To_UTM_Belt(double lat)
{
  Int16 	belt;
  const char	utm_belt[] = "CDEFGHJKLMNPQRSTUVWX";

  /*
   * range check actually done by Convert_Geodetic_To_UTM(),
   *   so we just make sure to keep in the valid range of UTM belts
   */
  lat = (lat > 84.0) ? 84.0 : lat;
  lat = (lat < -79.0) ? -79.0 : lat; 
   
  /* correction for zone X which extends from 72 to 84 */
  if (lat >= 80.0 && lat < 84.5) {
    lat =  72.0;
  }
  
  belt = (Int16) floor(lat / 8.0);
  belt += 10;
  return utm_belt[belt];
}


/*****************************************************************************
 * FUNCTION:	semi2deg 
 *
 * DESCRIPTION:	Convert semicircles to degrees
 *		degrees = semicircles * (180 / 2^31)
 *
 * PARAMETERS:	semi	-- angle in semicircles
 *
 * RETURNED:	angle in degrees
 ****************************************************************************/
double semi2deg(Int32 semi) {
  return ((double)(semi * (180.0/(1L<<31))));
}


/*****************************************************************************
 * FUNCTION:	deg2semi 
 *
 * DESCRIPTION:	Convert degrees to semicircles
 *		semicircles = degrees * (2^31 / 180)
 *
 * PARAMETERS:	deg	-- angle in degrees
 *
 * RETURNED:	angle in semicircles
 ****************************************************************************/
Int32 deg2semi(double deg)
{
  return ((Int32)(deg * ((1L<<31)/180.0)));
}


/*****************************************************************************
 * FUNCTION:	rad2deg
 *
 * DESCRIPTION:	Convert radians to degrees
 *
 * PARAMETERS:	rad	-- angle in radians
 *
 * RETURNED:	angle in degrees
 ****************************************************************************/
double rad2deg(double rad)
{
  return (rad / M_PI * 180.0);
}


/*****************************************************************************
 * FUNCTION:	deg2rad
 *
 * DESCRIPTION:	Convert degrees to radians
 *
 * PARAMETERS:	deg	-- angle in degrees
 *
 * RETURNED:	angle in radians
 ****************************************************************************/
double deg2rad(double deg)
{
  return (deg / 180 * M_PI);
}


/*****************************************************************************
 * FUNCTION:	deg_to_dm_str 
 *
 * DESCRIPTION:	Convert degrees to degrees and minutes string
 *		for latitude and longitude
 *
 * PARAMETERS:	lat		-- latitude in degrees
 *		lon		-- longitude in degrees
 *		lat_dm_str	-- string buffer for returning latitude
 *				   in degrees and minutes
 *		lon_dm_str	-- string buffer for returning longitude
 *				   in degrees and minutes
 *
 * RETURNED:	-
 ****************************************************************************/
void deg_to_dm_str(double lat, double lon, char *lat_dm_str, char *lon_dm_str)
{
  double	frac;		/* fractional part */
  double	ipart;		/* integer part */
  
  /* Latitude */
  if (lat < 0) {
    StrCopy(lat_dm_str, "-");
    lat = -lat;
  } else {
    StrCopy(lat_dm_str, "");
  }

  /* degrees */
  frac = modf(lat, &ipart) * 60.0;
  StrPrintF(&lat_dm_str[StrLen(lat_dm_str)], "%02u", (UInt16)ipart);
  StrCat(lat_dm_str, "");

  /* minutes */
  format_number(frac, 4, &lat_dm_str[StrLen(lat_dm_str)]);
  StrCat(lat_dm_str, "\'");
  
  /* Longitude */
  if (lon < 0) {
    StrCopy(lon_dm_str, "-");
    lon = -lon;
  } else {
    StrCopy(lon_dm_str, "");
  }

  /* degrees */
  frac = modf(lon, &ipart) * 60.0;
  StrPrintF(&lon_dm_str[StrLen(lon_dm_str)], "%03u", (UInt16)ipart);
  StrCat(lon_dm_str, "");

  /* minutes */
  format_number(frac, 4, &lon_dm_str[StrLen(lon_dm_str)]);
  StrCat(lon_dm_str, "\'");

}

/*****************************************************************************
 * FUNCTION:	deg_to_dms_str 
 *
 * DESCRIPTION:	Convert degrees to degrees, minutes and seconds string
 *		for latitude and longitude.
 *
 * PARAMETERS:	lat		-- latitude in degrees
 *		lon		-- longitude in degrees
 *		lat_dms_str	-- string buffer for returning latitude in
 *				   degrees, minutes and seconds
 *		lon_dms_str	-- string buffer for returning longitude in
 *				   degrees, minutes and seconds
 *
 * RETURNED:	-
 ****************************************************************************/
void deg_to_dms_str(double lat, double lon, char *lat_dms_str, char *lon_dms_str)
{
  double	sec;		/* seconds of arc */
  double	frac;		/* fractional part */
  double	ipart;		/* integer part */

  /* Latitude */
  if (lat < 0) {
    StrCopy(lat_dms_str, "-");
    lat = -lat;
  } else {
    StrCopy(lat_dms_str, "");
  }

  /* degrees */
  frac = modf(lat, &ipart) * 60.0;
  StrPrintF(&lat_dms_str[StrLen(lat_dms_str)], "%02u", (UInt16)ipart);
  StrCat(lat_dms_str, "");

  /* minutes */
  sec = modf(frac, &ipart) * 60.0;
  StrPrintF(&lat_dms_str[StrLen(lat_dms_str)], "%02u", (UInt16)ipart);
  StrCat(lat_dms_str, "'");

  /* seconds */
  format_number(sec, 3, &lat_dms_str[StrLen(lat_dms_str)]);
  StrCat(lat_dms_str, "\"");
  
  /* Longitude */
  if (lon < 0) {
    StrCopy(lon_dms_str, "-");
    lon = -lon;
  } else {
    StrCopy(lon_dms_str, "");
  }
  
  /* degrees */
  frac = modf(lon, &ipart) * 60.0;
  StrPrintF(&lon_dms_str[StrLen(lon_dms_str)], "%03u", (UInt16)ipart);
  StrCat(lon_dms_str, "");

  /* minutes */
  sec = modf(frac, &ipart) * 60.0;
  StrPrintF(&lon_dms_str[StrLen(lon_dms_str)], "%02u", (UInt16)ipart);
  StrCat(lon_dms_str, "'");

  /* seconds */
  format_number(sec, 3, &lon_dms_str[StrLen(lon_dms_str)]);
  StrCat(lon_dms_str, "\"");

}
