/*****************************************************************************
 *
 * $RCSfile: DataExchange.c,v $
 *
 * GPS4Palm Data Exchange using Garmin Protocol
 *
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2004-12-14 20:10:47 $
 *
 * $Revision: 1.12 $
 *
 * $Log: DataExchange.c,v $
 * Revision 1.12  2004-12-14 20:10:47  mp
 * fixed alert
 *
 * Revision 1.11  2004/12/08 20:51:52  mp
 * removed debug output
 *
 * Revision 1.10  2004/12/06 21:11:53  mp
 * implemented get_next_trk()
 *
 * Revision 1.9  2004/12/04 13:35:06  mp
 * implemented NumTrkPackets() and store_trk()
 *
 * Revision 1.8  2004/11/27 10:16:08  mp
 * replaced ErrFatalDisplay() by Die() for memory allocation failures
 *
 * Revision 1.7  2004/11/24 21:05:24  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.6  2004/11/23 17:54:06  mp
 * removed unused variables
 *
 * Revision 1.5  2004/11/17 19:46:33  mp
 * modified interface to Data.c ( PackWaypoint(), PackRouteHdr(), and
 * PackRouteWpt() ), removed debug output, added exception handling for
 * creation of new route/waypoint
 *
 * Revision 1.4  2004/11/11 18:48:47  mp
 * fixed triggering of table update, added TermStr()
 *
 * Revision 1.3  2004/06/23 18:32:16  mp
 * updated comments
 *
 * Revision 1.2  2004/06/23 18:25:39  mp
 * added NumRtePackets(), fixed Route transfers
 *
 * Revision 1.1  2004/04/29 20:11:35  mp
 * initial version
 *
 *
 ****************************************************************************/
#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include <SerialMgrOld.h>
#include "ResourceDefines.h"
#include "gpslib.h"
#include "Data.h"
#include "DataExchange.h"
#include "WaypointForm.h"
#include "RouteForm.h"
#include "TrackForm.h"
#include "Utils.h"
#include "Serial.h"
#include "Utils.h"
#include "common.h"

/* Global Variables */
extern PrefsType	gPrefs;			/* application preferences */

/* Static Functions */
static void TermStr(char *str, UInt16 len)		DATAEXCHANGE_SECTION;


/****************************************************************************/
/**
 * \brief	Callback function to retrieve waypoint from database
 *		for transmission to host.
 *              
 *
 * \param	current		current waypoint count (zero-based)
 * \param	format		Garmin Object format ID
 *				   (only G104Type supported)
 * \param	size		size of structure returned by get_next_wpt
 *
 * \note	The function's prototype is defined by gpslib
 *		  (GetNextWptCallbackProc).
 *
 * \return	Pointer to Waypoint Object structure,
 *		NULL if waypoint transfer done or wrong protocol ID.
 ****************************************************************************/
GarminObject *get_next_wpt(UInt16 current, LocalFormatID format, 
  unsigned char *size)
{
  static GarminObject	wpt;			/* waypoint (Garmin packet) */
  UInt16		rec = 0;		/* record nr */
  static MemHandle	waypointDBEntry = NULL;	/* record handle */
  waypoint_t		waypoint;		/* unpacked waypoint */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
  Int16			i = 0;			/* loop index */
  UInt32		tmp;			/* var. for conversion */
  
  if (format == G104Type) {
    /* Init waypoint */
    MemSet((Char *)&wpt.d104.ident, 6, 0);
    MemSet((Char *)&wpt.d104.dst, sizeof(wpt.d104.dst), 0);
    MemSet((Char *)&wpt.d104.cmnt, sizeof(wpt.d104.cmnt), 0);
    
    /* seek to record */
    DmSeekRecordInCategory(gWaypointDB, &rec /* index */, current /* offset */,
      dmSeekForward, dmAllCategories);
    
    if (waypointDBEntry) {
      MemHandleUnlock(waypointDBEntry);
      waypointDBEntry = NULL;
    }
    
    /* get record */
    waypointDBEntry = DmQueryNextInCategory(gWaypointDB, &rec, 
      dmAllCategories);

    /* unpack record */
    packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
    UnpackWaypoint(&waypoint, packed_waypoint);
    
    /* Copy Identifier */
    StrNCopy(wpt.d104.ident, waypoint.ident, 6);
    
    /* Copy Latitude and Longitude */
    tmp = waypoint.posn.lat;
    for (i=0; i<sizeof(wpt.d104.lat); i++) {
      wpt.d104.lat[i] = (G_byte)(tmp & 0xFF);
      tmp >>= 8;
    }
    
    tmp = waypoint.posn.lon;
    for (i=0; i<sizeof(wpt.d104.lon); i++) {
      wpt.d104.lon[i] = (G_byte)(tmp & 0xFF);
      tmp >>= 8;
    }
    
    /* 
     * Note: Proxmity Distance only valid during
     *       Proximity Waypoint Transfer Protocol
     */
    
    /* Copy Symbol */
    tmp = waypoint.smbl;
    for (i=0; i<sizeof(wpt.d104.smbl); i++) {
      wpt.d104.smbl[i] = (G_byte)(tmp & 0xFF);
      tmp >>= 8;
    }
   
    /* Copy Display Mode */
    wpt.d104.dspl = (G_byte)waypoint.dspl;
    
    /* Copy Comment */
    StrNCopy(wpt.d104.cmnt, waypoint.cmnt, 40);
    
  } else {
    /* unknown protocol */
    return NULL;
  }

  *size = sizeof(wpt.d104);
  
  /* last waypoint sent? */
  if (current == DmNumRecordsInCategory(gWaypointDB, dmAllCategories))
    return NULL;

  return &wpt;
}


/****************************************************************************/
/**
 * \brief	Callback function to retrieve route from database
 *		for transmission to host.
 *              
 *
 * \param	current		current item number (header or waypoint,
 *				  first=0)
 * \param	format		Garmin Object format ID
 *			 	  (only G104Type and G201Type supported)
 * \param	size		size of structure returned by get_next_rte
 * \param	h		True if current packet is a Route Header,
 *				  else false
 * \param	eor		True if current Route Waypoint is the end of
 *				  the current route, else false
 *
 * \note	The function's prototype is defined by gpslib
 *		  (GetNextRteCallbackProc).
 *
 * \return	Pointer to Waypoint/Route Object structure,
 *		NULL if transfer done or wrong protocol ID.
 ****************************************************************************/
GarminObject *get_next_rte(UInt16 current, LocalFormatID format, 
  unsigned char *size, Boolean * h, Boolean * eor)
{
  Err			err;			/* error code */
  static GarminObject	wpt;			/* waypoint (Garmin packet) */
  static GarminObject	rte;			/* route (Garmin packet) */ 
  static route_t	*route;			/* route */
  static MemHandle	waypointDBEntry = NULL;	/* waypoint record handle */
  static MemHandle	routeDBEntry = NULL;	/* route record handle */
  static UInt16		rte_nr = 0;		/* route number */
  static UInt16		wpt_nr = 0;		/* waypoint number */
  UInt16		rec = 0;		/* record nr */
  waypoint_t		waypoint;		/* unpacked waypoint */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
  Int16			i = 0;			/* loop index */
  UInt32		tmp;			/* var. for conversion */

  if (format == G104Type) {
/*  
 *  StrPrintF(buf, "%d/%d ", wpt_nr, route->items); 
 *  WinDrawChars(buf, StrLen(buf), cnt * 20, 15);
 *  cnt++;
 */
     
    /* Init waypoint */
    MemSet((Char *)&wpt.d104.ident, 6, 0);
    MemSet((Char *)&wpt.d104.dst, sizeof(wpt.d104.dst), 0);
    MemSet((Char *)&wpt.d104.cmnt, sizeof(wpt.d104.cmnt), 0);
        
    if (waypointDBEntry) {
      /* unlock old record handle */
      MemHandleUnlock(waypointDBEntry);
      waypointDBEntry = NULL;
    }

    err = DmFindRecordByID(gWaypointDB,
      route->wpt_rec_id[wpt_nr], &rec);

    if (err) {
      /* unique ID not found */
      *eor = false;
      return NULL;
    }
    
    /* get record */
    waypointDBEntry = DmQueryNextInCategory(gWaypointDB, &rec, 
      dmAllCategories);

    /* unpack record */
    packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
    UnpackWaypoint(&waypoint, packed_waypoint);
    
    /* Copy Identifier */
    StrNCopy(wpt.d104.ident, waypoint.ident, 6);
    
    /* Copy Latitude and Longitude */
    tmp = waypoint.posn.lat;
    for (i=0; i<sizeof(wpt.d104.lat); i++) {
      wpt.d104.lat[i] = (G_byte)(tmp & 0xFF);
      tmp >>= 8;
    }
    
    tmp = waypoint.posn.lon;
    for (i=0; i<sizeof(wpt.d104.lon); i++) {
      wpt.d104.lon[i] = (G_byte)(tmp & 0xFF);
      tmp >>= 8;
    }
    
    /* 
     * Note: Proxmity Distance only valid during
     *       Proximity Waypoint Transfer Protocol
     */
    
    /* Copy Symbol */
    tmp = waypoint.smbl;
    for (i=0; i<sizeof(wpt.d104.smbl); i++) {
      wpt.d104.smbl[i] = (G_byte)(tmp & 0xFF);
      tmp >>= 8;
    }
   
    /* Copy Display Mode */
    wpt.d104.dspl = (G_byte)waypoint.dspl;
    
    /* Copy Comment */
    StrNCopy(wpt.d104.cmnt, waypoint.cmnt, 40);
   
    /* size of waypoint struct */
    *size = sizeof(wpt.d104);

    /* last waypoint sent? set end-of-route flag */
    *eor = (wpt_nr == route->items-1) ? true : false;

    /* protocol error: past end of last waypoint */
    if (wpt_nr == route->items) {
      *h = false;
      return NULL;
    }
    
    /* increment waypoint count */
    wpt_nr++;

    /* current packet is a Route Waypoint */
    *h = false;

    return &wpt;
  
  } else if (format == G201Type) {
    /* route header, reset waypoint count */
    wpt_nr = 0;
    
    if (current == 0) {
      /* start of route transfer */
      rte_nr = 0;
    }
/*     
 *  StrPrintF(buf, "%d ", rte_nr); 
 *  WinDrawChars(buf, StrLen(buf), cnt * 15, 30);
 *  cnt++;
 */
    /* Init route */
    MemSet((Char *)&rte.d201.cmnt, sizeof(rte.d201.cmnt), 0);

    /* seek to record */
    DmSeekRecordInCategory(gRouteDB, &rec /* index */, rte_nr /* offset */,
      dmSeekForward, dmAllCategories);
    
    if (routeDBEntry) {
      /* unlock old record handle */
      MemHandleUnlock(routeDBEntry);
      routeDBEntry = NULL;
    }

    /* get record */
    routeDBEntry = DmQueryNextInCategory(gRouteDB, &rec, 
      dmAllCategories);
    
    route = MemHandleLock(routeDBEntry);
    
    /* Copy Route Number */
    rte.d201.nmbr = route->nmbr;
    
    /* Copy Comment */
    StrNCopy(rte.d201.cmnt, route->cmnt, 20);
        
    /* size of route struct */
    *size = sizeof(rte.d201);

    /* last route sent? */
    if (rte_nr == DmNumRecordsInCategory(gRouteDB, dmAllCategories)) {
      return NULL;
    }
    
    /* increment route counter */
    rte_nr++;

    /* current packet is a Route Header */
    *h = true;
    
    /* no End of Route */
    *eor = false;
    
    return &rte;
    
  } else {
    /* unknown protocol */
    return NULL;
  }
  
}

/****************************************************************************/
/**
 * \brief	Callback function to to retrieve track from database
 *		for transmission to host.
 *
 * \param	current		index of current element (first=0)
 * \param	format		format of element requested
 * \param	size		pointer to store size of element returned
 * \param	eot		true if the current item is the
 *				last element of current track
 *
 * \return 	pointer on element returned (Garmin format)
 ****************************************************************************/
GarminObject *get_next_trk(UInt16 current, LocalFormatID format,
  byte *size, Boolean * eot)
{
  static GarminObject	trkh;			/* track header (Garmin) */
  static GarminObject	trkp;			/* track point (Garmin) */
  trk_point_t		*trkpoint;		/* track point */
  trk_header_t		*trkheader;		/* track header */
  static UInt16		trkh_index;		/* track header index */
  static UInt16		trkp_index;		/* track point index */
  static DmOpenRef	db_ref;			/* data base reference */
  static UInt16		num_points;		/* number of track points */
  static UInt16		point = 0;		/* track point number */
  static UInt16		track;			/* track number */
  MemHandle		trkhDBEntry;		/* trk. header record handle */
  MemHandle		trkpDBEntry;		/* trk. point record handle */
  Char			dbname[dmDBNameLength];	/* data base name */
  UInt32		trackID;		/* track ID */
  static UInt16		num_tracks;		/* number of tracks */
  UInt32		tmp;			/* var. for conversion */
  UInt16		i;			/* loop index */
  
  if ((format != G300Type) && (format != G310Type)) {
    /* unknown protocol */
    return NULL;
  }
  
  if ((format == G310Type) || (current == 0) || (point == 0)) {
    /* track header */
    
    if (current == 0) {
      /* get number of all tracks in Track Header DB */
      num_tracks = DmNumRecordsInCategory(gTrackHdrDB, dmAllCategories);

      trkh_index = 0;
      point      = 0;
      track      = 0;
    }

    /* loop until non-empty track DB found */
    do {
      if (track == num_tracks) {
	*eot = true;
	*size = 0;
	return NULL;
      }

      trkp_index = 0;

      /* query next track */
      trkhDBEntry = DmQueryNextInCategory(gTrackHdrDB, &trkh_index, 
	dmAllCategories);

      trkhDBEntry = DmQueryRecord(gTrackHdrDB, trkh_index);

      /* get unique Track ID */
      DmRecordInfo(gTrackHdrDB, trkh_index, NULL /* *attrP */,
	&trackID /* *uniqueIDP */, NULL /* *chunkIDP */);      

      /* create track data base name */
      TrackDBName(dbname, trackID);

      /* open track point data base */
      db_ref = DmOpenDatabase(0 /* cardNo */,
	DmFindDatabase(0 /* cardNo */, dbname), dmModeReadOnly);

      trkh_index++;
      track++;

      if (db_ref) {
	/* open succeeded, get number of track points */
	num_points = DmNumRecordsInCategory(db_ref, dmAllCategories);

	if (num_points == 0) {
          /* close track point data base */
          DmCloseDatabase(db_ref);
	}

      } /* if (db_ref) */
    } while (num_points == 0);
  
    if (format == G310Type) {
      trkheader = (trk_header_t *)MemHandleLock(trkhDBEntry);
      
      /* copy track header for transfer to host */
      trkh.d310.dspl  = trkheader->dspl;
      trkh.d310.color = trkheader->color;
      StrNCopy(trkh.d310.trk_ident, trkheader->trk_ident, 49);
      
      MemHandleUnlock(trkhDBEntry);
      *size = sizeof(trkh.d310);
      return &trkh;
    }
  }
  
  if (format == G300Type) {
    /* track point */

    /* query next track point */
    trkpDBEntry = DmQueryNextInCategory(db_ref, &trkp_index, 
      dmAllCategories);

    trkpDBEntry = DmQueryRecord(db_ref, trkp_index);    

    trkpoint = (trk_point_t *)MemHandleLock(trkpDBEntry);

    /* Copy Latitude and Longitude */
    tmp = trkpoint->posn.lat;
    for (i=0; i<sizeof(trkp.d300.lat); i++) {
      trkp.d300.lat[i] = (G_byte)(tmp & 0xFF);
      tmp >>= 8;
    }
    
    tmp = trkpoint->posn.lon;
    for (i=0; i<sizeof(trkp.d300.lon); i++) {
      trkp.d300.lon[i] = (G_byte)(tmp & 0xFF);
      tmp >>= 8;
    }
    
    /* Copy Time */
    tmp = trkpoint->time;
    for (i=0; i<sizeof(trkp.d300.time); i++) {
      trkp.d300.time[i] = (G_byte)(tmp & 0xFF);
      tmp >>= 8;
    }

    trkp.d300.new_trk = trkpoint->new_trk;
    
    /* unlock record handle */
    MemHandleUnlock(trkpDBEntry);
    
    trkp_index++;
    
    /* last track point sent? */
    if (point == num_points-1) {
      /* close track point data base */
      DmCloseDatabase(db_ref);
      
      /* set end-of-track flag */
      *eot  = true;
      point = 0;
      
    } else {
      *eot = false;
      point++;
    }
    
    *size = sizeof(trkp.d300);
    return &trkp;
    
  }

  return NULL;  
}

  
/****************************************************************************/
/**
 * \brief	Callback function to store waypoint to database upon
 *		reception from host.
 *
 *		If a waypoint with the same waypoint ID already exists,
 *		it is modified if required.
 *
 * \param	wpt		ptr to GarminObject
 * \param	total		total number of packets
 * \param	current		packet number (one-based), for the
 *				D104 protocol, this can also be used as the
 *				number of the current waypoint
 * \param	format		Garmin Object format ID
 * 				  (only G104Type supported)
 * \param	size		size of structure pointed to by wpt
 *		
 * \note	The function's prototype is defined by gpslib
 *		  (StoreWptCallbackProc).
 *
 * \return	0 - success,
 *		1 - failure
 ****************************************************************************/
Err store_wpt(GarminObject* wpt, UInt16 total, UInt16 current,
  LocalFormatID format, byte size)
{
  waypoint_t	waypoint;		/* unpacked waypoint */
  UInt16	index;			/* record index */
  Int16		i;			/* loop index */
  MemHandle	h;			/* memory handle */
  Boolean	wpt_exists = false;	/* flag: waypoint already exists */
  
  if (format == G104Type) {
    /* Init waypoint */
    MemSet((Char *)waypoint.ident, 6, 0);
    waypoint.posn.lat = 0;
    waypoint.posn.lon = 0;
    waypoint.dst = 0.0;
    waypoint.smbl = 0;
    
    /* Copy Identifier */
    StrNCopy(waypoint.ident, wpt->d104.ident, 6);

    /* Copy Latitude and Longitude */
    for (i=sizeof(waypoint.posn.lat)-1; i>=0; i--) {
      waypoint.posn.lat <<= 8;
      waypoint.posn.lat |= (G_byte)wpt->d104.lat[i];
      waypoint.posn.lon <<= 8;
      waypoint.posn.lon |= (G_byte)wpt->d104.lon[i];
    }
    
    /* 
     * Note: Proxmity Distance only valid during
     *       Proximity Waypoint Transfer Protocol
     */
    
    /* Copy Symbol */
    for (i=sizeof(waypoint.smbl)-1; i>=0; i--) {
      waypoint.smbl <<= 8;
      waypoint.smbl |= (G_byte)wpt->d104.smbl[i];
    }
   
    /* Copy Display Mode */
    waypoint.dspl = (G_byte)wpt->d104.dspl;
    
    /* Copy Comment */
    for (i=0; i<40; i++) {
      waypoint.cmnt[i] = wpt->d104.cmnt[i];
    }
    TermStr(waypoint.cmnt, 40);

    /* Find sort position of new record in waypoint DB */  
    index = DmFindSortPosition(gWaypointDB, &waypoint, 0 /* newRecordInfo */,
      (DmComparF *)CompareWayptIdent /* DmComparF *compar */, 0 /* other */);

    if (index > 0) {
      /* check if waypoint with same ID already exists */
      packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
      
      h = DmQueryRecord(gWaypointDB, index - 1);
      packed_waypoint = MemHandleLock(h);
      if (StrNCompare(waypoint.ident, packed_waypoint->ident, 6) == 0) {
        wpt_exists = true;
	/* preserve proximity distance */
        waypoint.dst = packed_waypoint->dst;
	/* overwrite waypoint */
	index--;
      }
      MemHandleUnlock(h);
    }
    
    if (!wpt_exists) {
      h = DmNewRecord(gWaypointDB, &index, 
	1 /* size, will be set by PackWaypoint() */);
      
      if (h) {
        DmReleaseRecord(gWaypointDB, index, false /* dirty */);
      } else {
        Die("Cannot create waypoint record!");
      }
    }
    
    /* Pack Waypoint and store it in Database */
    PackWaypoint(&waypoint, index);    
 
    /* redraw the table */
    if ((current == total) && (FrmGetActiveFormID() == RouteForm))
      routeDrawTable();

    if ((current == total) && (FrmGetActiveFormID() == WaypointForm))
      wayptDrawTable();
    
    /* success */
    return 0;
  } else {
    /* failure (wrong protocol) */
    return gpslibErrParamErr;
  }
}


/****************************************************************************/
/**
 * \brief	Callback function to store route to database upon
 *		reception from host.
 *
 * \param	obj		ptr to GarminObject
 * \param	total		total count of items sent as indicated by gps 
 * \param	current		current item number (first=1)
 * \param	format		Garmin Object format ID
 *				  (only G104Type and G201Type supported)
 * \param	size		size of structure pointed to by rte
 *		
 * \note	The function's prototype is defined by gpslib
 *		  (StoreRteCallbackProc).
 *
 * \return	0 - success,
 *		1 - failure
 ****************************************************************************/
Err store_rte(GarminObject* obj, UInt16 total, UInt16 current,
  LocalFormatID format, byte size)
{
  Err			err;
  waypoint_t		waypoint;		/* unpacked waypoint */
  UInt32		wpt_rec_id;		/* unique wpt DB record ID */ 
  UInt16		wpt_index;		/* waypoint record index */
  static UInt16		rte_index;		/* route record index */
  route_t		route;			/* route */
  MemHandle		rteH;			/* route record handle */
  route_t		*rteP;			/* ptr to route record */
  MemHandle		h;			/* memory handle */
  UInt16		i;			/* loop index */
  Boolean		wpt_exists = false;	/* flag: waypoint exists */
  Boolean		rte_exists = false;	/* flag: route exists */

  if (format == G104Type) {
    /* store waypoint (as part of route) */

    /* Init waypoint */
    MemSet((Char *)waypoint.ident, 6, 0);
    waypoint.posn.lat = 0;
    waypoint.posn.lon = 0;
    waypoint.dst = 0.0;
    waypoint.smbl = 0;
    
    /* Copy Identifier */
    StrNCopy(waypoint.ident, obj->d104.ident, 6);

    /* Find sort position of new record in waypoint DB */  
    wpt_index = DmFindSortPosition(gWaypointDB, &waypoint,
      0 /* newRecordInfo */,
      (DmComparF *)CompareWayptIdent /* DmComparF *compar */, 0 /* other */);

    if (wpt_index > 0) {
      /* check if waypoint with same ID already exists */
      packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
      
      h = DmQueryRecord(gWaypointDB, wpt_index - 1);
      packed_waypoint = MemHandleLock(h);
      if (StrNCompare(waypoint.ident, packed_waypoint->ident, 6) == 0) {
        wpt_exists = true;
	/* preserve proximity distance */
        waypoint.dst = packed_waypoint->dst;
        /* overwrite waypoint */
	wpt_index--;
      }
      MemHandleUnlock(h);
    }

    if (!wpt_exists) {
      /* Copy Latitude and Longitude */
      for (i=sizeof(waypoint.posn.lat)-1; i>=0; i--) {
	waypoint.posn.lat <<= 8;
	waypoint.posn.lat |= (G_byte)obj->d104.lat[i];
	waypoint.posn.lon <<= 8;
	waypoint.posn.lon |= (G_byte)obj->d104.lon[i];
      }

      /* Copy Symbol */
      for (i=sizeof(waypoint.smbl)-1; i>=0; i--) {
	waypoint.smbl <<= 8;
	waypoint.smbl |= (G_byte)obj->d104.smbl[i];
      }

      /* Copy Display Mode */
      waypoint.dspl = (G_byte)obj->d104.dspl;

      /* Copy Comment */
      for (i=0; i<40; i++) {
	waypoint.cmnt[i] = obj->d104.cmnt[i];
      }
      TermStr(waypoint.cmnt, 40);

      /* create new waypoint record */
      h = DmNewRecord(gWaypointDB, &wpt_index, 
	1 /* size, will be set by PackWaypoint() */);

      if (h) {
        DmReleaseRecord(gWaypointDB, wpt_index, false /* dirty */);
      } else {
        Die("Cannot create waypoint record!");
      }
      
      /* Pack Waypoint and store it in Database */
      PackWaypoint(&waypoint, wpt_index);    

    }
    
    /* get record ID of waypoint */
    err = DmRecordInfo(gWaypointDB, wpt_index, NULL, &wpt_rec_id, NULL);
    
    /* add waypoint record ID to route record */
    PackRouteWpt(wpt_rec_id, rte_index);

    /* redraw the table */
    if ((current == total) && (FrmGetActiveFormID() == RouteForm))
      routeDrawTable();

    if ((current == total) && (FrmGetActiveFormID() == WaypointForm))
      wayptDrawTable();

    /* success */
    return 0;
  } else if (format == G201Type) {
    /* receiving route header */
    
    /* copy route number */
    route.nmbr = obj->d201.nmbr;
    
    /* copy route comment */
    MemSet((Char *)route.cmnt, 21, 0);
    StrNCopy(route.cmnt, obj->d201.cmnt, 20);
    TermStr(route.cmnt, 20);
    
    /* clear number of items (waypoints) */
    route.items = 0;
    
    /* Find sort position of new record in route DB */  
    rte_index = DmFindSortPosition(gRouteDB, &route, 0 /* newRecordInfo */,
      (DmComparF *)CompareRteNmbr /* DmComparF *compar */, 0 /* other */);

    if (rte_index > 0) {
      /* check if route with same number already exists */
      
      rteH = DmQueryRecord(gRouteDB, rte_index - 1);
      rteP = MemHandleLock(rteH);
      if (route.nmbr == rteP->nmbr) {
        rte_exists = true;
	/* overwrite route */
	rte_index--;
      }
      MemHandleUnlock(rteH);
    }

    if (!rte_exists) {
      /* create new route record */
      rteH = DmNewRecord(gRouteDB, &rte_index, 
	1 /* size, will be set by PackRoute() */);
    
      if (rteH) {
        DmReleaseRecord(gRouteDB, rte_index, false /* dirty */);
      } else {
        Die("Cannot create route record!");
      }
     
    }

    /* Store Route Header in Database */
    PackRouteHdr(&route, rte_index);
    
    /* success */
    return 0;
  } else {
    /* failure (wrong protocol) */
    return gpslibErrParamErr;
  }
}


/****************************************************************************/
/**
 * \brief	Callback function called when a track is received	
 *
 * \param	trk		pointer on a raw track point or track header
 *				received from gps
 * \param	total		total count of items sent as indicated by gps
 * \param	current		current item number (first=1)
 * \param	format		format of the object received
 * \param	size		size of object
 *
 * \return	error code (0 if no error)
 ****************************************************************************/
Err store_trk(GarminObject * trk, UInt16 total, UInt16 current,
  LocalFormatID format, byte size)
{
  UInt16		index;			/* track record index */
  MemHandle		trkhDBEntry;		/* track header record handle */
  trk_header_t		trkh;			/* track header */
  MemHandle		trkpDBEntry;		/* track point record handle */
  trk_point_t		trkp;			/* track point */
  UInt32		trackID;		/* track ID (unique) */
  Err			err;			/* error code */
  static DmOpenRef	db_ref;			/* track point DB reference */
  Char			dbname[dmDBNameLength];	/* data base name */
  Int16			i = 0;			/* loop index */
  
  if ((format != G300Type) && (format != G310Type)) {
    /* failure (wrong protocol) */
    return gpslibErrParamErr;
  } 
  
  if ((format == G310Type) || (current == 1)) {
    /* create and store track header */
    
    /* Limit number of Tracks */
    if ( DmNumRecordsInCategory(gTrackHdrDB, dmAllCategories) == 0x100 ) {
      FrmAlert(TrackCreateAlert);
      return gpslibErrBusy;
    }

    /* No particular order - add at the end of the database */
    index = dmMaxRecordIndex;

    trkhDBEntry = DmNewRecord(gTrackHdrDB, &index, sizeof(trk_header_t));

    if (!trkhDBEntry) {
      /* error occurred - fatal! */
      Die("Cannot create track header record!");
      return gpslibErrBusy;
    }

    /* release new record */
    DmReleaseRecord(gTrackHdrDB, index, false /* dirty */);
    
    if (format == G310Type) {
    
      if (current > 1) {
        /* not the first header, close previous track point DB */
        DmCloseDatabase(db_ref);
      }
    
      /* received track header */
      trkh.dspl  = trk->d310.dspl;
      trkh.color = trk->d310.color;
      StrNCopy(trkh.trk_ident, trk->d310.trk_ident, 49);
    } else {
      /* received track point without preceeding header */
      /* create header */
      trkh.dspl  = true;
      trkh.color = Default_Color;
      TrackIdent(trkh.trk_ident);
    }
    
    /* store track header to data base */
    StoreTrkHdr(&trkh, index);
    
    DmRecordInfo(gTrackHdrDB, index, NULL /* *attrP */,
      &trackID /* *uniqueIDP */, NULL /* *chunkIDP */);

    /* create data base name */
    TrackDBName(dbname, trackID);

    /* create track point data base */ 
    err = DmCreateDatabase(0 /* cardNo */, dbname,
      kCreatorId /* creator */, kTrkDBType /* type */, false /* resDB */);
      
    if (current != total) {
      /* track points will follow */
      db_ref = DmOpenDatabase(0 /* cardNo */,
        DmFindDatabase(0 /* cardNo */,  dbname), dmModeReadWrite);
    }
       
    if (current == total) { 
      /* close track point data base */
      DmCloseDatabase(db_ref);
      
      /* update Track Table */
      if (FrmGetActiveFormID() == TrackForm) {
	TablePtr		tblP;		/* table ptr */
	UInt16 			track_top = 0;	/* track table top */

	/* set top table row to ensure selected item is visible */
	tblP = (TablePtr)GetObjectFromActiveForm(TrackTable);

        if (index > TblGetNumberOfRows(tblP)-1)
          track_top = index-TblGetNumberOfRows(tblP)+1;

	trackDrawTable(track_top /* top */, &index);
      }
    }
    
    /* success */
    if (format == G310Type)
      return 0;
  } /* if ((format == G310Type) || (current == 1)) */
  
  if (format == G300Type) {
    /* receiving track point */

    /* copy track point */
    for (i=sizeof(trkp.posn.lat)-1; i>=0; i--) {
      trkp.posn.lat <<= 8;
      trkp.posn.lat |= (G_byte)trk->d300.lat[i];
      trkp.posn.lon <<= 8;
      trkp.posn.lon |= (G_byte)trk->d300.lon[i];
    }
    trkp.time        = 0;
    trkp.new_trk     = trk->d300.new_trk;
    
    /* add at the end of the database */
    index = dmMaxRecordIndex;

    /* create new track point record */
    trkpDBEntry = DmNewRecord(db_ref, &index, sizeof(trk_point_t));

    if (!trkpDBEntry) {
      Die("Cannot create track point record!");
      return gpslibErrBusy;
    }
 
    /* release new record */
    DmReleaseRecord(db_ref, index, false /* dirty */);

    /* store track point to data base */
    StoreTrkPoint(&trkp, index);

    /* update Track Table */
    if ((current == total) && (FrmGetActiveFormID() == TrackForm)) {
      TablePtr		tblP;		/* table ptr */
      Int32 		track_top = 0;	/* track table top */
     
      /* set top table row to ensure selected item is visible */
      tblP = (TablePtr)GetObjectFromActiveForm(TrackTable);
      
      if (index > TblGetNumberOfRows(tblP)-1)
        track_top = index-TblGetNumberOfRows(tblP)+1;

      trackDrawTable(track_top /* top */, &index);
    }
    
    /* success */
    return 0;
  } /* if (format == G300Type) */

  return gpslibErrParamErr;
}


/****************************************************************************/
/**
 * \brief	Get number of data packets for route transfer using
 *		Garmin protocol
 *
 * \return	Number of data packets (Route Headers + Route Waypoints) 	
 *
 ****************************************************************************/
UInt16 NumRtePackets(void)
{
  UInt16		num_routes;		/* nr. of routes */
  UInt16		num_packets;		/* nr. of packets */
  MemHandle		routeDBEntry;		/* record handle */
  route_t		*route;			/* route */
  UInt16		rec;			/* record nr */
  UInt16		i;			/* loop index */

  /* get number of all routes in Route DB */
  num_routes = DmNumRecordsInCategory(gRouteDB, dmAllCategories);
  
  num_packets = num_routes;
  
  rec = 0;
  
  /* seek to first record in Route DB */
  DmSeekRecordInCategory(gRouteDB, &rec /* index */, 0 /* offset */,
    dmSeekForward, dmAllCategories);
 
  /* get number of waypoints of all routes */
  for (i=0; i<num_routes; i++) {
  
    /* query next route */
    routeDBEntry = DmQueryNextInCategory(gRouteDB, &rec, 
      dmAllCategories);

    /* lock record */
    route = (route_t *)MemHandleLock(routeDBEntry);

    /* add number of route waypoints to packet count */
    num_packets += route->items;
	
    /* unlock record */
    MemHandleUnlock(routeDBEntry);
  }
  
  return num_packets;
}


/****************************************************************************/
/**
 * \brief	Get number of data packets for track transfer using
 *		Garmin protocol
 *
 * \return	Number of data packets (Track Headers + Track Waypoints) 	
 *
 ****************************************************************************/
UInt16 NumTrkPackets(void)
{
  UInt16		num_tracks;		/* nr. of tracks */
  UInt16		num_packets;		/* nr. of packets */
  DmOpenRef		db_ref;			/* data base reference */
  MemHandle		trkhDBEntry;		/* record handle */
  Char			dbname[dmDBNameLength];	/* data base name */
  UInt32		trackID;		/* track ID */
  UInt16		rec;			/* record nr */
  UInt16		i;			/* loop index */
  
  /* get number of all tracks in Track Header DB */
  num_tracks = DmNumRecordsInCategory(gTrackHdrDB, dmAllCategories);

  /* 
   * Note:
   * G310 Track Header currently not supported by emulated GARMIN device;
   * #packets = #track_points
   *
   * num_packets = num_tracks;
   */

  num_packets = 0;
 
  rec = 0;
  
  /* get number of track points of all tracks */
  for (i=0; i<num_tracks; i++) {

    /* query next route */
    trkhDBEntry = DmQueryNextInCategory(gTrackHdrDB, &rec, 
      dmAllCategories);

    trkhDBEntry = DmQueryRecord(gTrackHdrDB, rec);

    /* get unique Track ID */
    DmRecordInfo(gTrackHdrDB, rec, NULL /* *attrP */,
      &trackID /* *uniqueIDP */, NULL /* *chunkIDP */);      

    /* create track data base name */
    TrackDBName(dbname, trackID);

    /* open track point data base */
    db_ref = DmOpenDatabase(0 /* cardNo */,
      DmFindDatabase(0 /* cardNo */, dbname), dmModeReadOnly);

    rec++;

    if (db_ref) {
      /* open succeeded, get number of track points */
      num_packets += DmNumRecordsInCategory(db_ref, dmAllCategories);

      /* close DB */
      DmCloseDatabase(db_ref);

    } /* if (db_ref) */

  } /* for (i=0; i<num_tracks; i++) */
  
  return num_packets;
}


/****************************************************************************/
/**
 * \brief	Terminate string with null characters ('\0').
 *
 *		GPSMan seems to pad strings with spaces, which prevents
 *		proper packing of database records.
 *
 * \param	str	String pointer
 *
 * \param	len	String length	
 *
 ****************************************************************************/
static void TermStr(char *str, UInt16 len)
{
  UInt16	i;
  
  for (i = len-1; i>=0; i--) {
    if (str[i] == ' ' || str[i] == '\0') {
      str[i] = '\0';
    } else {
      break;
    }
  }
}
