/*****************************************************************************
 *
 * $RCSfile: FileselForm.c,v $
 *
 * GPS4Palm File Selector Form
 *   - search for map data bases and create directory data bases
 *   - show image files in table 
 *   - implement table scrolling, panning, and selection
 *   - implement file delete operation
 *
 * created: 2002-12-15
 *
 * History:
 * 2002-12-15:	first version
 * 2002-12-30:	added delete function, added Help->About Menu
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-16 09:57:17 $
 *
 * $Revision: 1.18 $
 *
 * $Log: FileselForm.c,v $
 * Revision 1.18  2005-05-16 09:57:17  mp
 * added display of size, date, and dimension of selected map on frmOpenEvent
 *
 * Revision 1.17  2005/05/14 11:40:18  mp
 * fixed bug in FileSelDeleteButton event handling
 *
 * Revision 1.16  2005/05/14 11:35:36  mp
 * added image dimensions display
 *
 * Revision 1.15  2005/05/13 18:33:10  mp
 * modified filename/filecard variable usage
 *
 * Revision 1.14  2005/02/26 14:52:59  mp
 * added map file deletion,
 * fixed table selection highlighting for some OS versions
 *
 * Revision 1.13  2005/02/25 18:40:34  mp
 * replaced gCardNo by gPrefs.filecard
 *
 * Revision 1.12  2005/02/19 20:05:21  mp
 * changed filename[32] to filename[FILENAMEMAXLEN]
 *
 * Revision 1.11  2005/02/19 14:36:50  mp
 * moved VFS Feature Set check to GPS.c
 *
 * Revision 1.10  2005/02/19 14:00:58  mp
 * added display of file creation date
 *
 * Revision 1.9  2005/02/19 12:18:32  mp
 * added VFS support and file size display, removed unused code
 *
 * Revision 1.8  2004/11/24 21:07:47  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.7  2004/11/23 17:53:39  mp
 * commented out unused variables
 *
 * Revision 1.6  2004/04/30 16:32:01  mp
 * replaced dopage()/doscroll() by common functions in Util.c
 *
 * Revision 1.5  2004/03/20 11:18:38  mp
 * changed scope of gTopRow
 *
 * Revision 1.4  2003/12/01 20:11:03  mp
 * removed old code, added AUTO button, USER/AUTO map selection
 *
 * Revision 1.3  2003/11/23 19:53:55  mp
 * moved CreateDatabase and DestroyDatabase to getdir.c
 *
 * Revision 1.2  2003/10/19 17:04:34  mp
 * removed/commented out table column <size>
 *
 * Revision 1.1  2003/10/18 16:29:39  mp
 * imported from piv, disabled Help->About and File->Delete functions
 *
 *
 ****************************************************************************/

#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include "ResourceDefines.h"
#include <Extensions/ExpansionMgr/VFSMgr.h>
#include "stringil.h"
#include "FileselForm.h"
#include "file_io.h"
#include "getdir.h"
#include "imginfo.h"
#include "Utils.h"
#include "common.h"

#define SORTPREFIX 4

/** Name of selected file (passed to/from FileselForm) */
#define SEL_FILENAME	gPrefs.filename

/** Card/Volume of selected file (passed to/from FileselForm) */
#define SEL_FILECARD	gPrefs.filecard

/* Global Variables */
extern PrefsType	gPrefs;			/* application preferences */
extern Boolean		gVfsManagerFtr;		/* VFS Feature Set available */
extern Boolean		gFtrSet35;		/* 3.5 Feature Set available */

/* Static Variables */
static Boolean		vfs_flag[10];		/* one entry per visible row */

/* Static Functions */
static void FileSelFormInit(FormPtr frmP)		FILESEL_SECTION;
static void FileSelFormDeinit(FormPtr frmP)		FILESEL_SECTION;
static void TableDrawCell(void *tblP, Int16 row,
  Int16 column, RectangleType *bounds)			FILESEL_SECTION;
static void DrawTable(UInt16 files, char **filelp,
                      char *filecard,Int32 top,
		      UInt16 *sel)			FILESEL_SECTION;
static UInt32 GetFileCreationDate(const char *name,
  UInt32 card)						FILESEL_SECTION;
static Err DeleteMap(const char *name,
  UInt32 card)						FILESEL_SECTION;

static void FileSelFormInit(FormPtr frmP)
{
#pragma unused(frmP)

  // warning-- don't do any drawing in this routine.
  // Also, don't call FrmSetFocus from here (it must be called *after*
  // FrmDrawForm)
}

static void FileSelFormDeinit(FormPtr frmP)
{
#pragma unused(frmP)
}


/*
 * TableDrawCell:
 *   Custom table cell drawing function.
 *   The function's prototype is defined by PalmOS
 *   (TableDrawItemFuncType). 
 */
TableDrawItemFuncType TableDrawCell;


/**********************************************************************/
/**
 * \brief	File table cell drawing function
 *
 * \note	The function's prototype is defined by PalmOS
 *   		(TableDrawItemFuncType).
 *
 * \note	On PalmOS 5.2.1 and 3.5.0 (and maybe others) the selected
 *		row has to be highlighted manually.
 *
 ***********************************************************************/
static void TableDrawCell(void *tblP, Int16 row, Int16 column,
                   RectangleType *bounds)
{
  Char		*c;		/* filename */
  Char		**filelp;	/* pointer to filelist entry */
  Int16 	sel_row;	/* selected table row */
  Int16		sel_col;	/* selected table column */

  filelp = (Char **)TblGetRowData(tblP, row);
  
  
  if (column == 0) {
    c = *filelp + SORTPREFIX;
    WinEraseRectangle(bounds, 0);
    
    TblGetSelection(tblP, &sel_row, &sel_col);
    if (row == sel_row) {
      /* Highlight selected row */
      if (gFtrSet35) {
        WinPushDrawState();
	WinSetBackColor(UIColorGetTableEntryIndex(UIObjectSelectedFill));
	WinSetTextColor(UIColorGetTableEntryIndex(UIObjectSelectedForeground));
	WinSetForeColor(UIColorGetTableEntryIndex(UIObjectSelectedForeground));
	WinFillRectangle(bounds, 0);
	WinPaintChars(c, strlen(c), bounds->topLeft.x + 1, bounds->topLeft.y);
      } else {
        WinDrawChars(c, strlen(c),
	  bounds->topLeft.x + 1, bounds->topLeft.y);
      }
    } else {
      WinDrawChars(c, strlen(c), bounds->topLeft.x + 1, bounds->topLeft.y);
    }
    
    /* draw card icon if file is located on card */
    if (vfs_flag[row]) {
      WinDrawChar(chrCardIcon, bounds->extent.x + 1 -
        FntCharWidth(chrCardIcon), bounds->topLeft.y);
    }
  }
  
  if (gFtrSet35)
    WinPopDrawState();
  
}


/**********************************************************************/
/**
 * \brief	File table drawing function
 *
 * \param	files		number of available files
 * \param	filelp		array of pointers to filelist entries
 * \param	filecard	file stream cards/VFS volumes per file
 * \param	top		top table row
 * \param	sel		selected table row
 *
 ***********************************************************************/
static void DrawTable(UInt16 files, char **filelp, char *filecard, 
                      Int32 top, UInt16 *sel)
{
  TablePtr 	tblP;			/* table ptr */
  FormPtr 	frmP;			/* form ptr */
  ScrollBarPtr	sclP;			/* scrollbar ptr */
  UInt16	row;			/* table row counter */
  UInt16	numRows;		/* number of rows */
  UInt32	card;			/* card / volume no. and FS/VFS flag */  
  
  frmP = FrmGetActiveForm();
  tblP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, FileSelTable));
  numRows = TblGetNumberOfRows(tblP);

  /* set column usable */
  TblSetColumnUsable(tblP, 0, true);
  
  for (row = 0; row < numRows; row++) {
    
    /* set row data to filelp entries */
    if (row + top < files) {
      TblSetRowData(tblP, row, (UInt32)(filelp + top + row));

      /* set all rows selectable */
      TblSetRowSelectable(tblP, row, true);
      
      /* set column to custom style, set custom cell drawing function */
      TblSetItemStyle(tblP, row, 0, customTableItem);  
      TblSetCustomDrawProcedure(tblP, 0, TableDrawCell);
      
      /* set vfs_flag-array for TableDrawCell() */
      memcpy(&card, &filecard[(top+row)*4], 4);
      vfs_flag[row] = (card & VFS_FLAG) ? true : false;
      
      /* mark all rows as invalid to trigger table update by FrmDrawForm() */
      TblMarkRowInvalid(tblP, row);
    } else {
      /* rows empty, mark as not usable */
      TblSetRowUsable(tblP, row, false);
    }

  }

  sclP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, FileSelScrl));

  /* set scrollbar: value, min, max, pagesize */
  if (files <= numRows) {
    SclSetScrollBar(sclP, 0, 0, 0, numRows);
  } else {
    SclSetScrollBar(sclP, top, 0, files - numRows, numRows);
  }

  
  /*
   * Selected item scrolled below visible table rows,
   *   selection is switched to last visible entry.
   */
  if ( *sel > (top + numRows - 1) ) {
    *sel = top + numRows - 1;
  }

  /*
   * Selected item scrolled above visible table rows,
   *   selection is switched to first visible entry.
   */  
  if (*sel < top) {
    *sel = top;
  }
  
  for (row = 0; row < numRows; row++, top++) {
    if ((top < files) && (top == *sel)) {
      TblSelectItem(tblP, row, 0);
      break;
    }
  }

  FrmDrawForm(frmP);
}


/**********************************************************************/
/**
 * \brief	Get file creation date (File Stream or VFS File)
 *
 * \param	name		pointer to filename
 * \param	card		File Stream card or VFS volume,
 *				MSB: VFS-Flag
 *
 * \return	File creation date (seconds)
 ***********************************************************************/
static UInt32 GetFileCreationDate(const char *name, UInt32 card)
{
  Err		err = 0;	/* error code */
  MemPtr	pathNameP;	/* ptr to path name */
  FileRef	fileRef;	/* file reference */
  LocalID	lid;		/* local database ID */
  UInt32	date;		/* date */

  /*
  if (gPrefs.mapprefs.scale == 8500) {
    StrCopy(FILEDIR, FILEDIRpre);
    StrCat(FILEDIR, "/8500");
  } else if (gPrefs.mapprefs.scale == 18000) {
    StrCopy(FILEDIR, FILEDIRpre);
    StrCat(FILEDIR, "/18k");
  } else {
    StrCopy(FILEDIR, FILEDIRpre);
  }
  */
  //  StrCopy(FILEDIR, FILEDIRpre);

  if (card & VFS_FLAG) {
    card &= ~VFS_FLAG;
    //    pathNameP = MemPtrNew(sizeof(FILEDIR) + StrLen(name) + 2);
    pathNameP = MemPtrNew(StrLen(FILEDIR) + StrLen(name) + 2);
    StrCopy(pathNameP, FILEDIR);
    StrCat(pathNameP, "/");
    StrCat(pathNameP, name);
    
    err = VFSFileOpen(card, pathNameP, vfsModeRead, &fileRef);
    MemPtrFree(pathNameP);
    
    if (!err) {
      err = VFSFileGetDate(fileRef, vfsFileDateCreated, &date);
      VFSFileClose(fileRef);
    }
  
  } else {
    lid = DmFindDatabase(card, name);
    
    if (lid) {
      err = DmDatabaseInfo(card,
                           lid,
			   NULL /*  *nameP */,
			   NULL /* *attributesP */,
			   NULL /* UInt16 *versionP */,
			   &date /* *crDateP */,
			   NULL /* UInt32 *modDateP */,
			   NULL /* *bckUpDateP */,
			   NULL /* UInt32 *modNumP */,
			   NULL /* LocalID *appInfoIDP */,
			   NULL /* LocalID *sortInfoIDP */,
			   NULL /* UInt32 *typeP */,
			   NULL /* UInt32 *creatorP*/
	    );
    }  
  }
  
  if (err) {
    date = 0;
  }
  return date;
}


/**********************************************************************/
/**
 * \brief	Delete Map File (File Stream or VFS File)
 *
 * \param	name		pointer to filename
 * \param	card		File Stream card or VFS volume,
 *				MSB: VFS-Flag
 *
 * \return	File creation date (seconds)
 ***********************************************************************/
static Err DeleteMap(const char *name, UInt32 card)
{
  Err		err = 0;	/* error code */
  MemPtr	pathNameP;	/* ptr to path name */

  if (card & VFS_FLAG) {
    card &= ~VFS_FLAG;
    //    pathNameP = MemPtrNew(sizeof(FILEDIR) + StrLen(name) + 2);
    pathNameP = MemPtrNew(StrLen(FILEDIR) + StrLen(name) + 2);
    StrCopy(pathNameP, FILEDIR);
    StrCat(pathNameP, "/");
    StrCat(pathNameP, name);
    
    err = VFSFileDelete(card, pathNameP);
  
  } else {
    err = FileDelete(card, name);  
  
  }
  
  return err;
}


/**********************************************************************/
/**
 * \brief	File Selection Form Event Handler.
 *
 * \param	eventP		pointer to event structure 
 *
 *
 * \return	Status flag: event handled
 ***********************************************************************/
Boolean FileSelFormHandleEvent(EventPtr eventP)
{
  EventType		event;			/* event to queue */
  Err			err;			/* error code */
  Boolean		handled = false;	/* event handled flag */
  FormPtr		frmP;			/* form ptr */
  TablePtr		tblP;			/* table ptr */
  static DmOpenRef	scratch;		/* scratchpad DB reference */
  static char		*filelist;		/* filelist DB record */
  static char		*filecard;		/* filecard DB record */
  static char		**filelp;		/* filelp DB record */
  static UInt16		files;			/* number of files found */
  UInt32		card;			/* card/volume */
  char			*filename;		/* ptr to filename */
  UInt32		seconds;		/* seconds (date) */
  DateTimeType		dt;			/* date/time structure */
  UInt16		i;			/* loop index */
  Char			**c;			/* pointer to file list entry */
  Char			tmp[dateStringLength +
                            timeStringLength +
			    2];			/* string buffer */
  UInt32		size;			/* file size */
  static Int32		file_top = 0;		/* file table top */
  static UInt16		file_sel = 0;		/* file table selected row */
  FileDescrType		filedescr;		/* file descriptor */
  UInt16		width;			/* image width */
  UInt16		height;			/* image height */
  UInt8			type;			/* image type */
  
  frmP = FrmGetActiveForm();
  tblP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, FileSelTable));
  
  switch (eventP->eType) {
  case frmOpenEvent:
    FileSelFormInit(frmP);
    FrmDrawForm(frmP);
    
    CreateDatabase(&scratch, &filelist, &filecard, &filelp);
    
    files = getdir(filelist, filecard, filelp);
    
    if (gVfsManagerFtr) {
      StrCopy(FILEDIR, FILEDIRpre);
      files = getdir_vfs(filelist, filecard, filelp, files);
    }
       
    if (files == 0) {
      /* No Image File found */
      if (gVfsManagerFtr) {
        FrmAlert(DirEmptyVfsAlert);
      } else {
        FrmAlert(DirEmptyAlert);
      }
    }
    
    /*
     * Compare current filename with filelist,
     * if matching entry is found, mark filename
     * in table as selected and store position in gFileSel.
     * The top row is adjusted to make sure the current
     * file is visible.
     */ 
    c = filelp;
    
    file_top = 0;
    file_sel = 0;
    
    /* find index of current file */
    for (i = 0; i<files; i++, c++) {
       memcpy(&card, &filecard[i*4], 4);
       
       if ( (StrCompare(SEL_FILENAME, *c+SORTPREFIX) == 0 ) &&
            (card == SEL_FILECARD) ) {
         
	 /* file found, set table top row and selection */
	 file_sel = i;
         file_top = (file_sel < TblGetNumberOfRows(tblP)) ? 0 :
	   (file_sel - TblGetNumberOfRows(tblP) + 1);     
         
	 /* show file size */
	 memcpy(&size, filelp[file_sel], 4);
	 StrIToA(tmp, size);
	 SetFieldText(FileSelSizeField, tmp, false, true);
    
	 /* show file creation date */
	 seconds = GetFileCreationDate(filelp[file_sel] + SORTPREFIX, card);

	 if (seconds > 0) {
	   TimSecondsToDateTime(seconds, &dt);

	   DateToAscii(dt.month, dt.day, dt.year, dfYMDWithDashes, tmp);
	   StrCat(tmp, " ");
	   TimeToAscii(dt.hour, dt.minute, tfColon24h, &tmp[StrLen(tmp)]);
	   SetFieldText(FileSelDateField, tmp, false, true);

	 } else {
	   SetFieldText(FileSelDateField, NULL, false, true);
	 }
    
	 /* show dimensions of selected image */
	 filedescr = fopen(card, filelp[file_sel] + SORTPREFIX, 
	   FILEDIR, &err);

	 if (!err) {
	   if (!imginfo(filedescr, &type, &width, &height)) {
	     StrPrintF(tmp, "%d x %d", width, height);
	     SetFieldText(FileSelInfo1Field, tmp, false, true);
	   } else {
             SetFieldText(FileSelInfo1Field, "", false, true);
	   }
	   fclose(filedescr);
	 } else {
	   SetFieldText(FileSelInfo1Field, "", false, true);
	 }         
         break;
         
       }
    }
    
    /* draw the table */
    DrawTable(files, filelp, filecard, file_top, &file_sel);

    handled = true;
    break;
  
  case tblSelectEvent:
    /* show size and creation date of selected file */
    file_sel = eventP->data.tblSelect.row + file_top;
    memcpy(&size, filelp[file_sel], 4);
    memcpy(&card, &filecard[file_sel*4], 4);
    
    StrIToA(tmp, size);
    SetFieldText(FileSelSizeField, tmp, false, true);
    
    seconds = GetFileCreationDate(filelp[file_sel] + SORTPREFIX, card);
    
    if (seconds > 0) {
      TimSecondsToDateTime(seconds, &dt);

      DateToAscii(dt.month, dt.day, dt.year, dfYMDWithDashes, tmp);
      StrCat(tmp, " ");
      TimeToAscii(dt.hour, dt.minute, tfColon24h, &tmp[StrLen(tmp)]);
      SetFieldText(FileSelDateField, tmp, false, true);
    
    } else {
      SetFieldText(FileSelDateField, NULL, false, true);
    }
    
    /* show dimensions of selected image */
    filedescr = fopen(card, filelp[file_sel] + SORTPREFIX, 
      FILEDIR, &err);
    
    if (!err) {
      if (!imginfo(filedescr, &type, &width, &height)) {
	StrPrintF(tmp, "%d x %d", width, height);
	SetFieldText(FileSelInfo1Field, tmp, false, true);
      } else {
        SetFieldText(FileSelInfo1Field, "", false, true);
      }
      fclose(filedescr);
    } else {
      SetFieldText(FileSelInfo1Field, "", false, true);
    }
    
    DrawTable(files, filelp, filecard, file_top, &file_sel);
    handled = true;
    break;
  
  case sclRepeatEvent:
    /* handle scrollbar */
    Doscroll(FileSelTable, FileSelScrl, files, &file_top);
    DrawTable(files, filelp, filecard, file_top, &file_sel);
    /* Note: scrollbar needs to handle the event, too */
    handled = false;
    break;
  
  case keyDownEvent:
    switch (eventP->data.keyDown.chr) {
    
    case pageUpChr:
    case pageDownChr:
      Dopage(FileSelTable, (eventP->data.keyDown.chr == pageDownChr),
        files, &file_top);
      DrawTable(files, filelp, filecard, file_top, &file_sel);
      /* Note: scrollbar needs to handle the event, too */
      handled = false;
      break;
    }
  
  case ctlSelectEvent:
    switch (eventP->data.ctlSelect.controlID) {

    case FileSelOkButton:
      if (files == 0) {
        /* no files, quit */
	event.eType = appStopEvent;
        EvtAddEventToQueue(&event);
        break;
      }
      if (file_sel < files) {
        /* copy selected name and card no. of selected file to preferences */
	StrCopy(SEL_FILENAME, filelp[file_sel] + SORTPREFIX);
	memcpy(&SEL_FILECARD, &filecard[file_sel*4], 4);
      }
      /* user selected map, keep this map even if not showing current position */
      gPrefs.mapprefs.select = SEL_USER;
      FrmGotoForm(MapForm);
      handled = true;
      break;
    
    case FileSelAutoButton:
      /* automatically select map */
      gPrefs.mapprefs.select = SEL_AUTO;
      /* clear filename */
      StrCopy(SEL_FILENAME, "");
      FrmGotoForm(MapForm);
      handled = true;
      break;
    
    case FileSelCancelButton:
      if (files == 0) {
        /* no files, quit */
	event.eType = appStopEvent;
        EvtAddEventToQueue(&event);
        break;
      }
      FrmGotoForm(MapForm);
      handled = true;
      break;

    case FileSelDeleteButton:
      if (file_sel < files) {
	MemHandle		strH = NULL;		/* string handle */
	MemPtr			strP = NULL;		/* string ptr */
        Boolean			confirm = false;	/* delete confirmed */
	
	filename = filelp[file_sel] + SORTPREFIX;
	memcpy(&card, &filecard[file_sel*4], 4);

	/* get string from resource */
	strH = DmGetResource(strRsc, (card & VFS_FLAG) ? MapExtStr : MapIntStr);

	if (strH != NULL) {
	  /* string handle valid, lock it */
	  strP = MemHandleLock(strH);
	}

        confirm = (FrmCustomAlert(MapDelConfirmAlert, filename, 
	  strP, "") == 0);

	if (strH != NULL) {
	  /* unlock string handle */
	  MemHandleUnlock(strH);

	  /* release string resource */
	  DmReleaseResource(strH);
	}
	if (confirm) {
	  DeleteMap(filename, card);
	  if (file_sel > 0) {
	    file_sel--;
	  }
	  /* clear filename */
	  StrCopy(SEL_FILENAME, "");

          FrmGotoForm(FileSelForm);
        }
      }
      handled = true;
      break;
      
    }
    break;
    
  case frmCloseEvent:
    SetFieldText(FileSelSizeField, NULL, false, true);
    SetFieldText(FileSelDateField, NULL, false, true);
    FileSelFormDeinit(FrmGetActiveForm());
    DestroyDatabase(scratch, filelist, filecard, filelp);
    handled = false;
    break;

  default:
    break;
  }
  
  return handled;
}
