/*****************************************************************************
 *
 * $RCSfile: Garmin.h,v $
 *
 * GPS4Palm
 * Garmin Data Type Definitions
 * According to GARMIN GPS Interface Specification Rev. 3 1999-12-06
 *
 * Imported from gpslib-1.31 by Christian Claveleira
 * <Christian.Claveleira@fnac.net>
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2004-02-08 18:07:43 $
 *
 * $Revision: 1.2 $
 *
 * $Log: Garmin.h,v $
 * Revision 1.2  2004-02-08 18:07:43  mp
 * minor editorial changes
 *
 *
 ****************************************************************************/

/*  Garmin types :

G_boolean  : 8 bit int
G_byte     : unsigned 8 bits int or char
G_char     : signed   8 bits int or char
G_word     : little-endian unsigned 16 bits int
G_int      : little-endian signed 16 bits int
G_longword : little-endian unsigned 32 bits int
G_long     : little-endian signed 32 bits int
G_float    : little-endian 32 bits IEEE-format floating point data
G_double   : little-endian 64 bits IEEE-format floating point data

*/
#ifndef __GARMIN_H__
#define __GARMIN_H__

#if defined __palmos__  // type castings for PalmOS
typedef unsigned char G_boolean;
typedef unsigned char G_byte;
typedef char G_char;
typedef UInt16 G_word;
typedef Int16 G_int;
typedef UInt32 G_longword;
typedef Int32 G_long;
typedef float G_float;
typedef double G_double;
typedef G_int Symbol_Type;

#elif defined GLIB_MAJOR_VERSION  // type castings for glib

typedef guchar G_boolean;
typedef guchar G_byte;
typedef char G_char;
typedef guint16 G_word;
typedef gint16 G_int;
typedef guint32 G_longword;
typedef gint32 G_long;
typedef float G_float;
typedef double G_double;
typedef G_int Symbol_Type;

#else    // default
#endif

typedef unsigned char byte;
typedef unsigned char boolean;

typedef struct
{				       /* packet structure */
    byte dle;			       /* start delimiter */
    byte type;			       /* type */
    byte length;		       /* size of packet data */
    byte datas[514];		       /* packet data (with large 
                                          additional space in case
					  of many DLEs */
}
Garmin_Packet;


enum				       /* Basic Packet IDs */
{
    Pid_Ack_Byte	= 6,
    Pid_Nack_Byte	= 21,
    Pid_Protocol_Array	= 253,
    Pid_Product_Rqst	= 254,
    Pid_Product_Data	= 255
};

enum				       /* L001 - Link Protocol 1 */
{
    Pid_Command_Data	= 10,
    Pid_Xfer_Cmplt	= 12,
    Pid_Date_Time_Data	= 14,
    Pid_Position_Data	= 17,
    Pid_Prx_Wpt_Data	= 19,
    Pid_Records		= 27,
    Pid_Rte_Hdr		= 29,
    Pid_Rte_Wpt_Data	= 30,
    Pid_Almanac_Data	= 31,
    Pid_Trk_Data	= 34,
    Pid_Wpt_Data	= 35,
    Pid_Speed_Req	= 48,  /* speed proposal ? */
    Pid_Speed_Ack	= 49,  /* speed ack ? */
    Pid_Pvt_Data	= 51,
    Pid_Rte_Link_Data	= 98,
    Pid_Trk_Hdr		= 99
};

enum				       /* A010 - Device Command Protocol 1 */
{
    Cmnd_Abort_Transfer	= 0,	       /* abort current transfer */
    Cmnd_Transfer_Alm	= 1,	       /* transfer almanac */
    Cmnd_Transfer_Posn	= 2,	       /* transfer position */
    Cmnd_Transfer_Prx	= 3,	       /* transfer proximity waypoints */
    Cmnd_Transfer_Rte	= 4,	       /* transfer routes */
    Cmnd_Transfer_Time	= 5,	       /* transfer time */
    Cmnd_Transfer_Trk	= 6,	       /* transfer track log */
    Cmnd_Transfer_Wpt	= 7,	       /* transfer waypoints */
    Cmnd_Turn_Off_Pwr	= 8,	       /* turn off power */
    Cmnd_Start_Pvt_Data = 49,	       /* start transmitting PVT data */
    Cmnd_Stop_Pvt_Data	= 50	       /* stop transmitting PVT data */
};


#ifdef DEVICE_COMMAND_PROTOCOL_2
enum				/* A011 - Device Command Protocol 2 */
{
    Cmnd_Abort_Transfer	= 0,	/* abort current transfer */
    Cmnd_Transfer_Alm	= 4,	/* transfer almanac */
    Cmnd_Transfer_Rte	= 8,	/* transfer routes */
    Cmnd_Transfer_Time	= 20,	/* transfer time */
    Cmnd_Transfer_Wpt	= 21,	/* transfer waypoints */
    Cmnd_Turn_Off_Pwr	= 26	/* turn off power */
};
#endif

typedef struct
{
    G_long lat;			       /* latitude in semicircles */
    G_long lon;			       /* longitude in semicircles */
}
Semicircle_Type;

typedef struct
{
    G_double lat;			       /* latitude in radians */
    G_double lon;			       /* longitude in radians */
}
Radian_Type;

typedef struct 
{ 
  G_int product_ID; 
  G_int software_version; 
  /* char product_description[]; null-terminated string */ 
  /* ... zero or more additional null-terminated strings */ 
} Product_Data_Type;

typedef struct 
{ 
  G_byte tag; 
  G_word data; 
} Protocol_Data_Type;

typedef Protocol_Data_Type Protocol_Array_Type[];

// id of Garmin data formats (sent in Protocol_Data_Type) :
enum GarminDataFormatID
{
    Garmin100Type = 100,
    Garmin101Type,
    Garmin102Type,
    Garmin103Type,
    Garmin104Type,
    Garmin105Type,
    Garmin106Type,
    Garmin107Type,
    Garmin108Type,
    Garmin109Type,
    Garmin200Type = 200, Garmin201Type, Garmin202Type, Garmin210Type = 210,
    Garmin300Type = 300, Garmin301Type, Garmin310Type = 310,
    Garmin600Type = 600,
    Garmin700Type = 700,
    Garmin800Type = 800
};
typedef enum GarminDataFormatID GarminDataFormatID;

// id of Garmin protocols (sent in Protocol_Data_Type) :
enum GarminProtoID
{
    A000 = 0, A001,
    A010 = 10, A011,
    A100 = 100,
    A200 = 200, A201,
    A300 = 300, A301,
    A400 = 400,
    A500 = 500,
    A600 = 600,
    A700 = 700,
    A800 = 800
};
typedef enum GarminProtoID GarminProtoID;

enum
{
/*--------------------------------------------------------------- 
Symbols for marine (group 0...0-8191...bits 15-13=000). 
---------------------------------------------------------------*/
    sym_anchor = 0,		       /* white anchor symbol */
    sym_bell = 1,		       /* white bell symbol */
    sym_diamond_grn = 2,	       /* green diamond symbol */
    sym_diamond_red = 3,	       /* red diamond symbol */
    sym_dive1 = 4,		       /* diver down flag 1 */
    sym_dive2 = 5,		       /* diver down flag 2 */
    sym_dollar = 6,		       /* white dollar symbol */
    sym_fish = 7,		       /* white fish symbol */
    sym_fuel = 8,		       /* white fuel symbol */
    sym_horn = 9,		       /* white horn symbol */
    sym_house = 10,		       /* white house symbol */
    sym_knife = 11,		       /* white knife & fork symbol */
    sym_light = 12,		       /* white light symbol */
    sym_mug = 13,		       /* white mug symbol */
    sym_skull = 14,		       /* white skull and crossbones symbol */
    sym_square_grn = 15,	       /* green square symbol */
    sym_square_red = 16,	       /* red square symbol */
    sym_wbuoy = 17,		       /* white buoy waypoint symbol */
    sym_wpt_dot = 18,		       /* waypoint dot */
    sym_wreck = 19,		       /* white wreck symbol */
    sym_null = 20,		       /* null symbol (transparent) */
    sym_mob = 21,		       /* man overboard symbol */
/*------------------------------------------------------ 
marine navaid symbols 
------------------------------------------------------*/
    sym_buoy_ambr = 22,		       /* amber map buoy symbol */
    sym_buoy_blck = 23,		       /* black map buoy symbol */
    sym_buoy_blue = 24,		       /* blue map buoy symbol */
    sym_buoy_grn = 25,		       /* green map buoy symbol */
    sym_buoy_grn_red = 26,	       /* green/red map buoy symbol */
    sym_buoy_grn_wht = 27,	       /* green/white map buoy symbol */
    sym_buoy_orng = 28,		       /* orange map buoy symbol */
    sym_buoy_red = 29,		       /* red map buoy symbol */
    sym_buoy_red_grn = 30,	       /* red/green map buoy symbol */
    sym_buoy_red_wht = 31,	       /* red/white map buoy symbol */
    sym_buoy_violet = 32,	       /* violet map buoy symbol */
    sym_buoy_wht = 33,		       /* white map buoy symbol */
    sym_buoy_wht_grn = 34,	       /* white/green map buoy symbol */
    sym_buoy_wht_red = 35,	       /* white/red map buoy symbol */
    sym_dot = 36,		       /* white dot symbol */
    sym_rbcn = 37,		       /* radio beacon symbol */
/*------------------------------------------------------ 
leave space for more navaids (up to 128 total) 
------------------------------------------------------*/
    sym_boat_ramp = 150,	       /* boat ramp symbol */
    sym_camp = 151,		       /* campground symbol */
    sym_restrooms = 152,	       /* restrooms symbol */
    sym_showers = 153,		       /* shower symbol */
    sym_drinking_wtr = 154,	       /* drinking water symbol */
    sym_phone = 155,		       /* telephone symbol */
    sym_1st_aid = 156,		       /* first aid symbol */
    sym_info = 157,		       /* information symbol */
    sym_parking = 158,		       /* parking symbol */
    sym_park = 159,		       /* park symbol */
    sym_picnic = 160,		       /* picnic symbol */
    sym_scenic = 161,		       /* scenic area symbol */
    sym_skiing = 162,		       /* skiing symbol */
    sym_swimming = 163,		       /* swimming symbol */
    sym_dam = 164,		       /* dam symbol */
    sym_controlled = 165,	       /* controlled area symbol */
    sym_danger = 166,		       /* danger symbol */
    sym_restricted = 167,	       /* restricted area symbol */
    sym_null_2 = 168,		       /* null symbol */
    sym_ball = 169,		       /* ball symbol */
    sym_car = 170,		       /* car symbol */
    sym_deer = 171,		       /* deer symbol */
    sym_shpng_cart = 172,	       /* shopping cart symbol */
    sym_lodging = 173,		       /* lodging symbol */
    sym_mine = 174,                    /* mine symbol */ 
    sym_trail_head = 175,              /* trail head symbol */ 
    sym_truck_stop = 176,              /* truck stop symbol */ 
    sym_user_exit = 177,               /* user exit symbol */ 
    sym_flag = 178,                    /* flag symbol */ 
    sym_circle_x = 179,                /* circle with x in the center */

/*--------------------------------------------------------------- 
 Symbols for land (group 1...8192-16383...bits 15-13=001). 
 ---------------------------------------------------------------*/
    sym_is_hwy = 8192,		       /* interstate hwy symbol */
    sym_us_hwy = 8193,		       /* us hwy symbol */
    sym_st_hwy = 8194,		       /* state hwy symbol */
    sym_mi_mrkr = 8195,		       /* mile marker symbol */
    sym_trcbck = 8196,		       /* TracBack (feet) symbol */
    sym_golf = 8197,		       /* golf symbol */
    sym_sml_cty = 8198,		       /* small city symbol */
    sym_med_cty = 8199,		       /* medium city symbol */
    sym_lrg_cty = 8200,		       /* large city symbol */
    sym_freeway = 8201,		       /* intl freeway hwy symbol */
    sym_ntl_hwy = 8202,		       /* intl national hwy symbol */
    sym_cap_cty = 8203,		       /* capitol city symbol (star) */
    sym_amuse_pk = 8204,	       /* amusement park symbol */
    sym_bowling = 8205,		       /* bowling symbol */
    sym_car_rental = 8206,	       /* car rental symbol */
    sym_car_repair = 8207,	       /* car repair symbol */
    sym_fastfood = 8208,	       /* fast food symbol */
    sym_fitness = 8209,		       /* fitness symbol */
    sym_movie = 8210,		       /* movie symbol */
    sym_museum = 8211,		       /* museum symbol */
    sym_pharmacy = 8212,	       /* pharmacy symbol */
    sym_pizza = 8213,		       /* pizza symbol */
    sym_post_ofc = 8214,	       /* post office symbol */
    sym_rv_park = 8215,		       /* RV park symbol */
    sym_school = 8216,		       /* school symbol */
    sym_stadium = 8217,		       /* stadium symbol */
    sym_store = 8218,		       /* dept. store symbol */
    sym_zoo = 8219,		       /* zoo symbol */
    sym_gas_plus = 8220,	       /* convenience store symbol */ 
    sym_faces = 8221,	               /* live theater symbol */ 
    sym_ramp_int = 8222,	       /* ramp intersection symbol */ 
    sym_st_int = 8223,	               /* street intersection symbol */ 
    sym_weigh_sttn = 8226,             /* inspection/weigh station symbol */ 
    sym_toll_booth = 8227,             /* toll booth symbol */ 
    sym_elev_pt = 8228, 	       /* elevation point symbol */ 
    sym_ex_no_srvc = 8229,             /* exit without services symbol */ 
    sym_geo_place_mm = 8230, 	       /* Geographic place name, man-made */ 
    sym_geo_place_wtr = 8231,          /* Geographic place name, water */ 
    sym_geo_place_lnd = 8232, 	       /* Geographic place name, land */ 
    sym_bridge = 8233,	               /* bridge symbol */ 
    sym_building = 8234,	       /* building symbol */ 
    sym_cemetery = 8235,	       /* cemetery symbol */ 
    sym_church = 8236,	               /* church symbol */ 
    sym_civil = 8237,	               /* civil location symbol */ 
    sym_crossing = 8238,               /* crossing symbol */ 
    sym_hist_town = 8239,              /* historical town symbol */ 
    sym_levee = 8240,	               /* levee symbol */ 
    sym_military = 8241,	       /* military location symbol */ 
    sym_oil_field = 8242,	       /* oil field symbol */ 
    sym_tunnel = 8243,	               /* tunnel symbol */ 
    sym_beach = 8244,	               /* beach symbol */ 
    sym_forest = 8245,	               /* forest symbol */ 
    sym_summit = 8246,	               /* summit symbol */ 
    sym_lrg_ramp_int = 8247,	       /* large ramp intersection symbol */ 
    sym_lrg_ex_no_srvc = 8248,	       /* large exit without services smbl */ 
    sym_badge = 8249,    	       /* police/official badge symbol */ 
    sym_cards = 8250,	               /* gambling/casino symbol */
    sym_snowski = 8251,	               /* snow skiing symbol */ 
    sym_iceskate = 8252,	       /* ice skating symbol */ 
    sym_wrecker = 8253,	               /* tow truck (wrecker) symbol */ 
    sym_border = 8254,	               /* border crossing (port of entry) */

/* --------------------------------------------- 
   Symbols for aviation (group 2...16383-24575...bits 15-13=010). 
   --------------------------------------------------------------- */
    sym_airport = 16384,	       /* airport symbol */
    sym_int = 16385,		       /* intersection symbol */
    sym_ndb = 16386,		       /* non-directional beacon symbol */
    sym_vor = 16387,		       /* VHF omni-range symbol */
    sym_heliport = 16388,	       /* heliport symbol */
    sym_private = 16389,	       /* private field symbol */
    sym_soft_fld = 16390,	       /* soft field symbol */
    sym_tall_tower = 16391,	       /* tall tower symbol */
    sym_short_tower = 16392,	       /* short tower symbol */
    sym_glider = 16393,		       /* glider symbol */
    sym_ultralight = 16394,	       /* ultralight symbol */
    sym_parachute = 16395,	       /* parachute symbol */
    sym_vortac = 16396,		       /* VOR/TACAN symbol */
    sym_vordme = 16397,		       /* VOR-DME symbol */
    sym_faf = 16398,		       /* first approach fix */
    sym_lom = 16399,		       /* localizer outer marker */
    sym_map = 16400,		       /* missed approach point */
    sym_tacan = 16401,		       /* TACAN symbol */
    sym_seaplane = 16402,	       /* Seaplane Base */
};

typedef struct
{
  G_char ident[6];		/* identifier */
  Semicircle_Type posn;		/* position */
  G_longword unused;		/* should be set to zero */
  G_char cmnt[40];		/* comment */
}
D100_Wpt_Type;
 
typedef struct
{
  G_char ident[6];		/* identifier */
  Semicircle_Type posn;		/* position */
  G_longword unused;		/* should be set to zero */
  G_char cmnt[40];		/* comment */
  G_float dst;			/* proximity distance (meters) */
  G_byte smbl;			/* symbol id */
}
D101_Wpt_Type;

typedef struct
{
  G_char ident[6];		/* identifier */
  Semicircle_Type posn;		/* position */
  G_longword unused;		/* should be set to zero */
  G_char cmnt[40];		/* comment */
  G_float dst;			/* proximity distance (meters) */
  Symbol_Type smbl;		/* symbol id */
}
D102_Wpt_Type;


typedef struct
{
    G_char ident[6];		/* identifier */
    Semicircle_Type posn;	/* position */
    G_longword unused;		/* should be set to zero */
    G_char cmnt[40];		/* comment */
    G_byte smbl;		/* symbol id */
    G_byte dspl;		/* display option */
}
D103_Wpt_Type;

enum
{
    smbl_dot = 0,		       /* dot symbol */
    smbl_house = 1,		       /* house symbol */
    smbl_gas = 2,		       /* gas symbol */
    smbl_car = 3,		       /* car symbol */
    smbl_fish = 4,		       /* fish symbol */
    smbl_boat = 5,		       /* boat symbol */
    smbl_anchor = 6,		       /* anchor symbol */
    smbl_wreck = 7,		       /* wreck symbol */
    smbl_exit = 8,		       /* exit symbol */
    smbl_skull = 9,		       /* skull symbol */
    smbl_flag = 10,		       /* flag symbol */
    smbl_camp = 11,		       /* camp symbol */
    smbl_duck = 12,		       /* duck symbol */
    smbl_deer = 13,		       /* deer symbol */
    smbl_buoy = 14,		       /* buoy symbol */
    smbl_back_track = 15	       /* back track symbol */
};

enum
{
    dspl_name = 0,		       /* Display symbol with waypoint name */
    dspl_none = 1,		       /* Display symbol by itself */
    dspl_cmnt = 2		       /* Display symbol with comment */
};

typedef struct
{
  G_char ident[6];		/* identifier */
  Semicircle_Type posn;		/* position */
  G_longword unused;		/* should be set to zero */
  G_char cmnt[40];		/* comment */
  G_float dst;			/* proximity distance (meters) */
  Symbol_Type smbl;		/* symbol id */
  G_byte dspl;			/* display option */
}
D104_Wpt_Type;

typedef struct { 
  Semicircle_Type posn;		/* position */ 
  Symbol_Type smbl;		/* symbol id */ 
  /* G_char wpt_ident[]; null-terminated string */ 
} D105_Wpt_Type;

enum
{
    dspl_smbl_none = 0,			/* Display symbol by itself */
    dspl_smbl_only = 1,			/* Display symbol by itself */
    dspl_smbl_name = 3,			/* Display symbol with waypoint name */
    dspl_smbl_cmnt = 5,			/* Display symbol with comment */
};

typedef struct				/*                             size */
{
    G_byte wpt_class;			/* class (see below)            1 */
    G_byte color;			/* color (see below)            1 */
    G_byte dspl;			/* display options (see below)  1 */
    G_byte attr;			/* attributes (see below)       1 */
    Symbol_Type smbl;			/* waypoint symbol              2 */
    G_byte subclass[18];		/* subclass                    18 */
    Semicircle_Type posn;		/* 32 bit semicircle            8 */
    G_float alt;			/* altitude in meters           4 */
    G_float dpth;			/* depth in meters              4 */
    G_float dist;			/* proximity distance in meters 4 */
    G_char state[2];			/* state                        2 */
    G_char cc[2];			/* country code                 2 */
    /* G_char ident[]; variable length string 1-51       */
    /* G_char comment[]; waypoint user comment 1-51      */
    /* G_char facility[]; facility name 1-31             */
    /* G_char city[]; city name 1-25                     */
    /* G_char addr[]; address number 1-51                */
    /* G_char cross_road[]; intersecting road label 1-51 */
}
D108_Wpt_Type;

/*  The enumerated values for the "wpt_class" member
    of the D108_Wpt_Type are defined as follows:  */
enum
{
    USER_WPT = 0x00,		       /* User waypoint */
    AVTN_APT_WPT = 0x40,	       /* Aviation Airport waypoint */
    AVTN_INT_WPT = 0x41,	       /* Aviation Intersection waypoint */
    AVTN_NDB_WPT = 0x42,	       /* Aviation NDB waypoint */
    AVTN_VOR_WPT = 0x43,	       /* Aviation VOR waypoint */
    AVTN_ARWY_WPT = 0x44,	       /* Aviation Airport Runway waypoint */
    AVTN_AINT_WPT = 0x45,	       /* Aviation Airport Intersection */
    AVTN_ANDB_WPT = 0x46,	       /* Aviation Airport NDB waypoint */
    MAP_PNT_WPT = 0x80,		       /* Map Point waypoint */
    MAP_AREA_WPT = 0x81,	       /* Map Area waypoint */
    MAP_INT_WPT = 0x82,		       /* Map Intersection waypoint */
    MAP_ADRS_WPT = 0x83,	       /* Map Address waypoint */
    MAP_LABEL_WPT = 0x84,	       /* Map Label Waypoint */
    MAP_LINE_WPT = 0x85,	       /* Map Line Waypoint */
};
/*  The "color" member can be one of the following values:  */
enum
{
    Black, Dark_Red, Dark_Green, Dark_Yellow,
    Dark_Blue, Dark_Magenta, Dark_Cyan, Light_Gray,
    Dark_Gray, Red, Green, Yellow,
    Blue, Magenta, Cyan, White,
    Default_Color = 0xFF
};

/*  The enumerated values for the "dspl" member of the D108_Wpt_Type are the same as the the "dspl" member of the D103_Wpt_Type.  */
/*  The "attr" member should be set to a value of 0x60.  */
/*  The "subclass" member of the D108_Wpt_Type is used for map waypoints only, and should be set to 0x0000 0x00000000 0xFFFFFFFF 0xFFFFFFFF 0xFFFFFFFF for other classes of waypoints.  */
/*  The "alt" and "dpth" members may or may not be supported on a given unit. A value of 1.0e25 in either of these fields indicates that this parameter is not supported or is unknown for this waypoint. */

typedef struct			/*                                 size */
{
    G_byte dtyp;		/* data packet type (0x01 for D109)1    */
    G_byte wpt_class;		/* class                           1    */
    G_byte dspl_color;		/* display & color (see below)     1    */
    G_byte attr;		/* attributes (0x70 for D109)      1    */
    Symbol_Type smbl;		/* waypoint symbol                 2    */
    G_byte subclass[18];	/* subclass                        18   */
    Semicircle_Type posn;	/* 32 bit semicircle               8    */
    G_float alt;		/* altitude in meters              4    */
    G_float dpth;		/* depth in meters                 4    */
    G_float dist;		/* proximity distance in meters    4    */
    G_char state[2];		/* state                           2    */
    G_char cc[2];		/* country code                    2    */
    G_longword ete;		/* outbound link ete in seconds    4    */
    /*  G_char	ident[];           variable length string          1-51 */
    /*  G_char	comment[];         waypoint user comment           1-51 */
    /*  G_char	facility[];        facility name                   1-31 */
    /*  G_char	city[];            city name                       1-25 */
    /*  G_char	addr[];            address number                  1-51 */
    /*  G_char	cross_road[];      intersecting road label         1-51 */
}
D109_Wpt_Type;

/*  All fields are defined the same as D108 except as noted below.                    */
/*  dtyp      - Data packet type, must be 0x01 for D109.                              */
/*  dsp_color - The 'dspl_color' member contains three fields; bits 0-4 specify       */
/*    the color, bits 5-6 specify the waypoint display attribute and bit 7 is unused  */
/*    and must be 0. Color values are as specified for D108 except that the default   */
/*    value is 0x1f. Display attribute values are as specified for D108.              */
/*  attr      - Attribute. Must be 0x70 for D109.                                     */
/*  ete       - Estimated time en route in seconds to next waypoint. Default value is */
/*  0xffffffff.                                                                       */


typedef G_byte D200_Rte_Hdr_Type;	       /* route number */

typedef struct
{
    G_byte nmbr;		/* route number */
    G_char cmnt[20];		/* comment */
}
D201_Rte_Hdr_Type;

typedef struct
{
    G_char rte_ident[254];	/* null-terminated string */
}
D202_Rte_Hdr_Type;

typedef struct
{
    G_word class;		/* link class; see below */
    G_byte subclass[18];	/* sublcass */
    G_char ident[50];		/* variable length string */
}
D210_Rte_Link_Type;

typedef struct
{
    Semicircle_Type posn;	/* position */
    G_longword time;		/* time */
    G_boolean new_trk;		/* new track segment? */
}
D300_Trk_Point_Type;

typedef struct
{
  Semicircle_Type posn;		/* position */
  G_longword time;		/* time */
  G_float alt;			/* altitude in meters */
  G_float dpth;			/* depth in meters */
  G_boolean new_trk;		/* new track segment? */
}
D301_Trk_Point_Type;

typedef struct
{
  G_boolean dspl;		/* display on the map? */
  G_byte color;			/* color (same as D108) */
  G_char trk_ident[50];		/* null-terminated string */
}
D310_Trk_Hdr_Type;

typedef struct
{
  G_int wn;			/* week number (weeks) */
  G_float toa;			/* almanac data reference time (s) */
  G_float af0;			/* clock correction coefficient (s) */
  G_float af1;			/* clock correction coefficient (s/s) */
  G_float e;			/* eccentricity (-) */
  G_float sqrta;		/* square root of semi-major axis (a) (m**1/2) */
  G_float m0;			/* mean anomaly at reference time (r) */
  G_float w;			/* argument of perigee (r) */
  G_float omg0;			/* right ascension (r) */
  G_float odot;			/* rate of right ascension (r/s) */
  G_float i;			/* inclination angle (r) */
  G_byte hlth;			/* almanac health */
}
D501_Almanac_Type;

typedef struct
{
    G_byte month;		/* month (1-12) */
    G_byte day;			/* day (1-31) */
    G_word year;		/* year (1990 means 1990) */
    G_int hour;			/* hour (0-23) */
    G_byte minute;		/* minute (0-59) */
    G_byte second;		/* second (0-59) */
}
D600_Date_Time_Type;

typedef struct
{
    G_float alt;		/* altitude above WGS 84 ellipsoid (meters) */
    G_float epe;		/* estimated position error, 2 sigma (meters) */
    G_float eph;		/* epe, but horizontal only (meters) */
    G_float epv;		/* epe, but vertical only (meters) */
    G_int fix;			/* type of position fix */
    G_double tow;		/* time of week (seconds) */
    Radian_Type posn;		/* latitude and longitude (radians) */
    G_float east;		/* velocity east (meters/second) */
    G_float north;		/* velocity north (meters/second) */
    G_float up;			/* velocity up (meters/second) */
    G_float msl_hght;		/* height of WGS 84 ellipsoid above MSL (meters) */
    G_int leap_scnds;		/* difference between GPS and UTC (seconds) */
    G_long wn_days;		/* week number days */
}
D800_Pvt_Data_Type;

enum
{
    unusable = 0,		       /* failed integrity check */
    invalid = 1,		       /* invalid or unavailable */
    _2D = 2,			       /* two dimensional */
    _3D = 3,			       /* three dimensional */
    _2D_diff = 4,		       /* two dimensional differential */
    _3D_diff = 5		       /* three dimensional differential */
};


/*                   Overlays :                              */


typedef struct
{
    byte id_l,			       /* product_ID */
      id_h,
      v_l,			       /* soft version */
      v_h,
      ident[24];
}
Raw_Product_Data_Type;		       /* overlay of type Product_Data_Type */

typedef struct
{
  byte tag,
  d_l,
  d_h;
}
Raw_Protocol_Data_type;		       /* overlay of type Protocol_Data_Type */

typedef struct
{				       /* version transmitted (packed) */
    char ident[6];
    byte lat[sizeof(G_long)];	       /* latitude                         4 */
    byte lon[sizeof(G_long)];	       /* longitude                        4 */
    byte unused[sizeof(G_longword)];
    char cmnt[40];
}
Raw_D100_Wpt_Type;

typedef struct
{					/* version transmitted (packed) */
    char ident[6];
    byte lat[sizeof(G_long)];		/* latitude                         4 */
    byte lon[sizeof(G_long)];		/* longitude                        4 */
    byte unused[sizeof(G_longword)];
    char cmnt[40];
    byte dst[sizeof(G_float)];		/* proximity distance (meters) */
    byte smbl[sizeof(Symbol_Type)];	/* symbol id */
}
Raw_D102_Wpt_Type;

typedef struct
{				       /* version transmitted (packed) */
    char ident[6];
    byte lat[sizeof(G_long)];	       /* latitude                         4 */
    byte lon[sizeof(G_long)];	       /* longitude                        4 */
    byte unused[sizeof(G_longword)];
    char cmnt [40];
    byte smbl;
    byte dspl;
}
Raw_D103_Wpt_Type;

typedef struct
{				       /* version transmitted (packed) */
    char ident[6];
    byte lat[sizeof(G_long)];	       /* latitude                         4 */
    byte lon[sizeof(G_long)];	       /* longitude                        4 */
    byte  unused[sizeof(G_longword)];
    char cmnt[40];
    byte dst[4];		       /* proximity distance (meters) */
    byte smbl[sizeof(Symbol_Type)];    /* symbol id */
    byte dspl;			       /* display option */
}
Raw_D104_Wpt_Type;

typedef struct
{
    byte lat[sizeof(G_long)];	       /* latitude                         4 */
    byte lon[sizeof(G_long)];	       /* longitude                        4 */
    byte smbl[sizeof(Symbol_Type)];    /* symbol id */
    char wpt_ident[244];	       /* don't know how much exactly */
}
Raw_D105_Wpt_Type;

typedef struct
{
    char ident[6];		       /* identifier */
    byte lat[sizeof(G_long)];	       /* latitude                         4 */
    byte lon[sizeof(G_long)];	       /* longitude                        4 */
    byte  unused[sizeof(G_longword)];
    char cmnt[40];		       /* comment */
    byte smbl;			       /* symbol id */
    byte dspl;			       /* display option */
    byte dst[4];		       /* proximity distance (meters) */
    byte color;			       /* waypoint color */
}
Raw_D107_Wpt_Type;

typedef struct
{
    byte wpt_class;		       /* class  1 */
    byte color;			       /* color  1 */
    byte dspl;			       /* display options  1 */
    byte attr;			       /* attributes  1 */
    byte smbl[sizeof(Symbol_Type)];    /* waypoint symbol 2 */
    byte subclass[18];		       /* subclass 18 */
    byte lat[sizeof(G_long)];	       /* latitude                         4 */
    byte lon[sizeof(G_long)];	       /* longitude                        4 */
    byte alt[sizeof(G_float)];	       /* altitude in meters               4 */
    byte dpth[sizeof(G_float)];	       /* depth in meters                  4 */
    byte dist[sizeof(G_float)];	       /* proximity distance in meters     4 */
    char state[2];		       /* state 2 */
    char cc[2];			       /* country code 2 */
    /* char ident[]; variable length string 1-51 */
    /* char comment[]; waypoint user comment 1-51 */
    /* char facility[]; facility name 1-31 */
    /* char city[]; city name 1-25 */
    /* char addr[]; address number 1-51 */
    /* char cross_road[]; intersecting road label 1-51 */
}
Raw_D108_Wpt_Type;

typedef struct
{
    byte dtyp;			       /* data packet type (0x01 for D109) 1 */
    byte wpt_class;		       /* class                            1 */
    byte dspl_color;		       /* display & color                  1 */
    byte attr;			       /* attributes (0x70 for D109)       1 */
    byte smbl[sizeof(Symbol_Type)];    /* waypoint symbol                  2 */
    byte subclass[18];		       /* subclass                        18 */
    byte lat[sizeof(G_long)];	       /* latitude                         4 */
    byte lon[sizeof(G_long)];	       /* longitude                        4 */
    byte alt[sizeof(G_float)];	       /* altitude in meters               4 */
    byte dpth[sizeof(G_float)];	       /* depth in meters                  4 */
    byte dist[sizeof(G_float)];	       /* proximity distance in meters     4 */
    char state[2];		       /* state                            2 */
    char cc[2];			       /* country code                     2 */
    byte ete[sizeof(G_longword)];      /* outbound link ete in seconds     4 */
    /* char ident[];                      variable length string 1-51 */
    /* char comment[];                    waypoint user comment 1-51 */
    /* char facility[];                   facility name 1-31 */
    /* char city[];                       city name 1-25 */
    /* char addr[];                       address number 1-51 */
    /* char cross_road[];                 intersecting road label 1-51 */
}
Raw_D109_Wpt_Type;  /* overlay of D109_Wpt_Type */

typedef struct
{
    byte class[sizeof(G_word)];		/* link class; see below */
    byte subclass[18];			/* sublcass */
    char ident[1];			/* variable length string */
}
Raw_D210_Rte_Link_Type;

typedef struct
{
    byte month;			       /* month (1-12) */
    byte day;			       /* day (1-31) */
    byte year_l;		       /* year (1990 means 1990) */
    byte year_h;
    byte hour_l;		       /* hour (0-23) */
    byte hour_h;
    byte minute;		       /* minute (0-59) */
    byte second;		       /* second (0-59) */
}
Raw_D600_Date_Time_Type;

typedef struct
{
  byte lat[sizeof(G_long)],	       /* Semicircle_Type posn */
  lon[sizeof(G_long)],
  time[sizeof(G_longword)];	       /* time */
  boolean new_trk;		       /* new track segment? */
}
Raw_D300_Trk_Point_Type;

typedef struct
{
  byte lat[sizeof(G_long)],	       /* Semicircle_Type posn */
  lon[sizeof(G_long)],
  time[sizeof(G_longword)];	       /* time */
  byte alt[sizeof(G_float)],	       /* altitude */
  dpth[sizeof(G_float)];	       /* depth */
  boolean new_trk;		       /* new track segment? */
}
Raw_D301_Trk_Point_Type;

typedef struct
{
    byte lat[sizeof (double)];	       /* latitude in radians */
    byte lon[sizeof (double)];	       /* longitude in radians */
}
Raw_Radian_Type;  /* overlay of the type Radian_Type */

typedef struct
{				       /* version transmise */
  byte alt[sizeof(G_float)],
  epe[sizeof(G_float)],
  eph[sizeof(G_float)],
  epv[sizeof(G_float)],
  fix[sizeof(G_int)],
  tow[sizeof(G_double)],
  posn[sizeof(Radian_Type)],
  east[sizeof(G_float)],
  north[sizeof(G_float)],
  up[sizeof(G_float)],
  msl_hght[sizeof(G_float)],
  leap_scnds[sizeof(G_int)],
  wn_days[sizeof(G_long)];
}
Raw_D800_Pvt_Data_Type;  /* overlay of D800_Pvt_Data_Type */


typedef struct
{				       /* speed negociation */
  byte speed[4];
}
Raw_Speed_Req_Type;


union GarminObject
{
    Raw_D100_Wpt_Type d100;
    Raw_D102_Wpt_Type d102;
    Raw_D103_Wpt_Type d103;
    Raw_D104_Wpt_Type d104;
    Raw_D105_Wpt_Type d105;
    Raw_D107_Wpt_Type d107;
    Raw_D108_Wpt_Type d108;
    Raw_D109_Wpt_Type d109;
    Raw_D300_Trk_Point_Type d300;
    Raw_D301_Trk_Point_Type d301;
    D310_Trk_Hdr_Type d310;
    Raw_D600_Date_Time_Type d600;
    D200_Rte_Hdr_Type d200;
    D201_Rte_Hdr_Type d201;
    D202_Rte_Hdr_Type d202;
};			/* generic Garmin object (data field of packets) */
typedef union GarminObject GarminObject;

//  local id of formats (must match order of FormatIDTable in gpslib.c) :
enum LocalFormatID
{
    Unknown = 0,
    G100Type, G101Type, G102Type, G103Type, G104Type, G105Type,
    G106Type, G107Type, G108Type,
    G200Type, G201Type, G202Type, G210Type,
    G300Type, G301Type, G310Type,
    G600Type,
    G700Type,
    G800Type,
    G109Type,
    GSpeedType,			       /* insert new Garmin types here */
    CustomWpt = 100,		       /* custom types */
    CustomTrkHdr, CustomTrkPt, CustomRteHdr, CompactTrkPt
    /* ,ToForceEnumSize=65535 */
} __attribute__((packed));
typedef enum LocalFormatID LocalFormatID;


#endif // __GARMIN_H__
