/*****************************************************************************
 *
 * $RCSfile: HandleMessage.c,v $
 *
 * GPS4Palm NMEA Message Handler function
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * Changes for Palm TX are Copyright (C) 07/2007 Frank Saurbier
 * <frank.saurbier@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2007-10-05 22:20:12 $
 *
 * $Revision: 1.22.2.1 $
 *
 * $Log: HandleMessage.c,v $
 * Revision 1.22.2.1  2007-10-05 22:20:12  mp
 * added Frank's changes (fix SkyviewForm on TX)
 *
 * Revision 1.23  2007/07/12 13:36:04  fs
 * delete sat view only one time
 *
 * Revision 1.22  2005/05/06 13:36:04  mp
 * added High-Density Display support for SkyviewForm
 *
 * Revision 1.21  2005/04/24 14:05:00  mp
 * added High-Density Display support for NavigationForm
 *
 * Revision 1.20  2005/02/19 15:50:47  mp
 * fixed invocation from WayptEditForm
 *
 * Revision 1.19  2004/12/19 10:07:09  mp
 * modified UpdateStatus()
 *
 * Revision 1.18  2004/12/11 13:45:39  mp
 * fixed GPS status display
 *
 * Revision 1.17  2004/12/10 19:52:40  mp
 * replaced DmGet1Resource by DmGetResource
 *
 * Revision 1.16  2004/12/09 17:34:03  mp
 * replaced strings by string resources
 *
 * Revision 1.15  2004/12/07 18:50:21  mp
 * added assignment to gGPSData.time
 *
 * Revision 1.14  2004/11/24 21:11:11  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.13  2004/11/23 17:52:49  mp
 * removed unused variables, fixed typo, added braces
 *
 * Revision 1.12  2004/03/20 13:37:06  mp
 * changed Position Form display
 *
 * Revision 1.11  2004/03/13 15:43:36  mp
 * Replaced usage of SetFieldTextFromStr() by SetFieldText(). Moved invocation
 * of Set_UTM_Parameters() to AppStart() (GPS.c). Modified message handling.
 *
 * Revision 1.10  2004/03/11 21:31:55  mp
 * Replaced gLon/gLat/gGPSok by equivalent gGPSData members, moved navigation
 * pointer display to NavigationForm.c. Changed from CMG to DTK.
 *
 * Revision 1.9  2004/03/06 15:57:22  mp
 * updated calling of cvt_atof() to new prototype
 *
 * Revision 1.8  2004/02/28 13:23:52  mp
 * modified UpdateLatAndLon(): now uses deg_to_dms_str()
 *
 * Revision 1.7  2004/01/18 19:02:48  mp
 * added UTP position display mode
 *
 * Revision 1.6  2003/12/31 14:59:21  mp
 * added comments, cleaned up code
 *
 * Revision 1.5  2003/12/30 19:47:58  mp
 * added unit conversion for PositionForm
 *
 * Revision 1.4  2003/12/28 18:41:49  mp
 * added GPS status flag for MapForm, added mapprefs debug info
 *
 * Revision 1.3  2003/11/20 20:37:23  mp
 * added update on request (STAT_REDRAW)
 *
 * Revision 1.2  2003/10/20 17:22:32  mp
 * changed UpdateLostSatellite to UpdateStatus
 *
 * Revision 1.1  2003/10/15 19:14:01  mp
 * moved code from GPS.c
 *
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <VFSMgr.h> // ABMHlogging
//#include <StringMgr.h>  // sprintf
#include "ResourceDefines.h"
#include "MathLib.h"			
#include "Sinetab.h"
#include "HandleMessage.h"
#include "GPS.h"
#include "NavigationForm.h"
#include "SkyviewForm.h"
#include "Utils.h"
#include "common.h"
#include "fp.h"
#include "utm.h"
#include "geo.h"

/* Satellite Information */
typedef struct { 
  Int16		prn;	/* Sat PRN */
  Int16		azi;	/* Azimuth, 0..359 deg. */
  Int16		ele;	/* Elevation, 0..90 deg. */
  UInt16	pwr;	/* SNR, 0..99 dBHz */
  Boolean	act;	/* Satellite active  flag (i.e. used for fix) */
} SatType;

/* Preferences data structure */
extern PrefsType gPrefs;

/* GPS Data */
GPSType gGPSData;

extern UInt32		gLastSuccessfulReception;
extern UInt32		gLastTimeDisplay;
extern UInt32		gNextReadTime;
extern WinHandle	gNavigationH;
extern WinHandle	gSkyViewH;
extern Boolean		gHdFtrSet;		/* High-Density Display */

#ifdef DEBUG_FORM
extern UInt32	gMinReceived;
extern UInt32	gMaxReceived;
extern UInt32	gChkErrs;
#endif


/* Constants for unit conversion and display */
const struct {
  char		*unit;
  double	conv;
} dst_c[] = {
            {"km", 1.0},
	    {"mi", 1.0/1.609344},
	    {"nm", 1.0/1.852}
	  };
 

const struct {
  char		*unit;
  double	conv;
} alt_c[] = {
            {"m",  1.0},
	    {"ft", 1.0/0.3048}
	  }; 


const struct {
  char		*unit;
  double	conv;
} spd_c[] = {
              {"km/h", 1.852},
	      {"m/s" , 1.852/3.6},
	      {"mi/h", 1.852/1.609344},
	      {"kts",  1.0}
	    };

/* Static Functions */
static Boolean GetField(const Char *buffer, UInt16 n,
  Char *result)        					MSGHANDLER_SECTION;
static void UpdatePosition(Char *message,
  double *p_lat, double *p_lon)				MSGHANDLER_SECTION;
static void UpdateLatAndLon(Char *message)		MSGHANDLER_SECTION;
static Boolean DecodeTimeOfFix(Char *message, Char *s)	MSGHANDLER_SECTION;


/*****************************************************************************
 * FUNCTION:	HandleMessage
 *
 * DESCRIPTION: Decodes NMEA message, updates global variables (position, GPS
 *		status and updates display depending on currently selected
 *		form.
 *
 * PARAMETERS:	message	-- pointer to message
 *
 * RETURNED:    updated display flag
 ****************************************************************************/
Boolean HandleMessage(Char *message)
{
  Char		s[20];				/* string buffer */
  UInt32	now = TimGetTicks();		/* current system time */
  Boolean	updatedDisplay = false;		/* flag */
  UInt16	form = FrmGetActiveFormID();	/* form ID */
  Char		*msg;				/* message pointer */
  RectangleType	rect;				/* rectangle structure */
  UInt8		bat;				/* battery status (percent) */
  static int	prn[12];			/* PRN if SAT is used for fix, 
  						   -1 otherwise */
  float		value;				/* var. for unit conversion */
  UInt16	i;				/* loop index */
  static SatType sat[12];			/* current satellite info */
  static SatType old[12];			/* previous satellite info */
  

  if (gGPSData.scroll) {
    gGPSData.valid = 1;
    return 1;
  }

  /*
   * Message type: RMC
   *
   * Data:	Date and time of fix
   *		Latitude and Longitude
   *		Status
   *		Speed over ground
   *		Course made good
   *		Magnetic variation
   *
   * Forms:	Position, Skyview, Navigation, Map
   */
  if ( (msg = StrStr(message, "RMC")) ) {
    Int32		x, y;			/* screen coordinates */
    Int32		x2, y2;			/*   for navigation form */
    Int32		r1, r2, r3;		/* radii (navigation form) */
    PointType		p1;			/* point (navigation form) */
    Int32		phi;			/* angle for navigation form */
    FlpCompDouble	dtrk;			/* track as double */
    Int32		trk;			/* track as integer */
    static Int32	otrk;			/* old track */
    const char 		compass[] = "N\0E\0S\0W\0"; /* compass directions */
    Boolean		evenMultipleOf5Seconds;
    
    gLastSuccessfulReception = now;	/* we successfully read */

    /*
     * Get Time
     */
    if (GetField(msg, 1, s)) {
      evenMultipleOf5Seconds = (s[5] == '0' || s[5] == '5');
      
      /* 
       * even multiple of five seconds OR it's been at 
       * least kMaxTicksWithoutTime seconds since a display
       * That way, if we lose 11:11:35, we won't have the
       * time go from 11:11:30 to 11:11:40. Instead, it'll go
       * 11:11:30, 11:11:36, 11:11:40
       */
      if (evenMultipleOf5Seconds ||
	(now - gLastTimeDisplay) > gMaxTicksWithoutTime) {

	/*
	 * Most of the time, we'll be on a multiple of five. 
	 * Thus, we want to read in four and a quarter more seconds.
	 * Otherwise, we want to read immediately
	 */
	if (evenMultipleOf5Seconds) {
	  gNextReadTime = gLastSuccessfulReception + 
		  4*gTicksPerSecond + gTicksPerSecond / 4;
	} else {
	  gNextReadTime = 0;
	}	
      } /* if (evenMultipleOf5Seconds || ...) */
      
      /* Update Time */
      if ( DecodeTimeOfFix(msg, s) ) {
	DateTimeType	dt;

	if (form == GPSMainForm || form == SkyviewForm || form == NavigationForm) {
	  /* update the time display */
	  SetFieldText(TimeField, s, false, true);
	}

	/* convert string to date/time struct */
	dt.second = StrAToI(&s[6]);
	dt.minute = StrAToI(&s[3]);
	dt.hour   = StrAToI(&s[0]);
	dt.day    = StrAToI(&s[9]);
	dt.month  = StrAToI(&s[12]);
	dt.year   = 2000 + StrAToI(&s[15]);

        /* copy GPS-time to gGPSData */
	gGPSData.time = TimDateTimeToSeconds(&dt);

	gLastTimeDisplay = gLastSuccessfulReception;
      }
   
    } /* if (GetField(msg, 1, s)) */

    /* 
     * Get GPS Status
     */ 
    if (GetField(msg, 2, s)) {
      if (s[0] == 'V') {
	UpdateStatus( STAT_WARNING );
	gGPSData.valid = false;
      } else {
	UpdateStatus( STAT_OK );
	gGPSData.valid = true;
      }
    } /* if (GetField(msg, 2, s)) */

    /*
     * Convert string to lat/lon double values
     */
    if (gGPSData.valid) {
      UpdatePosition(msg, &gGPSData.lat, &gGPSData.lon);
    }
    
    switch (form) {

      case GPSMainForm:
	if (gGPSData.valid) {      
	  /* display position */
	  UpdateLatAndLon(msg);

	  if (GetField(msg, 7, s)) {
	    /* Speed (with unit conversion) */
	    value = cvt_atof(s, NULL, NULL);
	    gGPSData.sog = value; 
	    value *= spd_c[gPrefs.units.spd_unit].conv;
	    format_number(value, 2, s);
	    StrCat(s, " ");
	    StrCat(s, spd_c[gPrefs.units.spd_unit].unit);

	    SetFieldText(SpeedField, s, false, true);
	  }
	  if (GetField(msg, 8, s)) {
	    /* Course Made Good */
	    StrCat(s, "");
	    SetFieldText(CMGField, s, false, true);
	  }
	  if (GetField(msg, 10, s) && 
	      GetField(msg, 11, s + StrLen(s))) {
	    /* Magnetic Variation */
	    StrCat(s, "");
	    SetFieldText(MagField, s, false, true);
	  }
	  updatedDisplay = true;
	} else {
	  SetFieldText(SpeedField, NULL, false, true);
	  SetFieldText(CMGField, NULL, false, true);
	  SetFieldText(MagField, NULL, false, true);
	} /* if (gGPSData.valid) */

      /* GPSMainForm */
      break;

      case SkyviewForm:
	SysBatteryInfo (false /* set */, NULL, NULL, NULL, NULL, NULL,
	                  &bat /* *percentP */);
        StrIToA(s, bat);
	StrCat(s, "%");
	SetFieldText(BatField, s, false, true);
	
	if (gGPSData.valid)
      	  updatedDisplay = true;
	
      /* SkyviewForm */
      break;
      
      case NavigationForm:
	    if (GetField(msg, 7, s)) {
	      /* Speed Over Ground */
	      value = cvt_atof(s, NULL, NULL);
	      gGPSData.sog = value; 
	    }

            if (GetField(msg, 8, s)) {
	      /* get CMG field */
              /* get floating point value from string */
              FlpBufferAToF(&dtrk.fd, s);
              gGPSData.cmg = dtrk.d;
	      
	      /* store "double"-field of union FlpCompDouble in variable trk */
              trk = (Int32)dtrk.d - 90;
	      
	      if (gHdFtrSet) {
		WinPushDrawState();
		WinSetCoordinateSystem(kCoordinatesNative);
	      } 
	      
	      /* Erase Compass Directions */
	      r1 = (gHdFtrSet) ? WinScaleCoord(NAV_R1, false) : NAV_R1;
	      r2 = (gHdFtrSet) ? WinScaleCoord(NAV_R2, false) : NAV_R2;
	      r3 = (gHdFtrSet) ? WinScaleCoord(NAV_R3, false) : NAV_R3;
	      p1.x = (gHdFtrSet) ? WinScaleCoord(NAV_X, false) : NAV_X;
	      p1.y = (gHdFtrSet) ? WinScaleCoord(NAV_Y, false) : NAV_Y;
	      for (i = 0, phi = 45; i <= 3; i++, phi += 90) {
	        Int16	w = FntCharWidth(compass[2*i]);
		Int16	h = FntCharHeight();
		
		x  = p1.x + r1 * icos(phi - otrk) / 32768;
	        y  = p1.y + r1 * isin(phi - otrk) / 32768;
	        x2 = p1.x + r2 * icos(phi - otrk) / 32768;
	        y2 = p1.y + r2 * isin(phi - otrk) / 32768;
	        WinEraseLine(x, y, x2, y2);
	        x  = p1.x + r3 * icos(180 - otrk + i*90) / 32768;
	        y  = p1.y + r3 * isin(180 - otrk + i*90) / 32768;
		WinEraseChars(&compass[2*i], 1, x - w/2, y - h/2);
	      }

	      /* Restore Compass */
	      p1.x = NAV_X-NAV_R2;
	      p1.y = NAV_Y-NAV_R2;
	      if (gHdFtrSet) {
		WinScalePoint(&p1, false);
	      }
	      x = y = (gHdFtrSet) ? WinScaleCoord(2 * NAV_R2, false)
	                          : 2 * NAV_R2;
	 
	      RctSetRectangle(&rect, 0, 0, x /* w */, y /* h */);

	      WinCopyRectangle(	gNavigationH, /* WinHandle srcWin */
	  			NULL, /* WinHandle dstWin, draw window */
				&rect, /* RectangleType *srcRect */
				p1.x, /* Coord destX */
				p1.y, /* Coord destY */
				winOverlay /* WinDrawOperation mode */);
	 
	      /* Draw Compass Directions */
	      p1.x = (gHdFtrSet) ? WinScaleCoord(NAV_X, false) : NAV_X;
	      p1.y = (gHdFtrSet) ? WinScaleCoord(NAV_Y, false) : NAV_Y;
     	      for (i = 0, phi = 45; i <= 3; i++, phi += 90) {
	        Int16	w = FntCharWidth(compass[2*i]);
		Int16	h = FntCharHeight();

	        x  = p1.x + r1 * icos(phi - trk) / 32768;
	        y  = p1.y + r1 * isin(phi - trk) / 32768;
	        x2 = p1.x + r2 * icos(phi - trk) / 32768;
	        y2 = p1.y + r2 * isin(phi - trk) / 32768;
	        WinDrawLine(x, y, x2, y2);
	        x  = p1.x + r3 * icos(180 - trk + i*90) / 32768;
	        y  = p1.y + r3 * isin(180 - trk + i*90) / 32768;
		WinDrawChars(&compass[2*i], 1, x - w/2, y - h/2);
	      }
	 
	      if (gHdFtrSet) {
		WinPopDrawState();
	      } 

	      /* save old track */
	      otrk = trk;
	    }

      /* NavigationForm: */
      break;

      case MapForm:
	// this was the reason why the direction was not displayed correctly
	    if (GetField(msg, 7, s)) {
	      /* Speed Over Ground */
	      value = cvt_atof(s, NULL, NULL);
	      gGPSData.sog = value; 
	    }

            if (GetField(msg, 8, s)) {
	      /* get CMG field */
              /* get floating point value from string */
              FlpBufferAToF(&dtrk.fd, s);
              gGPSData.cmg = dtrk.d;
	      
	      /* store "double"-field of union FlpCompDouble in variable trk */
              trk = (Int32)dtrk.d - 90;
	    }
	/* Note: further processing is done in MapFormHandleEvent() */	

      /* MapForm */
      break;

#ifdef DEBUG_FORM
     case DebugForm:
       {
         Char tmp[40];
	 
	 StrCopy(tmp, "RcvMin:           ");
	 StrPrintF(&tmp[8], "%5lu", gMinReceived);
         WinDrawChars(tmp, StrLen(tmp), 0, 15);
	 
	 StrCopy(tmp, "RcvMax:           ");
	 StrPrintF(&tmp[8], "%5lu", gMaxReceived);
	 WinDrawChars(tmp, StrLen(tmp), 0, 30);
	 
	 StrCopy(tmp, "ChkErr:           ");
	 StrPrintF(&tmp[8], "%lu", gChkErrs);
	 WinDrawChars(tmp, StrLen(tmp), 0, 45);

	 StrCopy(tmp, "Map Select: ");
	 StrCat(tmp,
	   (gPrefs.mapprefs.select == SEL_USER) ? "User" : "Auto");
	 WinDrawChars(tmp, StrLen(tmp), 0, 60);

	 StrCopy(tmp, "Fetch State: ");
	 if (gPrefs.mapprefs.fetch == FETCH_READY) {
	   StrCat(tmp, "Ready");
	 } else if (gPrefs.mapprefs.fetch == FETCH_START) {
	   StrCat(tmp, "Start");
	 } else {
	   StrCat(tmp, "Stop ");
	 }
	 WinDrawChars(tmp, StrLen(tmp), 0, 75);
       }
       updatedDisplay = true;
     /* DebugForm */
     break;     
#endif      
    } /* switch (form) */
  } /* if (StrStr(message, "RMC")) */

  /*
   * Message type: GSV
   *
   * Data:	Number of satellites in view,
   *		Satellite Nr (PNR), Elevation, Azimuth, Signal/Noise Ratio (SNR)
   *		(per satellite)
   *
   * Forms:	Skyview
   */
  if ((msg = StrStr(message, "$GPGSV")) && (form == SkyviewForm)) {
    UInt16		n;		/* loop index */
    UInt16		f;		/* field index */
    Int32		x, y;		/* screen coordinates */
    PointType		p1;		/* point */
    Coord		w;		/* object width */ 
    Coord		h;		/* object height */
    Int32		r;		/* radius (screen) */
    Int32		phi;		/* angle (screen) */
    UInt16		nr_sats = 0;	/* nr. of sats in view */
    UInt16		curr;		/* current GSV message */
    
    /* Sats in view */
    if (GetField(msg, 3, s)) {
      SetFieldText(SatsField, s, false, true);
      nr_sats = StrAToI(s);
    }
    
    /* clear data in unused entries */
    for (i=nr_sats; i<12; i++) {
      sat[i].prn = -1;
    } 
    
    /*
     * decode current sentence
     *   f: field index within message
     *   i: sat index   
     */
    if (GetField(msg, 2, s)) {
      curr = StrAToI(s) - 1;
      
      /* first sat data starts at 4th field in a message */
      f = 4;
      
      /* loop all sats in current message (max. 4) */
      for (i=curr*4; (i < nr_sats) && (i < curr*4+4); i++) {
	if (GetField(msg, f++, s)) {
	  sat[i].prn = StrAToI(s);
	}
	if (GetField(msg, f++, s)) {
	  sat[i].ele = StrAToI(s);
	}
	if (GetField(msg, f++, s)) {
	  sat[i].azi = StrAToI(s);
	}
	if (GetField(msg, f++, s)) {
	  sat[i].pwr = StrAToI(s);
	}
      }
    } /* if (GetField(msg, 2, s)) */
    
    if (gHdFtrSet) {
      WinPushDrawState();
      WinSetCoordinateSystem(kCoordinatesNative);
    }


/********************************************************************************
* (fs) delete sat-view one time - now I can see all satellites
*********************************************************************************/
    p1.x = (gHdFtrSet) ? WinScaleCoord(SKY_X-SKY_R2, false) : SKY_X-SKY_R2;
    p1.y = (gHdFtrSet) ? WinScaleCoord(SKY_Y-SKY_R2, false) : SKY_Y-SKY_R2;
    WinCopyRectangle(	gSkyViewH, /* WinHandle srcWin */
			NULL, /* WinHandle dstWin, draw window */
			&rect, /* RectangleType *srcRect */
			p1.x, /* Coord destX */
			p1.y, /* Coord destY */
			winOverlay /* WinDrawOperation mode */);


    for (i=0; i<12; i++) {
      /* check if current satellite is active (i.e. used for fix) */
      sat[i].act = false;
      for (n=0; n<12; n++) {
        if (prn[n] == sat[i].prn) {
	      sat[i].act = true;
	      break;
        }
      }

      /* erase previous data if it has been valid */
      if (old[i].prn > -1) {
	/* PRN */
	StrPrintF(s, "%02d", old[i].prn);
	p1.x = (gHdFtrSet) ? WinScaleCoord(9+(i*12), false) : 9+(i*12);
	p1.y = (gHdFtrSet) ? WinScaleCoord(SKY_L2+1, false) : SKY_L2+1;
        WinEraseChars(s, StrLen(s), p1.x, p1.y);
	
	/* PWR (bar graphs) */
	p1.y = (gHdFtrSet) ? WinScaleCoord(SKY_L2-old[i].pwr/4, false)
	                   : SKY_L2-old[i].pwr/4;
	h = (gHdFtrSet) ? WinScaleCoord(old[i].pwr/4, false) : old[i].pwr/4;
	if (old[i].act) {
	  w = (gHdFtrSet) ? WinScaleCoord(8, false) : 8;
	  RctSetRectangle(&rect, p1.x, p1.y, w, h);
	  WinEraseRectangle(&rect, 0 /* cornerDiam*/ );
	} else {
	  p1.x = (gHdFtrSet) ? WinScaleCoord(10+(i*12), false) : 10+(i*12);
	  w = (gHdFtrSet) ? WinScaleCoord(6, false) : 6;
	  RctSetRectangle(&rect, p1.x, p1.y, w, (h > 0) ? h-1 : 0 /* h */);
	  WinEraseRectangleFrame(simpleFrame, &rect);
	}
	
	/* Azimut and Elevation (PRNs on vector display) */
	p1.x = (gHdFtrSet) ? WinScaleCoord(SKY_X, false) : SKY_X;
	p1.y = (gHdFtrSet) ? WinScaleCoord(SKY_Y, false) : SKY_Y;
	phi = old[i].azi - 90;
	r = (gHdFtrSet) ? WinScaleCoord((90-old[i].ele)/2, false)
	                : (90-old[i].ele)/2;
	x = p1.x + r * icos(phi) / 32768;
	y = p1.y + r * isin(phi) / 32768;
	StrPrintF(s, "%02d", old[i].prn);
	WinEraseChars(s, StrLen(s), x - FntCharsWidth(s, StrLen(s))/2,
	  y - FntCharHeight()/2);
	w = h = (gHdFtrSet) ? WinScaleCoord(2*SKY_R2, false) : 2*SKY_R2;
        RctSetRectangle(&rect, 0, 0, w, h);

      } /* if (old[i].prn > -1) */
	
      /* show new data if valid (sat in view) */
      if (sat[i].prn != -1) {  
	/* PRN */
	StrPrintF(s, "%02d", sat[i].prn);
	p1.x = (gHdFtrSet) ? WinScaleCoord(9+(i*12), false) : 9+(i*12);
	p1.y = (gHdFtrSet) ? WinScaleCoord(SKY_L2+1, false) : SKY_L2+1;
	WinDrawChars(s, StrLen(s), p1.x, p1.y);

	/* PWR (bar graphs) */
	p1.y = (gHdFtrSet) ? WinScaleCoord(SKY_L2-sat[i].pwr/4, false)
	                   : SKY_L2-sat[i].pwr/4;
	h = (gHdFtrSet) ? WinScaleCoord(sat[i].pwr/4, false) : sat[i].pwr/4;
	/* sat active: draw solid bar / sat inactive: draw empty bar */	  
	if (sat[i].act) {
	  w = (gHdFtrSet) ? WinScaleCoord(8, false) : 8;
          RctSetRectangle(&rect, p1.x, p1.y, w, h);  
	  WinDrawRectangle(&rect, 0 /* cornerDiam*/ );
	} else {
	  p1.x = (gHdFtrSet) ? WinScaleCoord(10+(i*12), false) : 10+(i*12);
	  w = (gHdFtrSet) ? WinScaleCoord(6, false) : 6;
	  if (h > 0) {
	    RctSetRectangle(&rect, p1.x, p1.y, w, h-1);
            WinDrawRectangleFrame(simpleFrame, &rect);
	  }
	}

	/* Azimut and Elevation (PRNs on vector display) */
	p1.x = (gHdFtrSet) ? WinScaleCoord(SKY_X, false) : SKY_X;
	p1.y = (gHdFtrSet) ? WinScaleCoord(SKY_Y, false) : SKY_Y;
	phi = sat[i].azi - 90;
	r = (gHdFtrSet) ? WinScaleCoord((90-sat[i].ele)/2, false)
	                : (90-sat[i].ele)/2;
	x = p1.x + r * icos(phi) / 32768;
	y = p1.y + r * isin(phi) / 32768;
	StrPrintF(s, "%02d", sat[i].prn);
	WinDrawChars(s, StrLen(s), x - FntCharsWidth(s, StrLen(s))/2,
	  y - FntCharHeight()/2);

      } /* if (sat[i].prn != -1) */
      
      /* save current dataset */
      old[i] = sat[i];
    } /* for (i=0; i<12; i++) */
    
    /* Base line for S/N bar graph */
    /* get screen dimensions */
    WinGetDisplayExtent(&h, &w);
    p1.y = (gHdFtrSet) ? WinScaleCoord(SKY_L2, false) : SKY_L2;
    WinDrawLine(0, p1.y, w-1, p1.y);
     
#ifdef VERBOSE
    RctSetRectangle(&rect, 0, 45, w, 48 /* h */);
    WinEraseRectangle(&rect, 0);
#endif

    if (gHdFtrSet) {
      WinPopDrawState();
    }

    while ( (msg = StrStr(msg, "$GPGSV")) ) {
#ifdef VERBOSE
      WinDrawChars(msg, StrLen(msg), 0, i);
      i += 15;
#endif
      if ((msg = StrStr(msg, "*")) == NULL)
        break; 
    }
  } /* (StrStr(message, "GSV") */

  /*
   * Message type: GSA
   *
   * Data:	Type of fix (none/2D/3D)
   *		Satellites used for fix
   *		Precision of fix (PDOP/HDOP/VDOP)
   *
   * Forms:	Skyview
   */
  if ((msg = StrStr(message, "$GPGSA")) && (form == SkyviewForm)) {
    UInt16	i;
    
    /* Type of fix (none/2D/3D) */
    if (GetField(msg, 2, s)) {
      if ((s[0] == '2') || (s[0] == '3')) {
        StrCat(s, "D");
      } else {
        StrCopy(s, "none");
      }
      SetFieldText(FixField, s, false, true);
    }

    /* Satellites used for fix (either PRN or empty field) */
    for (i=0; i<=12;i++) {
      if (GetField(msg, 3+i, s)) {
        if (s[0] != '\0') {
	  /* store PRN */
	  prn[i] = StrAToI(s);
	} else {
	  /* not used */
	  prn[i] = -1;
	}
      }
    } /* for (i=0; i<=12;i++) */

    /* Position DOP */
    if (GetField(msg, 15, s)) {
      SetFieldText(PDOPField, s, false,	true);
    }
    
    /* Horizontal DOP */
    if (GetField(msg, 16, s)) {
      SetFieldText(HDOPField, s, false, true);
    }
    
    /* Vertical DOP */
    if (GetField(msg, 17, s)) {
      SetFieldText(VDOPField, s, false, true);
    }
  } /* if ((msg = StrStr(message, "$GPGSV")) && (form == SkyviewForm)) */

  /*
   * Message type: GGA
   *
   * Data:	Altitude
   *
   * Forms:	Position
   */
  if ((msg = StrStr(message, "$GPGGA")) && (form == GPSMainForm)) {
    if (gGPSData.valid) {
      /* Altitude */
      if (GetField(msg, 9, s)) {
	value = cvt_atof(s, NULL, NULL) * alt_c[gPrefs.units.alt_unit].conv;
	format_number(value, 1, s);
	StrCat(s, " ");
	StrCat(s, alt_c[gPrefs.units.alt_unit].unit);

	SetFieldText(AltField, s, false, true);
      }
    } else {
      SetFieldText(AltField, NULL, false, true);
    }
  }


  // ABMH logging
  {
    Err err = errNone;
    UInt32 volumeIterator = 0;
    UInt16 firstVolume = 0;
    FileRef f;
    //    UInt8 *bufferP = NULL;
    UInt8 bufferP[255];
    UInt32	seconds;		/* current time */
    DateTimeType	dt;			/* date/time structure */
    UInt32 written = 0;
    static UInt32	last_seconds = 0;
    double	tmp;
    float		value;				/* var. for unit conversion */

    seconds = TimGetSeconds();
    if ((gGPSData.valid) && (seconds - last_seconds >= 10)) {
      last_seconds = seconds;

      // Get first card
      volumeIterator = vfsIteratorStart;
      err = VFSVolumeEnumerate(&firstVolume, &volumeIterator);
      if (!(err == expErrEnumerationEmpty)) {
	
	f = NULL;
	err = VFSFileOpen(firstVolume, "/gps4palm.log", vfsModeWrite|vfsModeCreate, &f);
	if (!err) {
	  
	  TimSecondsToDateTime(seconds, &dt);
	  
	  /* write date to string */
	  DateToAscii(dt.month, dt.day, dt.year,
		      dfYMDWithDashes,
		      bufferP);
	  StrCat(bufferP, " ");
	  
	  /* write time to string */
	  TimeToAscii(dt.hour, dt.minute,
		      tfColon24h,
		      &bufferP[StrLen(bufferP)]);
	  StrCat(bufferP, " ");

	  /*  sprintf(bufferP, "%04d-%02d-%02d %02d:%02d ",
		  dt.year, dt.month, dt.day,
		  dt.hour, dt.minute);
	  */

	  format_number(gGPSData.lon, 7, &bufferP[StrLen(bufferP)]);
	  StrCat(bufferP, " ");
	  format_number(gGPSData.lat, 7, &bufferP[StrLen(bufferP)]);
	  StrCat(bufferP, " ");

	  tmp = gGPSData.sog * spd_c[gPrefs.units.spd_unit].conv;
	  format_number(tmp, 0, &bufferP[StrLen(bufferP)]);
	  StrCat(bufferP, " ");
	  StrCat(bufferP, spd_c[gPrefs.units.spd_unit].unit);
	  StrCat(bufferP, " ");

	  value = cvt_atof(s, NULL, NULL) * alt_c[gPrefs.units.alt_unit].conv;
	  format_number(value, 1, &bufferP[StrLen(bufferP)]);
	  StrCat(bufferP, " ");
	  StrCat(bufferP, alt_c[gPrefs.units.alt_unit].unit);
	  StrCat(bufferP, "\n");
	  
	  VFSFileSeek(f, vfsOriginEnd, 0);
	  err = VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
	  VFSFileClose(f);      
	}
      }
    }

  }

  return updatedDisplay;
}

/*****************************************************************************  
 * FUNCTION:	GetField
 *
 * DESCRIPTION:	returns n'th (0-based) comma-delimited field within buffer
 *
 * PARAMETERS:	buffer	-- pointer to message buffer
 *		n	-- (0-based field number)
 *		result	-- pointer to result buffer 
 *
 * RETURNED:    true if field found, false otherwise
 ****************************************************************************/
static
Boolean GetField(const Char *buffer, UInt16 n, Char *result)
{
  int	i;		/* loop index */

  /* skip n commas */
  for (i = 0; i < n; i++) {
    while (*buffer && *buffer != ',')
      buffer++;
    if (*buffer == '\0')
      return false;
    buffer++;
  }
  
  /* read until end of field (either comma or asterisk) */
  while (*buffer && *buffer != ',' && *buffer != '*')
    *result++ = *buffer++;
  
  /* terminate result string */
  *result = '\0';
  
  /* ok/fail */
  return *buffer == ',' || *buffer == '\0' || *buffer == '*';
}


/*****************************************************************************  
 * FUNCTION:	DecodeTimeOfFix
 *
 * DESCRIPTION:	Decodes date and time of fix from message
 *
 * PARAMETERS:	message	-- pointer to message buffer
 *		s	-- pointer to result buffer
 *
 * RETURNED:    true if field found, false otherwise
 ****************************************************************************/
static
Boolean DecodeTimeOfFix(Char *message, Char *s)
{
  Boolean result = false;
  
  /* Get time and date of fix (UTC) */
  if (GetField(message, 1, s) && GetField(message, 9, &s[8])) {
    /* change from HHMMSS.DDD to HH:MM:SS */
    s[7] = s[5];
    s[6] = s[4];
    s[5] = ':';
    s[4] = s[3];
    s[3] = s[2];
    s[2] = ':';

    /* change from DDMMYY to DD/MM/YY */
    s[17] = '\0';
    s[16] = s[13];
    s[15] = s[12];
    s[14] = '/';
    s[13] = s[11];
    s[12] = s[10];
    s[11] = '/';
    s[10] = s[9];
    s[ 9] = s[8];
    s[ 8] = ' ';
    result = true;
  } /* if (Time- & Date-Field available) */
  
  return result;
}


/*****************************************************************************  
 * FUNCTION:	UpdateLatAndLon
 *
 * DESCRIPTION:	Decodes and displays Latitude and Longitude according to
 *		user selected format.
 *
 * PARAMETERS:	message	-- pointer to message buffer
 *
 ****************************************************************************/
static 
void UpdateLatAndLon(Char *message)
{
  Char		s[40];		/* string buffer */
  Char		lat_str[20];	/* string buffer */
  Char		lon_str[20];	/* string buffer */
  Char 		*pt_pos;	/* decimal point pos. */
  double	lat;		/* latitude */
  double	lon;		/* longitude */
  UInt16	i;		/* string index */
  Int32		err;		/* error code */
  Int32		zone;		/* UTM Zone */
  char		hemisphere;	/* 'N' / 'S' */
  double	easting;	/* Easting in m */
  double	northing;	/* Northing in m */
  
  switch (gPrefs.units.pos_unit) {
    case POS_D:
      /* Degrees - convert from lat/lon double values */
      if (gGPSData.lat >= 0) {
        StrCopy(s, "N");
	format_number(gGPSData.lat, 7, &s[1]);
      } else {
        StrCopy(s, "S");
	format_number(-gGPSData.lat, 7, &s[1]);
      }   
      StrCat(s, "");
      SetFieldText(LatitudeField, s, false, true);
      
      if (gGPSData.lon >= 0) {
        StrCopy(s, "E");
	format_number(gGPSData.lon, 7, &s[1]);
      } else {
        StrCopy(s, "W");
	format_number(-gGPSData.lon, 7, &s[1]);
      }   
      StrCat(s, "");
      SetFieldText(LongitudeField, s, false, true);
      break;
    
    case POS_DM:
      /* Degrees and Minutes - only a little string patching needed */
      /* 4 is N or S for Lat direction, 3 is lat */
      if (GetField(message, 4, s) && 
	      GetField(message, 3, s + StrLen(s))) {

	/* find position of decimal point */
	pt_pos = StrStr(s, ".");

	/* insert "" between degree and integer minutes and "'" at the end */
	for (i = StrLen(s); &s[i]>=pt_pos-2 ; i--) {
	  s[i+1] = s[i];
	}
	s[++i] = '';
	StrCat(s, "'");

	SetFieldText(LatitudeField, s, false, true);
      }

      /* 6 is E or W for Lon direction, 5 is lon */
      if (GetField(message, 6, s) && 
	      GetField(message, 5, s + StrLen(s))) {

	/* find position of decimal point */
	pt_pos = StrStr(s, ".");

	/* insert "" between degree and integer minutes and "'" at the end */
	for (i = StrLen(s); &s[i]>=pt_pos-2 ; i--) {
	  s[i+1] = s[i];
	}
	s[++i] = '';
	StrCat(s, "'");

	SetFieldText(LongitudeField, s, false, true);
      }
      break;
      
    case POS_DMS:
      /* Degrees, Minutes, Seconds - convert from lat/lon double values */
      lat = (gGPSData.lat >= 0) ? gGPSData.lat : -gGPSData.lat;
      StrCopy(lat_str, (gGPSData.lat >= 0) ? "N" : "S");

      lon = (gGPSData.lon >= 0) ? gGPSData.lon : -gGPSData.lon;
      StrCopy(lon_str, (gGPSData.lon >= 0) ? "E" : "W");
      
      deg_to_dms_str(lat, lon, &lat_str[1], &lon_str[1]);
      
      SetFieldText(LatitudeField, lat_str, false, true);     
      SetFieldText(LongitudeField, lon_str, false, true);

      break;

    case POS_UTM:
      /* Universal Transverse Mercator - convert from lat/lon double values */ 
      err = Convert_Geodetic_To_UTM (
      				deg2rad(gGPSData.lat) /* Latitude (rad) */,
                                deg2rad(gGPSData.lon) /* Longitude (rad) */,
                                &zone /* long  *Zone */,
                                &hemisphere /* Hemisphere */,
                                &easting /* *Easting */,
                                &northing /* *Northing */);
      
      if (err == UTM_NO_ERROR) {
        /* print zone and belt to string */
	StrPrintF(s, "%ld%c ", zone, Lat_To_UTM_Belt(gGPSData.lat));
	
	/* print easting to string */
	format_number(easting, 1, &s[StrLen(s)]);
        SetFieldText(LatitudeField, s, false, true);
		
	/* print northing to string */
	format_number(northing, 1, s);
        SetFieldText(LongitudeField, s, false, true);
	
	
      } else {
        SetFieldText(LatitudeField, "n.a.", false, true);
        SetFieldText(LongitudeField, "", false, true);
      }
      
      break;
  } /* switch (gPrefs.units.pos_unit) */
}

/*****************************************************************************
 * FUNCTION:	UpdatePosition
 *
 * DESCRIPTION:	Updates floating point Latitude/Longitude variables
 *
 * PARAMETERS:	message	-- pointer to message buffer
 *		p_lat	-- pointer to latitude variable
 *		p_lon	-- pointer to longitude variable
 *
 ****************************************************************************/
void UpdatePosition(Char *message, double *p_lat, double *p_lon)
{
  Char		s[20];		/* string buffer */
  Char		tmp[8];		/* temp. string */
  Char 		*pt_pos;	/* decimal point pos. */
  double	lat;		/* latitude (deg) */
  double	lon;		/* longitude (deg) */
  FlpCompDouble	min;		/* minutes */
  
  /* 4 is N or S for Lat direction, 3 is lat */
  if (GetField(message, 4, s) && 
	  GetField(message, 3, s + StrLen(s))) {

    /* find position of decimal point */
    pt_pos = StrStr(s, ".");

    /* degrees */
    StrNCopy(tmp, pt_pos - 4, 2);
    tmp[2] = '\0';
    lat = (double)StrAToI(tmp);
    
    /* minutes */
    StrCopy(tmp, pt_pos - 2);
    
    /* get floating point value from string */
    FlpBufferAToF(&min.fd, tmp);
    
    lat += (double)min.d/60.0;
    
    
    lat *= (s[0] == 'S') ? -1 : 1; 
    
    *p_lat = lat;
  }
  
  
  /* 6 is E or W for Lon direction, 5 is lon */
  if (GetField(message, 6, s) && 
	  GetField(message, 5, s + StrLen(s))) {

    /* find position of decimal point */
    pt_pos = StrStr(s, ".");

    /* degrees */
    StrNCopy(tmp, pt_pos - 5, 3);
    tmp[3] = '\0';
    lon = (double)StrAToI(tmp);
    
    /* minutes */
    StrCopy(tmp, pt_pos - 2);

    /* get floating point value from string */
    FlpBufferAToF(&min.fd, tmp);
    
    lon += (double)min.d/60.0;
    
    lon *= (s[0] == 'W') ? -1 : 1; 
    
    *p_lon = lon;
  }
}


/*****************************************************************************
 * FUNCTION:	UpdateStatus
 *
 * DESCRIPTION: Update Status Fields
 *
 * PARAMETERS:	status
 *
 *		STAT_NO_GPS
 *		STAT_WARNING
 *		STAT_OK
 *
 ****************************************************************************/
void UpdateStatus(UInt16 status)
{
  UInt16		formID = FrmGetActiveFormID();	/* current form ID */
  static UInt16		prevID;				/* previous form ID */
  static UInt16		prevStat;			/* previous status */
  MemHandle		strH = NULL;			/* string handle */
  MemPtr		strP = NULL;			/* string ptr */
  UInt16		str_id;				/* string resource ID */

  /* Update Status Field */
  if ( (status != prevStat) || (formID != prevID) || (status == STAT_REDRAW) ) {
    
    if ( status == STAT_REDRAW ) {
      status = prevStat;
    }

    /* get status string ID */
    switch (status) {

      case STAT_WARNING:
	str_id = GPSWarnStr;
	break;

      case STAT_OK:
	str_id = GPSOkStr;
	break;

      case STAT_NO_GPS:
	str_id = GPSLostStr;
	break;

      default:
	str_id = GPSOkStr;
    }

    /* get string from resource */
    strH = DmGetResource(strRsc, str_id);

    if (strH != NULL) {
      /* string handle valid, lock it */
      strP = MemHandleLock(strH);
    }
    
    if ( status == STAT_WARNING ) {

      switch (formID) {
      
      case GPSMainForm:
	SetFieldText(PositionStatusField, strP, false, true);
	break;

      case SkyviewForm:
	SetFieldText(SkyViewStatusField, strP, false, true);
	break;
	
      case NavigationForm:
	SetFieldText(NavigationStatusField, strP, false, true);
	break;
	
      } /* switch (formID) */
    } else if ( status == STAT_OK ){
      
      switch (formID) {
      
      case GPSMainForm:
	SetFieldText(PositionStatusField, strP, false, true);
	break;

      case SkyviewForm:
	SetFieldText(SkyViewStatusField, strP, false, true);
	break;
	
      case NavigationForm:
	SetFieldText(NavigationStatusField, strP, false, true);
	break;

      case MapForm:
	break;
	
      } /* switch (formID) */
    } else if ( status == STAT_NO_GPS ) {
      
      switch (formID) {
      
      case GPSMainForm:
	SetFieldText(PositionStatusField, strP, false, true);
	break;

      case SkyviewForm:
	SetFieldText(SkyViewStatusField, strP, false, true);
	break;
	
      case NavigationForm:
	SetFieldText(NavigationStatusField, strP, false, true);
	break;
      	
      } /* switch (formID) */
    }
  }
  
  prevID = formID;
  prevStat = status;
  
  if (strH != NULL) {
    /* unlock string handle */
    MemHandleUnlock(strH);

    /* release string resource */
    DmReleaseResource(strH);
  }
}
