/*****************************************************************************
 *
 * $RCSfile: PortForm.c,v $
 *
 * GPS4Palm Port Form
 *
 * Based on the example "GPS" from the book
 * "Palm Programming, the Developer's Guide, 2nd edition"
 * by O'Reilly.
 * (Copyright (c) 1998-2001, Neil Rhodes and Julie McKeehan  neil@pobox.com)
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-04-20 21:13:34 $
 *
 * $Revision: 1.12 $
 *
 * $Log: PortForm.c,v $
 * Revision 1.12  2005-04-20 21:13:34  mp
 * implemented opening of Bluetooth Virtual Serial Port
 * with connecting to known device (address)
 *
 * Revision 1.11  2005/02/06 14:57:59  mp
 * added GPSPortRTSCtrlCheckbox
 *
 * Revision 1.10  2005/02/05 17:40:15  mp
 * Added support for two different ports (Host and GPS),
 * storage of selected connection profile  (if Connection Manager
 * Feature Set is available) or Port Name/Number in Application Preferences.
 *
 * Revision 1.9  2004/12/18 16:12:52  mp
 * fixed serial port re-opening for PDAs with New Serial Manager
 *
 * Revision 1.8  2004/11/25 17:10:16  mp
 * fixed comment
 *
 * Revision 1.7  2004/11/25 17:02:39  mp
 * fixed file header
 *
 * Revision 1.6  2004/11/24 21:33:37  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.5  2004/11/23 18:05:15  mp
 * added deallocation of connection profile list
 *
 * Revision 1.4  2004/04/29 19:22:38  mp
 * fixed typo
 *
 * Revision 1.3  2004/04/29 19:04:40  mp
 * added RouteForm, modified for doxygen
 *
 * Revision 1.2  2004/03/01 17:18:43  mp
 * Added some comments, moved global variables from GPS.c, moved creation of list
 * from PortFormInit() to PortFormHandleEvent.
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <BtCommVdrv.h>
#include "ResourceDefines.h"
#include "PortForm.h"
#include "Utils.h"
#include "Serial.h"
#include "gpslib.h"		/* GpslibSetSerialPort() */
#include "common.h"

extern UInt16		gPortID;
extern PrefsType	gPrefs;
extern UInt32		gSerialPort;

/* Variables used by Connection Manager */
extern Char **gProfiles;
extern UInt16 gNumProfiles;

/* Static Functions */
static void PortFormInit(FormPtr form);
static void PortFormDeinit();


/*****************************************************************************
 * FUNCTION:	PortFormInit
 *
 * DESCRIPTION: Port Form init function
 *
 * PARAMETERS:	form		- Form Pointer (unused)
 *
 * RETURNED:    %
 *
 ****************************************************************************/
static 
void PortFormInit(FormPtr frmP)
{
#pragma unused(form)
  // warning-- don't do any drawing in this routine.
  // Also, don't call FrmSetFocus from here (it must be called *after*
  // FrmDrawForm)
}


/*****************************************************************************
 * FUNCTION:	PortFormDeinit
 *
 * DESCRIPTION: Port Form deinit function
 *
 *		Deallocation of memory allocated by CncGetProfileList()
 *
 * PARAMETERS:	%
 *
 * RETURNED:    %
 *
 ****************************************************************************/
static 
void PortFormDeinit()
{
  FreeCncProfileList();
}


/****************************************************************************/
/**
 * \brief	Port Form event handler.
 *
 *		Allows selection of connection profile. Available only
 *		with New Serial Manager. 
 *
 * \param	eventP	Pointer to event structure
 *
 * \return	Status flag: event handled
 ****************************************************************************/
Boolean PortFormHandleEvent(EventPtr eventP)
{
  Boolean	handled;			/* flag: event handled */
  FormPtr	frmP;				/* form ptr */
  ListPtr	list;				/* list ptr */
  Err		err;				/* error code */
  UInt16	listSelection;			/* list index */
  const UInt16	kMaxVisibleItems = 14;		/* max. visible list items */
  UInt16	numVisibleItems = kMaxVisibleItems;

  frmP = FrmGetActiveForm();
  handled = false;
  
  switch (eventP->eType) {
    case frmOpenEvent:
      PortFormInit(frmP);
      FrmDrawForm(frmP);

      /* Set Auto Power-Off Check Box from Preferences */
      FrmSetControlValue(frmP,
	FrmGetObjectIndex(frmP, GPSPortRTSCtrlCheckbox),
	gPrefs.serprefs.rts_ctrl);

      /* close current port */
      SrmClose(gPortID);

      /* create list of available connection profiles */
      err = CncGetProfileList(&gProfiles, &gNumProfiles);
      if (gNumProfiles < kMaxVisibleItems)
	numVisibleItems = gNumProfiles;
      
      /* configure GPSPortList */
      list = (ListPtr)GetObjectFromForm(frmP, GPSPortList);
      LstSetListChoices(list, gProfiles, gNumProfiles);
      LstSetHeight(list, numVisibleItems);
      
      if ((gCncManagerFtr) && (CncProfileGetIndex(gPrefs.serprefs.gps_port,
	     &listSelection) == errNone)) {
        LstSetSelection(list, listSelection);
	CtlSetLabel(GetObjectFromActiveForm(GPSPortPopupTrigger),
	  gProfiles[listSelection]);     
      } else {
        LstSetSelection(list, noListSelection);
      }
      
      /* configure HostPortList */
      list = (ListPtr)GetObjectFromForm(frmP, HostPortList);
      LstSetListChoices(list, gProfiles, gNumProfiles);
      LstSetHeight(list, numVisibleItems);
      
      if ((gCncManagerFtr) && (CncProfileGetIndex(gPrefs.serprefs.host_port,
	     &listSelection) == errNone)) {
        LstSetSelection(list, listSelection);
	CtlSetLabel(GetObjectFromActiveForm(HostPortPopupTrigger),
	  gProfiles[listSelection]);     
      } else {
        LstSetSelection(list, noListSelection);
      }
      
      handled = true;
      break;

    case popSelectEvent:
      /* update label to selected list item */
      CtlSetLabel(eventP->data.popSelect.controlP,
	gProfiles[eventP->data.popSelect.selection]);
      handled = true;
      break;

    case ctlSelectEvent:
      if (eventP->data.ctlSelect.controlID == GPSPortUsePortButton) {
	Err	err;			/* error code */
	UInt32	host_port;		/* host port */
	UInt32	baud;			/* baud rate */
	ListPtr list;			/* list prt */
	
	/* check Host port */ 
	list = (ListPtr)GetObjectFromActiveForm(HostPortList);
	
	listSelection = LstGetSelection(list);

	if (listSelection == noListSelection) {
	  /* button "OK" pressed, but no profile selected yet */
	  FrmAlert(NoProfileSelectedAlert);
	  break;
	} else {
	  /* get port ID and baud rate from selected profile */
	  err = CncGetProfileInfo(gProfiles[listSelection],
            &host_port, &baud, 0, 0, 0, 0, 0, 0);
	}

	if (err == 0) {
	  err = SrmOpen(host_port, baud, &gPortID);
	  
	  if (gCncManagerFtr) {
	    /* save Connection Mannager profile ID */
	    CncProfileGetIDFromIndex(listSelection, &gPrefs.serprefs.host_port);
	  } else {
	    /* save port name/logical port number and baud rate */
	    gPrefs.serprefs.host_port = host_port;
	    gPrefs.serprefs.host_baud = baud;
	  }
	}
	
	if (err == serErrAlreadyOpen) {
	  FrmAlert(SerialInUseAlert);
	  DoClose(gPortID);
	} else if (err) {
	  FrmAlert(CantOpenSerialAlert);
	} else {
	  /* set Serial Port to be used by gpslib */
	  GpslibSetSerialPort(host_port);
	  
	}
	
	/* close Host port */
	SrmClose(gPortID);
		
	
	/* check and open GPS port */
	list = (ListPtr)GetObjectFromActiveForm(GPSPortList);
	
	listSelection = LstGetSelection(list);

	if (listSelection == noListSelection) {
	  /* button "OK" pressed, but no profile selected yet */
	  FrmAlert(NoProfileSelectedAlert);
	  break;
	} else {
	  /* get port ID and baud rate from selected profile */
	  err = CncGetProfileInfo(gProfiles[listSelection],
            &gSerialPort, &baud, 0, 0, 0, 0, 0, 0);			
	}

	if (err == 0) {
	  if (gSerialPort == sysFileCVirtRfComm) {
	   /*
	    * Bluetooth Vitrual Serial Port --
	    * Connect to known device (address)
	    */
	    UInt16		btLibRefNum;	/* BT Library reference */
	    SrmOpenConfigType	config;		/* port's properties */
	    BtVdOpenParams	btParams;	/* BT parameters */
	    int			i;		/* loop index */
	    
	    /* get BT Library reference */
	    err = SysLibFind(btLibName, &btLibRefNum);
	    if (err) break;
	    
	    /* open Bluetooth Library */
	    err = BtLibOpen(btLibRefNum, false /* allowStackToFail */);
	    if (err) break;
	    
	    /*
	     * Discover all available devices, present them
	     * in the user interface, and allow the user to
	     * select one of these devices.
	     * Same as SrmOpen would do, but this provides us
	     * the device address which allows us to reconnect
	     * later to the same device automatically! 
	     */
	    err = BtLibDiscoverSingleDevice(
	      btLibRefNum,
	      NULL, /* *instructionTxt */
	      NULL, /* *deviceFilterList */
	      0, /* deviceFilterListLen */
	      (BtLibDeviceAddressType *)btParams.u.client.remoteDevAddr.address, /* *selectedDeviceP */
	      false, /* addressAsName */
	      false /* showLastList */
	    );
	    if (err) break;
            
	    /* close Bluetooth Library */
	    BtLibClose(btLibRefNum);
	    
	    config.baud = baud;				/* baud rate */
	    config.function = 0;			/* default */
	    config.drvrDataP = (MemPtr)&btParams;
	    config.drvrDataSize = sizeof(BtVdOpenParams);
	    btParams.role = btVdClient;			/* PDA is client */
	    btParams.u.client.method = btVdUseUuidList; /* no particular channel */
	    btParams.u.client.u.uuidList.len = 0;	/* any service class */
	    btParams.authenticate = false;		/* no authentification */
	    btParams.encrypt = false;			/* no encryption */
	    
	    /* open BT serial port */ 
	    err = SrmExtOpen(gSerialPort, &config, sizeof(config), &gPortID);
	                
	    if (!err) {
	      /*
	       * if opening succeeded, copy device address
	       * to Application Preferences
	       */
	      for (i=0; i<btLibDeviceAddressSize; i++) {
	        gPrefs.serprefs.gps_btaddr[i] =
		  btParams.u.client.remoteDevAddr.address[i];
	      }
	    }
	    
	  } else {
	    /* open non-BT serial port */
	    err = SrmOpen(gSerialPort, baud, &gPortID);
	  } /* if (gSerialPort == sysFileCVirtRfComm) */
	  
	  if (gCncManagerFtr) {
	    /* save Connection Mannager profile ID */
	    CncProfileGetIDFromIndex(listSelection, &gPrefs.serprefs.gps_port);
	  } else {
	    /* save port name/logical port number and baud rate */
	    gPrefs.serprefs.gps_port = gSerialPort;
	    gPrefs.serprefs.gps_baud = baud;
	  }
	}
	
	if (err == serErrAlreadyOpen) {
	  FrmAlert(SerialInUseAlert);
	  DoClose(gPortID);
	} else if (err) {
	  FrmAlert(CantOpenSerialAlert);
	} else {
	  /* port opened successfully */
	  err = DoSetReceiveBuffer(gPortID, gSerialBuffer, 
	    sizeof(gSerialBuffer));

	  sermux_enable(gPrefs.serprefs.rts_ctrl);

	  if (gPrefs.serprefs.rts_ctrl) {
	    sermux_select(false);
          }

	  FrmGotoForm(gPrefs.form);
	}
	handled = true;
      }
      
      if (eventP->data.ctlSelect.controlID == GPSPortRTSCtrlCheckbox) {
        gPrefs.serprefs.rts_ctrl =
	  GetControlValueFromId(GPSPortRTSCtrlCheckbox);
	
      }
      break;
    
    case frmCloseEvent:
      PortFormDeinit(FrmGetActiveForm());
      handled = false;
      break;

  } /* switch (eventP->eType) */
  return(handled);
}
