/*****************************************************************************
 *
 * $RCSfile: Serial.c,v $
 *
 * GPS4Palm Serial Manager API Wrapper Functions -
 *   either the new or the old API is used
 *
 * Based on the example "GPS" from the book
 * "Palm Programming, the Developer's Guide, 2nd edition"
 * by O'Reilly.
 * (Copyright (c) 1998-2001, Neil Rhodes and Julie McKeehan  neil@pobox.com)
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-06 13:33:41 $
 *
 * $Revision: 1.5 $
 *
 * $Log: Serial.c,v $
 * Revision 1.5  2005-05-06 13:33:41  mp
 * added BluetoothOpen()
 *
 * Revision 1.4  2005/01/30 21:26:18  mp
 * added gCncManagerFtr, gProfiles, gNumProfiles, and FreeCncProfileList()
 *
 * Revision 1.3  2004/12/18 16:12:27  mp
 * fixed serial port re-opening for PDAs with New Serial Manager
 *
 * Revision 1.2  2004/04/30 15:54:59  mp
 * modified comments for doxygen
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <SerialMgrOld.h>
#include <BtCommVdrv.h>
#include "Serial.h"

/** New Serial Manager available */
Boolean gNewSerialManager;

/** Connection Manager feature set available */ 
Boolean gCncManagerFtr;

/** Serial port ID */
UInt16	gPortID;

/** four-character port name or logical port number (New Serial Manager) */
UInt32	gSerialPort;

/** List of Connection Manager Profiles */
Char	**gProfiles;

/** Number of Connection Manager Profiles */
UInt16	gNumProfiles;

/**
 * Serial port receive buffer.
 * 900 bytes should be more than enough for one second of
 * data: 10 lines @ 80 chars per line.
 */
char   gSerialBuffer[900];


/**********************************************************************/
/**
 * Serial Manager/New Serial Manager wrapper function for data reception.
 *
 * \param	portId		port ID
 * \param	rcvBufP		ptr to receive buffer
 * \param	count		number of bytes to receive
 * \param	timeout		timeout in ticks
 * \param	errP		error code
 *
 * \return	Number of bytes received.
 **********************************************************************/
UInt32 DoReceive(UInt16 portId, void *rcvBufP, UInt32 count, Int32 timeout, Err *errP)
{
  if (gNewSerialManager)
    return SrmReceive(portId, rcvBufP, count, timeout, errP);
  else
    return SerReceive(portId, rcvBufP, count, timeout, errP);
}


/**********************************************************************/
/**
 * Serial Manager/New Serial Manager wrapper function to set receive
 *   buffer.
 *
 * \param	portId		port ID
 * \param	bufP		ptr to receive buffer
 * \param	bufSize		size of buffer in bytes
 *
 * \return	Returns 0 if successful.
 ********************************************************************/
Err DoSetReceiveBuffer(UInt16 portId, void *bufP, UInt16 bufSize)
{
  if (gNewSerialManager)
    return SrmSetReceiveBuffer(portId, bufP, bufSize);
  else
    return SerSetReceiveBuffer(portId, bufP, bufSize);
}


/**********************************************************************/
/**
 * Serial Manager/New Serial Manager wrapper function to wait for
 * at least specified number of bytes in receive buffer.
 *
 * \param	portId		port ID
 * \param	bytes		number of bytes desired
 * \param	timeout		timeout in ticks	
 *
 * \return	Returns
 *		- 0 if successful
 *		- serErrTimeOut if timeout occurred
 *		- serErrLineErr if line error ocurred
 ********************************************************************/
Err DoReceiveWait(UInt16 portId, UInt32 bytes, Int32 timeout)
{
  if (gNewSerialManager)
    return SrmReceiveWait(portId, bytes, timeout);
  else
    return SerReceiveWait(portId, bytes, timeout);
}

Err DoReceiveFlush(UInt16 portId, Int32 timeout)
{
  if (gNewSerialManager)
    return SrmReceiveFlush(portId, timeout);
  else {
    SerReceiveFlush(portId, timeout);
    return 0;
  }
}


/**********************************************************************/
/**
 * Serial Manager/New Serial Manager wrapper function to return
 * the count of bytes presently in the receive queue.
 *
 * \param	portId		port ID
 * \param	numBytesP	number of bytes in receive queue
 *
 * \return
 *		- 0 if no error
 *		- serErrLineErr if line error pending
 **********************************************************************/
Err DoReceiveCheck(UInt16 portId,  UInt32 *numBytesP)
{
  if (gNewSerialManager)
    return SrmReceiveCheck(portId, numBytesP);
  else
    return SerReceiveCheck(portId, numBytesP);
}


/**********************************************************************/
/**
 * Serial Manager/New Serial Manager wrapper function to close port.
 *
 * \param	portId		port ID
 *
 * \return
 *		- 0 if no error
 *		- serErrNotOpen if port wasn't open
 *		- serErrStillOpen if port still held open by
 *		    another process.
 **********************************************************************/
Err DoClose(UInt16 portId)
{
  if (gNewSerialManager)
    return SrmClose(portId);
  else
    return SerClose(portId);
}


/****************************************************************************/
/** Deallocation of memory allocated by CncGetProfileList().
 *
 ****************************************************************************/
void FreeCncProfileList(void)
{

  if (gProfiles) {
    UInt16 i;

    for (i = 0; i < gNumProfiles; i++)
      if (gProfiles[i])
	      MemPtrFree(gProfiles[i]);
    MemPtrFree(gProfiles);
    gProfiles =  NULL;
  }
}


/**************************************************************************/
/**
 * Open Bluetooth Virtual Serial Port and connect to known device (address)
 * 
 *
 * \param	port		port name or logical port number
 * \param	baud		baud rate
 * \param	bt_addr		Bluetooth Device Address
 * \param	PortID		port ID	returned by reference
 *
 * \return	see SrmExtOpen()
 **************************************************************************/
Err BluetoothOpen(UInt32 port, UInt32 baud, UInt8 *bt_addr, UInt16 *PortID)
{
  /*
   * Bluetooth Vitrual Serial Port --
   * Connect to known device (address)
   */
  SrmOpenConfigType	config;			/* port's properties */
  BtVdOpenParams	btParams;		/* BT parameters */
  int			i;			/* loop index */
  Err			err;			/* error code */

  config.baud = baud;	/* baud rate */
  config.function = 0;				/* default */
  config.drvrDataP = (MemPtr)&btParams;
  config.drvrDataSize = sizeof(BtVdOpenParams);
  btParams.role = btVdClient;			/* PDA is client */
  /* UUID: Universal Unique Identifier */
  btParams.u.client.method = btVdUseUuidList;	/* no particular channel */
  btParams.u.client.u.uuidList.len = 0;		/* any service class */
  btParams.authenticate = false;		/* no authentification */
  btParams.encrypt = false;			/* no encryption */

  /* Copy device address from Application Preferences */
  /* User will be queried if zero */
  for (i=0; i<btLibDeviceAddressSize; i++) {
    btParams.u.client.remoteDevAddr.address[i] = bt_addr[i];
  }

  /* open BT serial port */ 
  err = SrmExtOpen(port, &config, sizeof(config), PortID);

  return err;
}
