/*****************************************************************************
 *
 * $RCSfile: Utils.c,v $
 *
 * GPS4Palm Utility Functions
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * The field scrolling utility functions UpdateScrollbar(),
 * ScrollLines(), and PageScroll()) are from the book:
 * 
 * Neil Rhodes and Julie McKeehan:
 * "Palm Programming, the Developer's Guide, 2nd edition"
 * by O'Reilly.
 *
 *
 * $Author: mp $
 *
 * $Date: 2005-02-06 15:00:17 $
 *
 * $Revision: 1.14 $
 *
 * $Log: Utils.c,v $
 * Revision 1.14  2005-02-06 15:00:17  mp
 * modified sermux_enable() and sermux_select() to control RTS with SrmControl()
 *
 * Revision 1.13  2004/12/18 16:14:35  mp
 * added PalmOS-compliant serial port muxing with DTR pin (New Serial Manager)
 *
 * Revision 1.12  2004/11/27 10:17:43  mp
 * added Die()
 *
 * Revision 1.11  2004/11/24 21:26:23  mp
 * added UpdateScrollbar(),
 * >  * ScrollLines(), and PageScroll()
 *
 * Revision 1.10  2004/11/23 17:47:16  mp
 * removed unused variable
 *
 * Revision 1.9  2004/11/14 11:02:43  mp
 * added MyFrmGetObjectIndexFromPtr()
 *
 * Revision 1.8  2004/04/30 16:31:12  mp
 * modified/added comments for doxygen
 *
 * Revision 1.7  2004/02/28 17:46:49  mp
 * implemented Serial Mux control for Palm III
 *
 * Revision 1.6  2004/01/10 18:01:01  mp
 * added SERMUX_ENABLE and comments
 *
 * Revision 1.5  2003/11/16 17:42:49  mp
 * removed debug messages
 *
 * Revision 1.4  2003/11/02 11:13:55  mp
 * added backlight(), sermux_select() and sermux_enable()
 *
 * Revision 1.3  2003/10/20 17:31:17  mp
 * added SetFieldText(), fixed comment for SetFieldTextFromStr()
 *
 * Revision 1.2  2003/10/15 19:01:25  mp
 * added function nibble2hex
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>			
#include "ResourceDefines.h"

/* Global Variables */
/* Serial port ID */
extern UInt16	gPortID;

/* New Serial Manager available */
extern Boolean	gNewSerialManager;

/*
 * Backlight Control
 *
 * Note: This is strictly hardware specific and will most likely work on
 *       Handspring Visor Deluxe only!!!
 *
 * Backlight is controlled by PGDATA[0]
 *   PGDATA[0] = 0 - Backlight On
 *   PGDATA[1] = 1 - Backlight Off 
 */ 
#define PGDATA	((volatile UInt8 *)0xFFFFF431L)
#define BLDATA	PGDATA			/* I/O Data */
#define BLMASK	((UInt8)  0x01)

#define BACKLIGHT_ON	(*BLDATA &= ~BLMASK)
#define BACKLIGHT_OFF	(*BLDATA |=  BLMASK)

/*
 * Serial Mux Control
 *
 * Note: This is strictly hardware specific and will most likely work on
 *       the devices explicitely specified only!!!
 *       For any other PDA, define SERMUX_* only if you know what you are
 *       doing! Otherwise the PDA might not work properly or even be damaged!
 *
 * Handspring Visor (Deluxe):
 * --------------------------
 * External Serial Port Multiplexer is controlled by either
 *   PE[7] - Cradle Connector Keyboard Input (/KBD) (R18: 330R)
 *   or
 *   PD[4] - Cradle Connector Hotsync Input (/HS) (R19: 100R)
 * To use either of these pins as output, the port pull-up has to be disabled
 * and the port direction has to be changed. The Visor's Cradle Connector pins
 * are both connected to a series resistor (R18/R19) on the button board.
 *
 * Palm III:
 * ---------
 * External Serial Port Multiplexer is controlled by
 * PE[6] - /RTS
 */

#if defined(SERMUX_VISOR_HS) | \
    defined(SERMUX_VISOR_KBD) | \
    defined(SERMUX_PALMIII_RTS)

#define PEDIR	((volatile UInt8 *)0xFFFFF420L)
#define PEDATA	((volatile UInt8 *)0xFFFFF421L)
#define PEPUEN	((volatile UInt8 *)0xFFFFF422L)
#define PESEL	((volatile UInt8 *)0xFFFFF423L)

#define PDDIR	((volatile UInt8 *)0xFFFFF418L)
#define PDDATA	((volatile UInt8 *)0xFFFFF419L)
#define PDPUEN	((volatile UInt8 *)0xFFFFF41AL)
#define PDSEL	((volatile UInt8 *)0xFFFFF41BL)
#endif

#if defined(SERMUX_VISOR_KBD) | defined(SERMUX_PALMIII_RTS)
  #define MUXDIR PEDIR		/* I/O Direction: 0 - In / 1 - Out */
#endif
#ifdef SERMUX_VISOR_HS
  #define MUXDIR PDDIR		/* I/O Direction: 0 - In / 1 - Out */
#endif

#define MUXDATA	(MUXDIR+1)	/* I/O Data */
#define MUXPUEN	(MUXDIR+2)	/* I/O Pull-up: 0 - Disabled / 1 - Enabled */
#define MUXSEL	(MUXDIR+3)	/* I/O Select: 0 - Peripheral / 1 - GPIO */

#ifdef SERMUX_PALMIII_RTS
  #define MUXMASK	((UInt8) 0x40)	/* PE[6] = /RTS */ 
#endif
#ifdef SERMUX_VISOR_KBD
  #define MUXMASK	((UInt8) 0x80)	/* PE[7] = /KBD */
#endif
#ifdef SERMUX_VISOR_HS
  #define MUXMASK	((UInt8) 0x10)	/* PD[4] = /HS */
#endif

/**********************************************************************/
/**
 * \brief	Get pointer to object of specified form.
 *
 * \param	form		form ptr
 * \param	objectID	object ID
 *
 * \return	ptr to object
 **********************************************************************/
void *GetObjectFromForm(FormPtr form, UInt16 objectID)
{
	return FrmGetObjectPtr(form, 
		FrmGetObjectIndex(form, objectID));
}

/**********************************************************************/
/**
 * \brief	Get pointer to object of active form.
 *
 * \param	objectID	object ID
 *
 * \return	ptr to object
 **********************************************************************/
void *GetObjectFromActiveForm(UInt16 objectID)
{
	return GetObjectFromForm(FrmGetActiveForm(), objectID);
}


/**********************************************************************/
/**
 * \brief	Get ID of object of active form given an object pointer.
 *
 * \param	frmP		form ptr
 * \param	objP		object ptr	
 *
 * \return	object ID
 *
 * \note	This function implements FrmGetObjectIndexFromPtr()
 *		on PalmOS versions < 4.0
 **********************************************************************/
UInt16 MyFrmGetObjectIndexFromPtr(const FormPtr frmP, void *objP)
{
  UInt16 index;
  UInt16 objIndex = frmInvalidObjectId;
  UInt16 numObjects = FrmGetNumberOfObjects(frmP);

  for (index = 0; index < numObjects; index++) {
    if (FrmGetObjectPtr(frmP, index) == objP) {
      /* Found it */
      objIndex = index;
      break;
    }
  }

  return objIndex;
}


/****************************************************************************/
/**
 * \brief	generate SysFatalAlert and stop application
 *
 * \param	msg		pointer to error message string
 *
 *****************************************************************************/
void Die(const Char *msg)
{
  EventType	event;			/* event structure */

  event.eType = appStopEvent;
  SysFatalAlert(msg);
  EvtAddEventToQueue(&event);
}


/****************************************************************************/
/**
 * \brief	Table paging function using scrollbar.
 *
 * \param	tbl_id		table ID
 * \param	down		paging direction (up/down)
 * \param	nr_items	number of items
 * \param	top_row		ptr to variable containing top row
 *
 *****************************************************************************/
void Dopage(UInt16 tbl_id, Boolean down, UInt16 nr_items,
  Int32 *top_row)
{
  TablePtr	tblP;		/* table ptr */
  FormPtr	frmP;		/* form ptr */
  UInt32	n;		/* number of table rows */

  frmP = FrmGetActiveForm();
  tblP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, tbl_id));
  n = TblGetNumberOfRows(tblP);
  *top_row += (down == true) ? n : -n;
  if ((*top_row + n) >= nr_items) {
    *top_row = nr_items - n;
  }
  if (*top_row < 0) {
    *top_row = 0;
  }
}


/****************************************************************************/
/**
 * \brief	Table scrolling function (using scrollbar).
 *
 * \param	tbl_id		table ID
 * \param	scl_id		scrollbar ID
 * \param	nr_items	number of items
 * \param	top_row		ptr to variable containing top row
 *
 *****************************************************************************/
void Doscroll(UInt16 tbl_id, UInt16 scl_id, UInt16 nr_items, Int32 *top_row)
{
  FormPtr	frmP;		/* form ptr */
  TablePtr	tblP;		/* table ptr */ 
  Int32		n;		/* number of table rows */
  ScrollBarPtr	sclP;		/* scrollbar ptr */
  Int16		v;		/* current scrollbar value */
  Int16		mn;		/* min scrollbar value (top) */
  Int16		mx;		/* max scrollbar value (bottom) */
  Int16		ps;		/* scrollbar page size */

  frmP = FrmGetActiveForm();
  sclP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, scl_id));
  SclGetScrollBar(sclP, &v, &mn, &mx, &ps);
  tblP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, tbl_id));
  n = TblGetNumberOfRows(tblP);
  *top_row = v;
  if ((*top_row + n) >= nr_items) {
    *top_row = nr_items - n;
  }
  if (*top_row < 0) {
    *top_row = 0;
  }
}


/****************************************************************************/
/**
 * \brief	Field scrolling function (using scrollbar).
 *
 *		Update scrollbar according to field's values.
 *
 * \param	form		form ptr
 * \param	fieldID		field ID
 * \param	scrollbarID	scrollbar ID
 *
 *****************************************************************************/
void UpdateScrollbar(FormPtr form, UInt16 fieldID, UInt16 scrollbarID)
{
   ScrollBarPtr   scroll;
   FieldPtr       field;
   UInt16         currentPosition;
   UInt16         textHeight;
   UInt16         fieldHeight;
   UInt16         maxValue;
   
   field = (FieldPtr) FrmGetObjectPtr(form, 
     FrmGetObjectIndex(form, fieldID));
   FldGetScrollValues(field, &currentPosition, &textHeight, &fieldHeight);

   // if the field is 3 lines, and the text height is 4 lines
   // then we can scroll so that the first line is at the top 
   // (scroll position 0) or so the second line is at the top
   // (scroll postion 1). These two values are enough to see
   // the entire text.
   if (textHeight > fieldHeight)
      maxValue = textHeight - fieldHeight;
   else if (currentPosition)
      maxValue = currentPosition;
   else
      maxValue = 0;
      
   scroll = (ScrollBarPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, scrollbarID));
   
   // on a page scroll, want to overlap by one line (to provide context)
   SclSetScrollBar(scroll, (Int16) currentPosition, 0, 
     (Int16) maxValue, (Int16) fieldHeight - 1);
}


/****************************************************************************/
/**
 * \brief	Field scrolling function (using scrollbar).
 *
 *		Scroll field's contents by specified number of lines.
 *
 * \param	form			form ptr
 * \param	fieldID			field ID
 * \param	scrollbarID		scrollbar ID
 * \param	numLinesToScroll	number of lines to scroll
 *					(negative means up)
 * \param	redraw			flag: redraw field if true
 *
 *****************************************************************************/
void ScrollLines(FormPtr form, UInt16 fieldID, UInt16 scrollbarID,
  Int16 numLinesToScroll, Boolean redraw)
{
   FormPtr        frm = FrmGetActiveForm();
   FieldPtr       field;
   
   field = (FieldPtr) FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
   if (numLinesToScroll < 0)
      FldScrollField(field, (UInt16) -numLinesToScroll, winUp);
   else
      FldScrollField(field, (UInt16) numLinesToScroll, winDown);
      
   // if there are blank lines at the end and we scroll up, FldScrollField
   // makes the blank lines disappear. Therefore, we've got to update
   // the scrollbar
   if ((FldGetNumberOfBlankLines(field) && numLinesToScroll < 0) ||
      redraw)
      UpdateScrollbar(form, fieldID, scrollbarID);
}


/****************************************************************************/
/**
 * \brief	Field scrolling function (using scrollbar).
 *
 *		Scroll field's contents by entire page size.
 *
 * \param	form			form ptr
 * \param	fieldID			field ID
 * \param	scrollbarID		scrollbar ID
 * \param	direction		winUp/winDown
 *
 *****************************************************************************/
void PageScroll(FormPtr form, UInt16 fieldID, UInt16 scrollbarID,
  WinDirectionType direction)
{
   FieldPtr       field;
   
   field = (FieldPtr) FrmGetObjectPtr(form, 
     FrmGetObjectIndex(form, fieldID));
   if (FldScrollable(field, direction)) {
      Int16 linesToScroll = (Int16) FldGetVisibleLines(field) - 1;
      
      if (direction == winUp)
         linesToScroll = -linesToScroll;
      ScrollLines(form, fieldID, scrollbarID, linesToScroll, true);
   }
}


/**********************************************************************/
/**
 * \brief	Get PalmOS ROM version.
 *
 * \return	Encoded ROM version.
 **********************************************************************/
UInt32 GetRomVersion()
{
	UInt32 romVersion;
	
	// The system records the version number in a feature.  A feature is a
	// piece of information which can be looked up by a creator and feature
	// number.
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	return romVersion;
}


/**********************************************************************/
/**
 *
 * \brief	Set field object's text handle.
 *
 *		Will reuse an existing text handle, if any.
 *
 * \param	field	field object pointer
 * \param	s	source text pointer
 * \param	redraw	if true, field will be redrawn
 *
 * \return	Zero if successful, memory manager error code if
 *		allocation/resizing of memory chunk failed.
 *
 ***********************************************************************/
Err SetFieldTextFromStr(FieldPtr field, Char *s, Boolean redraw)
{
   MemHandle      h;
   
   h = FldGetTextHandle(field);
   if (h) {
     Err	err;
     
     FldSetTextHandle(field, NULL);
     err = MemHandleResize(h, StrLen(s) + 1);
     if (err) {
       FldSetTextHandle(field, h);  // restore handle
       return err;
     }
   } else {
     h = MemHandleNew(StrLen(s) + 1);
     if (!h)
       return memErrNotEnoughSpace;
   }
   // at this point, we have a handle of the correct size

   // copy the string to the locked handle.
   StrCopy((Char *) MemHandleLock(h), s);
   // unlock the string handle.
   MemHandleUnlock(h);
   
   FldSetTextHandle(field, h);
   if (redraw)
   	 FldDrawField(field);
   return 0;
} 


/**********************************************************************/
/**
 *
 * \brief	Set field object's text handle.
 *
 *		Will reuse an existing text handle, if any.
 * \note	Set srcP to NULL to free memory allocated for text.
 *
 * \param	fldID	field object id
 * \param	srcP	source text pointer
 * \param	append  true: append / false: replace
 * \param	redraw	if true, field will be redrawn
 *
 ***********************************************************************/
void SetFieldText(UInt16 fldID, Char *srcP, Boolean append, Boolean redraw)
{
  MemHandle 		stringH;
  UInt16		oldLength = 0;
  UInt16		newSize;
  Char 			*stringP;
  FieldPtr		fldP;
  FormPtr		frmP;

  frmP = FrmGetActiveForm();
  fldP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP,fldID));

  if (!srcP || !*srcP)
	{				// handle clearing field as well
	  FldFreeMemory(fldP);
	  if ( redraw )
		{
		  FldEraseField( fldP );
		  FldDrawField( fldP );
		}
	  return;
	}


  newSize = StrLen(srcP) + 1;		// initialize new size
  stringH = FldGetTextHandle( fldP );	// get the current text handle
  FldSetTextHandle( fldP, 0 );		// release this handle from field

  // Resize the existing handle, if any
  if ( stringH )
	{
	  if ( append )
		{
		  stringP = MemHandleLock( (MemHandle)stringH );
		  oldLength = StrLen(stringP);
		  newSize += oldLength;
		  MemPtrUnlock( stringP );
		}
	  if ( MemHandleResize ( (MemHandle)stringH, newSize) )
		  goto Exit;
	} // Resize the existing handle, if any

  // Otherwise, allocate a new handle
  else
	{
	  stringH = MemHandleNew( newSize );	// allocate a new chunk
	  if ( !stringH )	return;
	}

  // Append the new text
  stringP = MemHandleLock( (MemHandle)stringH );
  StrCopy( stringP + oldLength, srcP );			// copy the new text
  MemPtrUnlock( stringP );

Exit:
  FldSetTextHandle( fldP, stringH );		// set the new text handle
  if ( redraw )
	{
	  FldEraseField( fldP );
	  FldDrawField( fldP );
	}
}


/**********************************************************************/
/**
 * \brief	Check if appropriate ROM version is available.
 *
 * \param	requiredVersion		required version
 * \param	launchFlags		application launch flags
 *
 * \return	0 on success, sysErrRomIncompatible on failure
 *
 ***********************************************************************/
Err RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags)
{
	UInt32 romVersion = GetRomVersion();
	
	// See if we're on in minimum required version of the ROM or later.
	if (romVersion < requiredVersion)
		{
		// If the user launched the app from the launcher, explain
		// why the app shouldn't run.  If the app was contacted for 
		// something else, like it was asked to find a string by the  
		// system find, then don't bother the user with a warning dialog.   
		// These flags tell how the app was launched to decided if a 
		// warning should be displayed.
		if ((launchFlags & 
			(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) 
			== (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
			FrmAlert(RomIncompatibleAlert);
		
			// Pilot 1.0 will continuously relaunch this app unless we switch  
			// to another safe one.  The sysFileCDefaultApp is 
			// considered "safe".
			if (sysGetROMVerMajor(romVersion) < 2) {
				AppLaunchWithCommand(sysFileCDefaultApp,
					sysAppLaunchCmdNormalLaunch, NULL);
			}
		}
		return sysErrRomIncompatible;
	}
	return 0;
}


/**********************************************************************/
/**
 * \brief	Convert nibble value to hex digit (upper case).
 *
 *              This function can be used instead of the PalmOS
 *              functions StrAToH (which always returns a 32 bit
 *              hex string) and StrPrintF (which ignores the field
 *              width specified in the format string). 
 *
 * \param	nibble	nibble value
 *
 * \return    0..9, A..F if nibble is valid, else '\0'.
 *
 ***********************************************************************/
char nibble2hex(const char nibble)
{
  
  if (nibble < 10) {
    return ('0' + nibble);
  } else if (nibble < 16) {
    return ('A' + nibble - 10);
  } else {
    return '\0';
  }
}


/**********************************************************************/
/**
 * \brief	Switch backlight on/off.
 * 
 * \note	The function directly accesses hardware registers,
 *		  this will work on specific devices only. 
 *
 * \param	on	true: on / false: off
 *
 ***********************************************************************/
void backlight(Boolean on)
{
  if (on) {
    BACKLIGHT_ON;
  } else {
    BACKLIGHT_OFF;
  }

}


/**********************************************************************/
/**
 * \brief	Enable serial multiplexer select output.
 * \note	If any of the SERMUX_* flags is defined, the
 *		  function directly accesses hardware registers,
 *		  this will work on specific devices only.
 *		If the new serial manager is available (and the PDA
 *		  hardware supports it), the RTS pin is used instead.		  
 *
 * \param	enable	true: enable / false: disable
 *
 ***********************************************************************/
void sermux_enable(Boolean enable)
{

#if defined(SERMUX_VISOR_HS) | \
    defined(SERMUX_VISOR_KBD) | \
    defined(SERMUX_PALMIII_RTS)
  if (enable) {
    
    /* disable pull-up */
    *MUXPUEN &= ~MUXMASK;
    
    /* port bit is output */
    *MUXDIR  |= MUXMASK;
  
    /* I/O function */
    *MUXSEL  |= MUXMASK;
  } else {
  
    /* port bit is input */ 
    *MUXDIR  &= ~MUXMASK;
    
    /* enable pull-up */
    *MUXPUEN |= MUXMASK;
    
    /* I/O function */
    *MUXSEL  |= MUXMASK;
  }
#else
  if (gNewSerialManager) {
    UInt32		flags;		/* port control flags */
    UInt16		valueLen;	/* value length */
    Err			err;		/* error code */    
    
    err = SrmControl(gPortID, srmCtlGetFlags, &flags, &valueLen);
    
    if (enable) {
      /* disable RTS Flow Control */
      flags &= ~srmSettingsFlagRTSAutoM;
    
    } else {
      /* enable RTS Flow Control */
      flags |= srmSettingsFlagRTSAutoM;
      
    }    
    
    valueLen = sizeof(UInt32);
    err = SrmControl(gPortID, srmCtlSetFlags, &flags, &valueLen);
  } /* if (gNewSerialManager) */
#endif

}


/**********************************************************************/
/**
 * \brief	Set serial multiplexer select output.
 * \note	If any of the SERMUX_* flags is defined, the
 *		  function directly accesses hardware registers,
 *		  this will work on specific devices only.
 *		If the new serial manager is available (and the PDA
 *		  hardware supports it), the RTS pin is used instead.		  
 *
 * \param	sel	true: set output high / false: set output low
 *
 ***********************************************************************/
void sermux_select(Boolean sel)
{
  
#if defined(SERMUX_VISOR_HS) | \
    defined(SERMUX_VISOR_KBD) | \
    defined(SERMUX_PALMIII_RTS)

  if (sel) {
    *MUXDATA |= MUXMASK;
  } else {
    *MUXDATA &= ~MUXMASK;
  }

#else
  
  if (gNewSerialManager) {
    UInt32		flags;		/* port control flags */
    UInt16		valueLen;	/* value length */
    Err			err;		/* error code */    
    
    err = SrmControl(gPortID, srmCtlGetFlags, &flags, &valueLen);
    
    if (sel) {
      /* disable RTS override */
      flags &= ~srmSettingsFlagRTSInactive;
    
    } else {
      /* set RTS inactive (low) */
      flags |= srmSettingsFlagRTSInactive;
    }

    valueLen = sizeof(UInt32);
    err = SrmControl(gPortID, srmCtlSetFlags, &flags, &valueLen);
      
  } /* if (gNewSerialManager) */

#endif
}
