/*****************************************************************************
 *
 * $RCSfile: WaypointForm.c,v $
 *
 * GPS4Palm Waypoint Form
 *
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-06 13:39:42 $
 *
 * $Revision: 1.32 $
 *
 * $Log: WaypointForm.c,v $
 * Revision 1.32  2005-05-06 13:39:42  mp
 * added workaround for Bluetooth-Open Bug
 *
 * Revision 1.31  2005/01/30 21:22:13  mp
 * added PortsMenu event handling
 *
 * Revision 1.30  2005/01/23 10:01:32  mp
 * changed re-opening of GPS serial port after data transfer
 *
 * Revision 1.29  2004/12/18 16:12:04  mp
 * fixed serial port re-opening for PDAs with New Serial Manager
 *
 * Revision 1.28  2004/12/09 17:27:24  mp
 * replaced waypoint display mode strings by WaypointDispList object
 *
 * Revision 1.27  2004/12/02 19:02:47  mp
 * improved string alignment for compatibility
 *
 * Revision 1.26  2004/11/30 20:41:33  mp
 * added TransferTrkReqEvent and TrackMenu event handling, added GarminGetTrks()
 *
 * Revision 1.25  2004/11/26 20:11:03  mp
 * modified TableDrawPosnCell()
 *
 * Revision 1.24  2004/11/24 21:30:07  mp
 * moved static function declarations from header to implementation file,
 * added AboutForm menu event handling
 *
 * Revision 1.23  2004/11/23 17:46:55  mp
 * removed unused variables
 *
 * Revision 1.22  2004/11/20 16:54:01  mp
 * added disabling of active route when selecting active waypoint
 *
 * Revision 1.21  2004/11/18 17:35:07  mp
 * added QueryRouteWaypoint() and modified WaypointDeleteButton event handling:
 * if a waypoint is part of any route, deletion is denied
 *
 * Revision 1.20  2004/11/14 22:44:14  mp
 * code cleanup
 *
 * Revision 1.19  2004/11/13 18:52:20  mp
 * added freeing of WaypointGpslibField memory, fixed rec index data type
 *
 * Revision 1.18  2004/11/11 18:52:04  mp
 * fixed table update after waypoint download (PC -> PDA)
 *
 * Revision 1.17  2004/06/23 18:06:31  mp
 * Modified comments, modified table scroll/page functions, updated DataExchange functions.
 *
 * Revision 1.16  2004/04/23 16:05:05  mp
 * Added error handling for serial port.
 *
 * Revision 1.15  2004/04/22 19:48:00  mp
 * Changed handling of waypoint reception if a record with same waypoint ID
 * already exists: existing waypoint is updated, no new record is created.
 *
 * Revision 1.14  2004/04/21 18:37:23  mp
 * fixed handling of empty waypoint DB, fixed transmission of comments
 *
 * Revision 1.13  2004/04/18 18:19:14  mp
 * added waypoint upload/download
 *
 * Revision 1.12  2004/03/28 18:56:36  mp
 * modified waypoint displaying
 *
 * Revision 1.11  2004/03/28 17:45:51  mp
 * added waypoint symbol bitmap display
 *
 * Revision 1.10  2004/03/20 13:34:44  mp
 * fixed waypoint table selection, changed waypoint data type to double
 *
 * Revision 1.9  2004/03/15 17:19:23  mp
 * added Waypoint Copy function
 *
 * Revision 1.8  2004/03/13 12:39:47  mp
 * moved call of Set_UTM_Parameters() to AppStart()
 *
 * Revision 1.7  2004/03/10 22:00:50  mp
 * added active waypoint check, modified comments
 *
 * Revision 1.6  2004/03/10 17:26:59  mp
 * Modified due to changes in Data.h. Added handling of WaypointGotoButton.
 *
 * Revision 1.5  2004/03/09 19:17:58  mp
 * added WaypointDispField handling
 *
 * Revision 1.4  2004/03/07 10:39:03  mp
 * Moved database open/close to AppStart()/AppStop() (GPS.c)
 *
 * Revision 1.3  2004/03/03 19:02:59  mp
 * added WaypointNewButton handling
 *
 * Revision 1.2  2004/03/01 19:58:46  mp
 * Fixed waypoint deletion.
 *
 * Revision 1.1  2004/03/01 19:00:36  mp
 * initial version, deleting not working properly
 *
 *
 ****************************************************************************/

#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include <SerialMgrOld.h>
#include "ResourceDefines.h"
#include "gpslib.h"
#include "WaypointForm.h"
#include "Data.h"
#include "DataExchange.h"
#include "Utils.h"
#include "Serial.h"
#include "utm.h"
#include "geo.h"
#include "fp.h"
#include "common.h"

/* Waypoint Symbol Coordinates */
#define WPT_X	146
#define WPT_Y	94

/* GPS Data */
/*   defined and modified in HandleMessage() (HandleMessage.c) */
extern GPSType gGPSData;

/* Global Variables */
extern PrefsType	gPrefs;			/* application preferences */
ActWayptType		gSelWaypoint;		/**< selected waypoint */
extern UInt32		gSerialPort;		/* serial port name/no. */


/* Local Variables */
static Int32		gWayptTopRow;		/* top table row */
static UInt16		gWayptSel;		/* selected waypoint (table) */


/* Static Functions */
static void WaypointFormInit(FormPtr frmP)		WAYPOINT_SECTION;
static void WaypointFormDeinit(FormPtr frmP)		WAYPOINT_SECTION;
static Boolean QueryRouteWaypoint(UInt16 wpt_index,
  G_byte *nmbr)						WAYPOINT_SECTION;
static void TableDrawIdentCell(void *tblP, Int16 row,
  Int16 column, RectangleType *bounds)			WAYPOINT_SECTION;
static void TableDrawPosnCell(void *tblP, Int16 row,
  Int16 column, RectangleType *bounds)			WAYPOINT_SECTION;
static void ChangePorts(void)				WAYPOINT_SECTION;
		   
/*****************************************************************************
 * FUNCTION:	WayptEditFormInit
 *
 * DESCRIPTION: Waypoint Form Init Function (not used)
 *
 ****************************************************************************/
static void WaypointFormInit(FormPtr frmP)
{
#pragma unused(frmP)
  // warning-- don't do any drawing in this routine.
  // Also, don't call FrmSetFocus from here (it must be called *after*
  // FrmDrawForm)
}


/*****************************************************************************
 * FUNCTION:	WayptEditFormInit
 *
 * DESCRIPTION: Waypoint Form De-Init Function (not used)
 *
 ****************************************************************************/
static void WaypointFormDeinit(FormPtr frmP)
{
#pragma unused(frmP)
}


/****************************************************************************/
/**
 * \brief	Check if waypoint is part of any route
 *
 * 
 * \param	wpt_index	waypoint record index
 * \param	nmbr		route number returned by reference,
 *				  undefined if waypoint not found
 *
 * \return
 *		- true if waypoint found in route
 *		- false if waypoint not found in route
 *****************************************************************************/
static
Boolean QueryRouteWaypoint(UInt16 wpt_index, G_byte *nmbr)
{
  MemHandle	routeDBEntry;	/* route record handle */
  route_t	*route;		/* route record ptr */
  UInt16	rte_index;	/* route record index */
  UInt16	rte;		/* route loop index */
  UInt16	wpt;		/* waypoint loop index */
  UInt32	wpt_rec_id;	/* waypoint record ID */
  UInt16	nr_routes;	/* number of routes */
  Boolean	found = false;	/* flag: waypoint found */
  
  /* get waypoint record ID */
  DmRecordInfo(gWaypointDB, wpt_index, NULL /* *attrP */,
	  &wpt_rec_id /* *uniqueIDP */, NULL /* *chunkIDP */);
  
   /* get number of routes */
  nr_routes = DmNumRecordsInCategory(gRouteDB, dmAllCategories);
 
  for (rte=0; rte < nr_routes; rte++) {
    /* start seeking from begin */
    rte_index = 0;
    
    /* get next route record */
    DmSeekRecordInCategory(gRouteDB, &rte_index /* index */,
      rte /* offset */, dmSeekForward, dmAllCategories);

    /* get record handle */
    routeDBEntry = DmQueryRecord(gRouteDB, rte_index);

    /* lock record */
    route = MemHandleLock(routeDBEntry);

    for (wpt=0; wpt < route->items; wpt++) {
      if (route->wpt_rec_id[wpt] == wpt_rec_id) {
	/* found the waypoint! */
	found = true;
	
	/* return route number */
	*nmbr = route->nmbr;
	break;
      }
    }
    
    /* unlock record */
    MemHandleUnlock(routeDBEntry);
  
    if (found == true)
      break;
  }
    
  return found;
}


/* Table Drawing Callback Functions */
TableDrawItemFuncType TableDrawIdentCell;
TableDrawItemFuncType TableDrawPosnCell;


/*****************************************************************************
 * FUNCTION:	TableDrawIdentCell
 *
 * DESCRIPTION:	Custom table cell drawing function for Waypoint Identifier
 *		(string, max. 6 characters). The DB record number is
 *              retrieved via TblGetRowData().
 *
 * GLOBALS:	gWaypointDB	-- waypoint DB reference
 *
 * PARAMETERS:	Note: The function's prototype is defined by PalmOS
 *		      (TableDrawItemFuncType).
 *
 * RETURNED:	-
 ****************************************************************************/
static void TableDrawIdentCell(void *tblP, Int16 row, Int16 column,
                   RectangleType *bounds)
{
  Char			tmp[8];			/* string buffer */
  MemHandle		waypointDBEntry;	/* record handle */
  UInt16		rec;			/* record index */
  waypoint_t		waypoint;		/* unpacked waypoint */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */

  /* get record index from row data */
  rec = (UInt16)TblGetRowData(tblP, row);

  if (column == 0) {
    WinEraseRectangle(bounds, 0);
    
    if (TblRowUsable(tblP, row)) {
      waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
      packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
      UnpackWaypoint(&waypoint, packed_waypoint);
      MemSet((Char *)tmp, 7, 0);
      StrNCopy(tmp, (Char *)&waypoint.ident, 6);
      MemHandleUnlock(waypointDBEntry);

      WinDrawChars(tmp, StrLen(tmp), bounds->topLeft.x + 1, bounds->topLeft.y);
    }
  }
}


/*****************************************************************************
 * FUNCTION:	TableDrawPosnCell
 *
 * DESCRIPTION:	Custom table cell drawing function for Waypoint Position
 *		Column 1 - Latitude or UTM Zone and Easting
 *		Column 2 - Longitude or UTM Northing
 *		The DB record number is retrieved via TblGetRowData().
 *		The display format is determined by gPrefs.units.pos_unit
 *		(Application Preferences). The table width is not sufficient
 *		for the DMS format.
 *
 * GLOBALS:	gWaypointDB	-- waypoint DB reference
 *
 * PARAMETERS:	Note: The function's prototype is defined by PalmOS
 *		      (TableDrawItemFuncType).
 *
 * RETURNED:	-
 ****************************************************************************/
static void TableDrawPosnCell(void *tblP, Int16 row, Int16 column,
                   RectangleType *bounds)
{
  Char			lat_str[20];		/* latitude string buffer */
  Char			lon_str[20];		/* longitude string buffer */
  MemHandle		waypointDBEntry;	/* record handle */
  UInt16		rec;			/* record index */
  waypoint_t		waypoint;		/* unpacked waypoint */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
  double		lat;			/* latitude */
  double		lon;			/* longitude */
  Int32			err;			/* error code */
  Int32			zone;			/* UTM Zone */
  char			hemisphere;		/* 'N' / 'S' */
  double		easting;		/* Easting in m */
  double		northing;		/* Northing in m */
  Int16			cell_width;		/* table cell width in pixels */
  Int16			length;			/* string length */
  Int16			offset = 0;		/* string alignment offset */
  Boolean		fits;			/* flag: string fits in cell */

  /* get record index from row data */
  rec = (UInt16)TblGetRowData(tblP, row);
  
  if (column == 1) {
    /* erase table cell */
    WinEraseRectangle(bounds, 0);
    
    /* get record */
    waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
    
    /* unpack record */
    packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
    UnpackWaypoint(&waypoint, packed_waypoint);
    
    /* convert semicircles to floating point */
    lat = semi2deg(waypoint.posn.lat);
    lon = semi2deg(waypoint.posn.lon);

    /* unlock record */
    MemHandleUnlock(waypointDBEntry);

    /*
     * Convert latitude from floating pont to string
     * (depending on selected position unit from preferences)
     */
    switch (gPrefs.units.pos_unit) {
      case POS_D:
        /* align numbers */
	offset = (lat < 0) ? 0 : FntCharWidth('-');

	deg_to_d_str(lat, lon, lat_str, lon_str);
	break;
	
      case POS_DM:
        /* align numbers */
        offset = (lat < 0) ? 0 : FntCharWidth('-');
	
	deg_to_dm_str(lat, lon, lat_str, lon_str);
	break;
	
      case POS_DMS:
        /* align numbers */
	offset = (lat < 0) ? 0 : FntCharWidth('-');
	
	deg_to_dms_str(lat, lon, lat_str, lon_str);
	break;
      
      case POS_UTM:
        err = Convert_Geodetic_To_UTM (
      			(double)(lat/180.0 * M_PI) /* Latitude (rad) */,
                        (double)(lon/180.0 * M_PI) /* Longitude (rad) */,
                        &zone /* long  *Zone */,
                        &hemisphere /* Hemisphere */,
                        &easting /* *Easting */,
                        &northing /* *Northing */);

        if (!err) {
          /* print zone and belt to string */
	  StrPrintF(lat_str, "%ld%c ", zone, Lat_To_UTM_Belt(lat));
	
	  /* print easting to string */
	  format_number(easting, 1, &lat_str[StrLen(lat_str)]);
        }
	break;
    }

    /* table cell width in pixels */
    cell_width = bounds->extent.x - offset;

    /* truncate string to fit in cell */
    length = StrLen(lat_str);
    FntCharsInWidth(lat_str, &cell_width, &length, &fits);
    
    /* print to table cell */
    WinDrawChars(lat_str, length,
      bounds->topLeft.x + offset + 1, bounds->topLeft.y);
  }
  
  if (column == 2) {
    /* erase table cell */
    WinEraseRectangle(bounds, 0);
    
    /* get record */
    waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
    
    /* unpack record */
    packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
    UnpackWaypoint(&waypoint, packed_waypoint);
    
    /* convert semicircles to floating point */
    lat = semi2deg(waypoint.posn.lat);
    lon = semi2deg(waypoint.posn.lon);

    /* unlock record */
    MemHandleUnlock(waypointDBEntry);

    /*
     * Convert longitude from floating pont to string
     * (depending on selected position unit from preferences)
     */
    switch (gPrefs.units.pos_unit) {
      case POS_D:
        /* align numbers */
	offset = (lon < 0) ? 0 : FntCharWidth('-');

	deg_to_d_str(lat, lon, lat_str, lon_str);
	break;
	
      case POS_DM:
        /* align numbers */
	offset = (lon < 0) ? 0 : FntCharWidth('-');
	
	deg_to_dm_str(lat, lon, lat_str, lon_str);
	break;
	
      case POS_DMS:
        /* align numbers */
	offset = (lon < 0) ? 0 : FntCharWidth('-');
	
	deg_to_dms_str(lat, lon, lat_str, lon_str);
	break;
      
      case POS_UTM:
        err = Convert_Geodetic_To_UTM (
      			deg2rad(lat) /* Latitude (rad) */,
                        deg2rad(lon) /* Longitude (rad) */,
                        &zone /* long  *Zone */,
                        &hemisphere /* Hemisphere */,
                        &easting /* *Easting */,
                        &northing /* *Northing */);
	
	/* print northing to string */
	if (!err) {
	  StrCopy(lon_str, "  ");
	  format_number(northing, 1, &lon_str[2]);
        }
	break;
    }
    
    /* table cell width in pixels */
    cell_width = bounds->extent.x - offset;

    /* truncate string to fit in cell */
    length = StrLen(lon_str);
    FntCharsInWidth(lon_str, &cell_width, &length, &fits);
    
    /* print to table cell */
    WinDrawChars(lon_str, length,
      bounds->topLeft.x + offset + 1, bounds->topLeft.y);
  }
}


/****************************************************************************/
/**
 * \brief	Waypoint table drawing function
 *
 *****************************************************************************/
void wayptDrawTable(void)
{
  TablePtr 		tblP;			/* table ptr */
  FormPtr 		frmP;			/* form ptr */
  ScrollBarPtr		sclP;			/* scrollbar ptr */
  ListPtr		lstP;			/* list ptr */
  UInt16		top;			/* table row counter */
  UInt16		row;			/* table row counter */
  UInt16		rec;			/* record nr */
  UInt16		max_rows;		/* total nr. of waypoints */
  UInt16		numRows;		/* number of visible rows */
  MemHandle		waypointDBEntry;	/* record handle */
  waypoint_t		waypoint;		/* unpacked waypoint */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
  Char			tmp[20];		/* string buffer */
  UInt16		smbl;			/* waypoint symbol ID */
  MemHandle		resH;			/* resource handle */
  BitmapType		*bmpP;			/* bitmap ptr */
  RectangleType		rect;			/* rectangle */
  
  frmP = FrmGetActiveForm();
  tblP = GetObjectFromForm(frmP, WaypointTable);
  lstP = GetObjectFromForm(frmP, WaypointDispList);
  top = gWayptTopRow;
  
  max_rows = DmNumRecordsInCategory(gWaypointDB, dmAllCategories);

  /* number of visible rows */
  numRows = TblGetNumberOfRows(tblP);

  /* set columns usable */
  TblSetColumnUsable(tblP, 0, true);
  TblSetColumnUsable(tblP, 1, true);
  TblSetColumnUsable(tblP, 2, true);
  
  rec = 0;
  
  /* seek to record at top of the table */
  DmSeekRecordInCategory(gWaypointDB, &rec /* index */, top /* offset */,
    dmSeekForward, dmAllCategories);
  
  /* set row data to Waypoint DB entries (unique waypoint IDs) */
  for (row = 0; row < numRows; row++) {
    if (row + top < max_rows) {
      /* get next record */
      waypointDBEntry = DmQueryNextInCategory(gWaypointDB, &rec, 
        dmAllCategories);
      
      /* set row data to record number */
      TblSetRowData(tblP, row, (UInt32)(rec++));
    }
  }

  top = gWayptTopRow;
  
  for (row = 0; row < numRows; row++, top++) {
    
    if (top < max_rows) {
      /* set all rows selectable */
      TblSetRowSelectable(tblP, row, true);
      
      /* set columns to custom style, set custom cell drawing functions */
      TblSetItemStyle(tblP, row, 0, customTableItem);  
      TblSetCustomDrawProcedure(tblP, 0, TableDrawIdentCell);
      TblSetItemStyle(tblP, row, 1, customTableItem);    
      TblSetCustomDrawProcedure(tblP, 1, TableDrawPosnCell);
      TblSetItemStyle(tblP, row, 2, customTableItem);    
      TblSetCustomDrawProcedure(tblP, 2, TableDrawPosnCell);
      
      /* mark row as usable */
      TblSetRowUsable(tblP, row, true);
      
      /* mark all rows as invalid to trigger table update by FrmDrawForm() */
      TblMarkRowInvalid(tblP, row);
    } else {
      /* rows empty, mark as not usable */
      TblSetRowUsable(tblP, row, false);
    }
  }

  sclP = GetObjectFromForm(frmP, WaypointScrl);

  /* set scrollbar: value, min, max, pagesize */
  if (max_rows <= numRows) {
    SclSetScrollBar(sclP, 0, 0, 0, numRows);
  } else {
    SclSetScrollBar(sclP, gWayptTopRow, 0, max_rows - numRows, numRows);
  }
  
  top = gWayptTopRow;
  
  /*
   * Selected waypoint scolled below visible table rows,
   *   selection is switched to last visible entry.
   */
  if ( gWayptSel > (gWayptTopRow + numRows - 1) ) {
    gWayptSel = gWayptTopRow + numRows - 1;
  }

  /*
   * Selected waypoint scolled above visible table rows,
   *   selection is switched to first visible entry.
   */  
  if (gWayptSel < gWayptTopRow) {
    gWayptSel = gWayptTopRow;
  }
  
  if (max_rows == 0) {
    /* no data - clear fields */
    SetFieldText(WaypointCmntField, NULL, false, true);
    SetFieldText(WaypointProxField, NULL, false, true);    
    SetFieldText(WaypointDispField, NULL, false, true);

    /* clear waypoint */
    RctSetRectangle(&rect, WPT_X-1, WPT_Y-1, 16 /* w */, 14 /* h */);
    WinEraseRectangle(&rect, 0 /* cornerDiam */ );
  }
  
  for (row = 0; row < numRows; row++, top++) {
    if ((top < max_rows) && (top == gWayptSel)) {
      TblSelectItem(tblP, row, 0);

      /* Update additinal info for selected waypoint */
      /* get record, record# = row# */
      waypointDBEntry = DmQueryRecord(gWaypointDB, top);

      /* unpack waypoint */
      packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
      UnpackWaypoint(&waypoint, packed_waypoint);

      /* set comment field */
      SetFieldTextFromStr((FieldPtr) GetObjectFromActiveForm(WaypointCmntField),
	(Char *)waypoint.cmnt, true);

      /* set proximity distance field */
      format_number(waypoint.dst, 0, tmp);
      SetFieldTextFromStr((FieldPtr) GetObjectFromActiveForm(WaypointProxField),
	tmp, true);

      /* set display mode field */
      SetFieldTextFromStr((FieldPtr) GetObjectFromActiveForm(WaypointDispField),
	LstGetSelectionText(lstP, waypoint.dspl), true);

      /* show waypoint symbol */
      /* correction for limited Resource ID range */
      smbl = (waypoint.smbl >= 16383) ? waypoint.smbl-16000 : waypoint.smbl;
      resH = DmGetResource('Tbmp', smbl);

      /* use default if symbol unavailable */
      if (!resH) {
	resH = DmGetResource('Tbmp', WayptDefBmp);
      }

      /* draw bitmap */
      if (resH) { 
	RctSetRectangle(&rect, WPT_X-1, WPT_Y-1, 16 /* w */, 14 /* h */);
	WinEraseRectangle(&rect, 0 /* cornerDiam */ );
	bmpP = MemHandleLock(resH);
	WinDrawBitmap(bmpP, WPT_X, WPT_Y);
	MemHandleUnlock(resH);
	DmReleaseResource(resH);
      }

      /* unlock record */
      MemHandleUnlock(waypointDBEntry);
      
      break;
    }
  }

  FrmDrawForm(frmP);
}


/****************************************************************************/
/**
 * \brief	Terminate Garmin mode, close host port and open GPS port.
 *
 * \note	This function could actually be called
 *              from WaypointFormHandleEvent() upon reception of a
 *              frmCloseEvent. Unfortunately, when a Bluetooth GPS device
 *		is used and SrmOpen/SrmExtOpen opens the connection progress
 *		window, the frmOpenEvent queued by calling FrmGotoForm()
 *		previously seems to have been lost.
 *		As a (rather awkward) workaround, we have to call
 *		ChangePorts() _before_ FrmGotoForm() (scattered throughout
 *		WaypointFormHandleEvent()).
 *
 ****************************************************************************/
static void ChangePorts(void)
{
  Err		err;		/* error code */

  /* Terminate Garmin mode and close serial port */
  GPSEndOfOperation();

  /*
   * open GPS port
   *
   * Note:	At this stage we have to cut corners. Even if the
   *		NewSerialManager is installed, we do not open the
   *		PortForm again to let the user select the connection
   *		profile. The error checking is also slightly simplified
   *          (because we have already checked that the serial port
   *          is available).
   */
  if (gNewSerialManager) {

    if (gSerialPort == sysFileCVirtRfComm) {
      /*
       * Bluetooth Virtual Serial Port --
       * Connect to known device (address)
       */
      err = BluetoothOpen(gSerialPort, gPrefs.serprefs.gps_baud,
	gPrefs.serprefs.gps_btaddr, &gPortID);

    } else {
      /* open non-BT serial port */
      err = SrmOpen(gSerialPort, gPrefs.serprefs.gps_baud, &gPortID);
    }
  } else {
    err = SerOpen(gPortID, 0, kBaudRate);
  }

  if (err) {
    EventType		event;			/* event to queue */

    FrmAlert(CantOpenSerialAlert);
    event.eType = appStopEvent;
    EvtAddEventToQueue(&event);
  }
}


/****************************************************************************/
/**
 * \brief	Waypoint Form event handler
 *
 * \param	eventP		pointer to event structure
 *
 * \return	Status flag: event handled
 ****************************************************************************/
Boolean WaypointFormHandleEvent(EventPtr eventP)
{
  Boolean		handled = false;	/* event handled flag */
  Err			err;			/* error */
  Char			tmp[20];		/* string buffer */
  FormPtr		frmP;			/* form ptr */
  FieldPtr		fldP;			/* field ptr */
  TablePtr		tblP;			/* table ptr */
  ListPtr		lstP;			/* list ptr */
  UInt16		rec;			/* record index */
  MemHandle		waypointDBEntry;	/* waypoint record handle */
  waypoint_t		waypoint;		/* unpacked waypoint */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
  UInt16		smbl;			/* waypoint symbol ID */
  MemHandle		resH;			/* resource handle */
  BitmapType		*bmpP;			/* bitmap ptr */
  RectangleType		rect;			/* rectangle */
  UInt16		nr_waypoints;		/* number of waypoints */
  G_byte		rte_nmbr;		/* 1st route where wpt found */

  frmP = FrmGetActiveForm();
  tblP = GetObjectFromForm(frmP, WaypointTable);

  /* Number of Waypoints/Routes (could have been modified by download) */
  nr_waypoints = DmNumRecordsInCategory(gWaypointDB, dmAllCategories);
  
  switch (eventP->eType) {
  case frmOpenEvent:
    fldP = (FieldPtr)GetObjectFromActiveForm(WaypointGpslibField);
    
    /* close serial port - Garmin uses 9600 baud */
    DoClose(gPortID);
    
    /* configure GPS4Palm to emulate Garmin III */
    GarminSetHostMode(fldP, GIII /* EmulatedGPS */);
    
    /* receive waypoints from host */
    GarminGetWpts(store_wpt, fldP);
    
    /* receive routes from host */
    GarminGetRtes(store_rte, fldP);
    
    /* receive tracks from host */
    GarminGetTrks(store_trk, fldP);
    
    frmP = FrmGetActiveForm();
    WaypointFormInit(frmP);
    FrmDrawForm(frmP);
        
    /* draw the table */
    wayptDrawTable();

    handled = true;
    break;

  case nilEvent:
    GpslibTicks();
    handled = true;
    break;

  case TransferWptReqEvent:
    /* gpslib - request to transfer waypoints */
    fldP = (FieldPtr)GetObjectFromActiveForm(WaypointGpslibField);
    GarminPutWpts(get_next_wpt, fldP, nr_waypoints);
    handled = true;
    break;
  
  case TransferRteReqEvent:
    /* gpslib - request to transfer routes */
    fldP = (FieldPtr)GetObjectFromActiveForm(WaypointGpslibField);
    GarminPutRtes(get_next_rte, fldP, NumRtePackets());
    handled = true;
    break;
  
  case TransferTrkReqEvent:
    /* gpslib - request to transfer tracks */
    fldP = (FieldPtr)GetObjectFromActiveForm(WaypointGpslibField);
    GarminPutTrks(get_next_trk, fldP, NumTrkPackets());
    handled = true;
    break;
        
  case menuEvent:
    switch (eventP->data.menu.itemID) {

      case PositionMenu:
	ChangePorts();
	FrmGotoForm(GPSMainForm);
	handled = true;
	break;

      case SkyviewMenu:
	ChangePorts();
	FrmGotoForm(SkyviewForm);
	handled = true;
	break;

      case NavigationMenu:
	ChangePorts();
	FrmGotoForm(NavigationForm);
	handled = true;
	break;

      case MapMenu:
	ChangePorts();
	FrmGotoForm(MapForm);
	handled = true;
	break;

      case MiscOptsMenu:
	ChangePorts();
	FrmGotoForm(MapOptsForm);
	handled = true;
	FrmGotoForm(MiscOptsForm);
	handled = true;
	break;

      case MapOptsMenu:
	ChangePorts();
	FrmGotoForm(MapOptsForm);
	handled = true;
	break;

      case PortsMenu:
	if (gNewSerialManager) {
	  ChangePorts();
	  FrmGotoForm(GPSPortForm);
	}
	handled = true;
	break;

      case RouteMenu:
	ChangePorts();
	FrmGotoForm(RouteForm);
	handled = true;
	break;

      case TrackMenu:
	ChangePorts();
	FrmGotoForm(TrackForm);
	handled = true;
	break;

      case AboutMenu:
	ChangePorts();
	FrmGotoForm(AboutForm);
	handled = true;
	break;

    }
    break;  

  case tblSelectEvent:
    gWayptSel = eventP->data.tblSelect.row + gWayptTopRow;
    
    /* always select column 0 */
    if (eventP->data.tblSelect.column != 0) {
      TblSelectItem(tblP, eventP->data.tblSelect.row, 0);
    }

    /* get record index from row data */
    rec = (UInt16)TblGetRowData(tblP, gWayptSel - gWayptTopRow);

    /* get record */
    waypointDBEntry = DmQueryRecord(gWaypointDB, rec);

    /* unpack waypoint */
    packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
    UnpackWaypoint(&waypoint, packed_waypoint);

    /* set comment field */
    SetFieldTextFromStr((FieldPtr) GetObjectFromActiveForm(WaypointCmntField), 
      (Char *)waypoint.cmnt, true);

    /* set proximity distance field */
    format_number(waypoint.dst, 0, tmp);
    SetFieldTextFromStr((FieldPtr) GetObjectFromActiveForm(WaypointProxField), 
      tmp, true);

    /* set display mode field */
    lstP = GetObjectFromActiveForm(WaypointDispList);
    SetFieldTextFromStr((FieldPtr) GetObjectFromActiveForm(WaypointDispField), 
      LstGetSelectionText(lstP, waypoint.dspl), true);
    
    /* show waypoint symbol */
    /* correction for limited Resource ID range */
    smbl = (waypoint.smbl >= 16383) ? waypoint.smbl-16000 : waypoint.smbl;
    resH = DmGetResource('Tbmp', smbl);
    
    /* use default if symbol unavailable */
    if (!resH) {
      resH = DmGetResource('Tbmp', WayptDefBmp);
    }
    
    /* draw bitmap */
    if (resH) { 
      RctSetRectangle(&rect, WPT_X-1, WPT_Y-1, 16 /* w */, 14 /* h */);
      WinEraseRectangle(&rect, 0 /* cornerDiam */ );
      bmpP = MemHandleLock(resH);
      WinDrawBitmap(bmpP, WPT_X, WPT_Y);
      MemHandleUnlock(resH);
      DmReleaseResource(resH);
    }
    
    /* unlock record */
    MemHandleUnlock(waypointDBEntry);
    
    handled = true;
    break;
  
  case sclRepeatEvent:
    /* handle scrollbar */
    Doscroll(WaypointTable, WaypointScrl, nr_waypoints, &gWayptTopRow);
    wayptDrawTable();
    /* Note: scrollbar needs to handle the event, too */
    handled = false;
    break;
  
  case keyDownEvent:
    switch (eventP->data.keyDown.chr) {
    
    case pageUpChr:
    case pageDownChr:
      Dopage(WaypointTable, (eventP->data.keyDown.chr == pageDownChr),
        nr_waypoints, &gWayptTopRow);
      wayptDrawTable();
      /* Note: scrollbar needs to handle the event, too */
      handled = false;
      break;
    }
    break;
  
  case ctlSelectEvent:
    switch (eventP->data.ctlSelect.controlID) {

      case WaypointDeleteButton:
        if (nr_waypoints == 0)
	  break;

	/* get record index from row data */
	rec = (UInt16)TblGetRowData(tblP, gWayptSel - gWayptTopRow);

        if ( QueryRouteWaypoint(rec, &rte_nmbr) ) {
	  StrIToA(tmp, rte_nmbr);
	  FrmCustomAlert(WaypointDelErrorAlert, tmp, "", ""); 
	  break;
	}

	if ( FrmAlert(WaypointDelConfirmAlert) == 0 ) {
          
          /*
           * Only data base backup, no synchronizing available.
	   * This means a record can simply be deleted.
           */
	  DmRemoveRecord(gWaypointDB, rec);

          /* update number of records */
	  nr_waypoints = DmNumRecordsInCategory(gWaypointDB, dmAllCategories);
	  
	  /* redraw the table */
	  gWayptTopRow = 0;
	  gWayptSel = 0;
	  wayptDrawTable();
	}
	handled = true;
	break;
    
      case WaypointNewButton:
        /* create new waypoint */
	ChangePorts();
	FrmGotoForm(WayptEditForm);
	gSelWaypoint.valid = false;
	handled = true;
	break;

      case WaypointCopyButton:
        /* create new waypoint starting from existing one */
        if (nr_waypoints == 0)
	  break;

	/* get record index from row data */
	rec = (UInt16)TblGetRowData(tblP, gWayptSel - gWayptTopRow);
	
	/* get unique waypoint ID */
        err = DmRecordInfo(gWaypointDB, rec, NULL /* *attrP */,
	  &gSelWaypoint.waypointID /* *uniqueIDP */, NULL /* *chunkIDP */);
	
	/* mark selected waypoint as valid/invalid */
	gSelWaypoint.valid = (err == errNone) ? true : false;

	ChangePorts();
	FrmGotoForm(WayptEditForm);
	handled = true;
	break;
      
      case WaypointGotoButton:
        if (nr_waypoints == 0)
	  break;

	/* get record index from row data */
	rec = (UInt16)TblGetRowData(tblP, gWayptSel - gWayptTopRow);
	
	/* get unique waypoint ID */
        err = DmRecordInfo(gWaypointDB, rec, NULL /* *attrP */,
	  &gPrefs.act_wpt.waypointID /* *uniqueIDP */, NULL /* *chunkIDP */);
	
	/* mark active waypoint as valid/invalid */
	gPrefs.act_wpt.valid = (err == errNone) ? true : false;
	
	/* mark active route as invalid */
	gPrefs.act_rte.valid = false;
	
        ChangePorts();
	if (gGPSData.scroll) {
	  MemHandle		waypointDBEntry;	/* waypoint record handle */	  
	  packed_waypoint_t	*packed_waypoint;	/* packed waypoint (ptr) */
	  waypoint_t		waypoint;		/* unpacked waypoint */
	  int rec_ = rec;

	  /* get next record */
	  waypointDBEntry = DmQueryNextInCategory(gWaypointDB,
						  &rec_, 
						  dmAllCategories);
	  /* unpack waypoint */
	  packed_waypoint = (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
	  UnpackWaypoint(&waypoint, packed_waypoint);
	  /* convert semicircles to floating point */
	  gGPSData.lat = semi2deg(waypoint.posn.lat);
	  gGPSData.lon = semi2deg(waypoint.posn.lon);

	  MemHandleUnlock(waypointDBEntry);

	  FrmGotoForm(MapForm);
	} else {
	  FrmGotoForm(NavigationForm);
	}
	handled = true;
	break;
    }
    break;

  case frmCloseEvent:    
    if (gPrefs.act_wpt.valid) {
      /* check if active waypoint still exists, mark as invalid if not */
      err = DmFindRecordByID(gWaypointDB, gPrefs.act_wpt.waypointID, &rec);
      gPrefs.act_wpt.valid = (err) ? false : true;
    }
    
    /* Deallocate Fields' Text Memory */
    SetFieldText(WaypointCmntField, NULL, false, true);
    SetFieldText(WaypointDispField, NULL, false, true);
    SetFieldText(WaypointProxField, NULL, false, true);
    SetFieldText(WaypointGpslibField, NULL, false, true);

    WaypointFormDeinit(FrmGetActiveForm());
    handled = false;
    break;

  default:
    break;
  }
  
  return handled;
}
