/*****************************************************************************
 *
 * $RCSfile: WayptEditForm.c,v $
 *
 * GPS4Palm Waypoint Edit Form
 *
 * This program is Copyright (C) 03/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-03-25 13:33:55 $
 *
 * $Revision: 1.19 $
 *
 * $Log: WayptEditForm.c,v $
 * Revision 1.19  2005-03-25 13:33:55  mp
 * increased field_str buffer length
 *
 * Revision 1.18  2005/01/30 21:00:44  mp
 * corrected comment field truncation
 *
 * Revision 1.17  2005/01/29 10:22:26  mp
 * fixed comment
 *
 * Revision 1.16  2005/01/29 10:12:09  mp
 * added waypoint import from GeoDB
 *
 * Revision 1.15  2004/12/10 19:57:33  mp
 * replaced DmGet1Resource by DmGetResource and WayptEditSymScrl by
 * WayptEditPrevSymButton and WayptEditNextSymButton,
 * implemented clipboard, SysKeyboardDialog, and SysGraffitiReferenceDialog
 *
 * Revision 1.14  2004/12/09 17:26:19  mp
 * replaced default waypoint identifier/comment strings by string resources
 *
 * Revision 1.13  2004/11/27 10:16:57  mp
 * replaced ErrFatalDisplay() by Die() for memory allocation failure
 *
 * Revision 1.12  2004/11/24 21:28:14  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.11  2004/11/23 17:46:31  mp
 * removed unused variable
 *
 * Revision 1.10  2004/11/16 20:44:18  mp
 * added check after creating new record, modified PackWaypoint() call
 *
 * Revision 1.9  2004/04/17 12:47:13  mp
 * modified waypoint display mode encoding according to Garmin I/O Spec
 *
 * Revision 1.8  2004/03/31 15:47:59  mp
 * added default waypoint display on frmOpenEvent
 *
 * Revision 1.7  2004/03/30 19:32:40  mp
 * added Waypoint Symbol selection and display
 *
 * Revision 1.6  2004/03/15 17:20:21  mp
 * added Waypoint Copy function
 *
 * Revision 1.5  2004/03/11 18:54:31  mp
 * fixed handling of sign (for DM/DMS and special cases)
 *
 * Revision 1.4  2004/03/10 17:35:57  mp
 * Modifications due to changes in Data.h, fixed UTM input.
 *
 * Revision 1.3  2004/03/09 19:12:35  mp
 * improved input field handling
 *
 * Revision 1.2  2004/03/08 20:26:44  mp
 * Added Proxmity Distance field, fixed Waypoint Identifier input, fixed
 * added handling of active field when WayptEditOkButton is pressed.
 *
 * Revision 1.1  2004/03/07 14:25:22  mp
 * initial version
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include "ResourceDefines.h"
#include <Unix/sys_types.h>
#include "stringil.h"
#include "WayptEditForm.h"
#include "GPS.h"		/*  ReadFromGPS() */
#include "Serial.h"
#include "Utils.h"
#include "Data.h"
#include "utm.h"
#include "geo.h"
#include "geodb.h"
#include "fp.h"
#include "common.h"

#define WPT_X	90
#define WPT_Y	30

/* Global Variables */

/** store the location data with importGeoDBLocation() */
Location  location;

extern PrefsType	gPrefs;		/* Preferences data structure */
extern DmOpenRef	gWaypointDB;	/* Waypoint Data Base reference */
extern GPSType		gGPSData;	/* GPS Data */
extern ActWayptType	gSelWaypoint;	/* Selected waypoint */


/* Static Functions */
static void WayptEditFormInit(void)			WAYPTEDIT_SECTION;
static Boolean scan_position(UInt16 fieldID,
  double *lat, double *lon)				WAYPTEDIT_SECTION;
static Boolean scan_zone(UInt16 *utm_zone,
  Char *utm_belt)					WAYPTEDIT_SECTION;
static void add_waypoint(double lat, double lon,
  UInt16 smbl)						WAYPTEDIT_SECTION;
static Boolean insert_position(double lat, double lon,
  double *eastP, double *northP)			WAYPTEDIT_SECTION;
static Boolean field_active(UInt16 fieldID)		WAYPTEDIT_SECTION;
static void select_symbol(UInt16 *sym, Int16 dir)	WAYPTEDIT_SECTION;
static Boolean geodb_str_to_deg(Location *loc, 
  double *lat, double *lon)				WAYPTEDIT_SECTION;

/*****************************************************************************
 * FUNCTION:	WayptEditFormInit
 *
 * DESCRIPTION: Waypoint Edit Form Init Function (not used)
 *
 ****************************************************************************/
static 
void WayptEditFormInit(void)
{  

}


/*****************************************************************************
 * FUNCTION:	field_active
 *
 * DESCRIPTION: Determines if the specified field is active, i.e. has the
 *		input focus.
 *
 * PARAMETERS:	fieldID		-- Field ID
 *
 * RETURNED:    Status flag: specified field has focus
 ****************************************************************************/
static
Boolean field_active(UInt16 fieldID)
{
  FormPtr		frmP = FrmGetActiveForm();
  FieldAttrType		fld_attr;
  
  FldGetAttributes(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, fieldID)),
    &fld_attr);
  
  return (fld_attr.hasFocus == 1);
}


/*****************************************************************************
 * FUNCTION:	scan_position
 *
 * DESCRIPTION: Scan position fields to latitude/longitude
 *		depending on selevted position units.
 *
 *		Note: The code assumes to run on a device with
 *		      latin character set.
 *		Note: The degree sign '' is named chrDegreeSign in
 *                    System/CharLatin.h (but we are sloppy...)
 *
 * GLOBALS:	gPrefs		-- Application Preferences
 *
 * PARAMETERS:	fieldID		-- ID of selected field
 *		lat		-- pointer to latitude
 *		lon		-- pointer to longitude
 *
 * RETURNED:    Status flag: valid entry
 ****************************************************************************/
static
Boolean scan_position(UInt16 fieldID, double *lat, double *lon)
{
  FormPtr	frmP;		/* form ptr */
  Char		*strP;		/* string ptr */
  Char		*trailP;	/* trailing string ptr */
  double	tmp;		/* temp. double */
  short		frac;		/* number of fractional digits */
  Int16		sign;		/* sign */
  
  frmP = FrmGetActiveForm();
  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
	    FrmGetObjectIndex(frmP, fieldID)));
  
  if (strP[0] == '-') {
    sign = -1;
    strP++;
  } else {
    sign = 1;
  }
  
  /* first value is either angle in degrees or cartesian coordinate in meters */
  tmp = cvt_atof(strP, &trailP, &frac);

  /* handle degrees part or esting/northing */
  switch (gPrefs.units.pos_unit) {
    case POS_D:
      /*
       * Field's contents considered as a single floating point number, 
       * which may or may not end with a degree sign (but nothing else).
       */
      if (*trailP == '\0' || *trailP == '') {
        /* valid entry */
	if (fieldID == LatitudeField) {
	  /* Latitude */
	  if ((tmp > 90.0) || (tmp < -90.0)) {
	    /* range error */
	    return false;
	  }
	  /* o.k., update position */
	  *lat = tmp * sign;

	} else {
	  /* Longitude */
	  if ((tmp > 180.0) || (tmp < -180.0)) {
	    /* range error */
	    return false;
	  }
	  /* o.k., update position */
	  *lon = tmp * sign;
	
	}
	return true;
	
      } else {
        /* invalid entry */
	return false;
      }
      break;
      
    case POS_DM:
    case POS_DMS:
      /*
       * Field's contents must start with an integer number, 
       * which must end with a degree sign.
       */
      if (*trailP != '' || frac > 0) {
        /* invalid entry */
	return false;
      } else {
        /* valid entry */
	if (fieldID == LatitudeField) {
	  /* Latitude */
	  if ((tmp > 90.0) || (tmp < -90.0)) {
	    /* range error */
	    return false;
	  }
	  /* o.k., update position */
	  *lat = tmp;

	} else {
	  /* Longitude */
	  if ((tmp > 180.0) || (tmp < -180.0)) {
	    /* range error */
	    return false;
	  }
	  /* o.k., update position */
	  *lon = tmp;
	}
      }
      break;
      
    case POS_UTM:
      /*
       * Field's contents must consist of a single floating point number only 
       */
      if (*trailP != '\0') {
        /* invalid entry */
	return false;
      }
      if (fieldID == LatitudeField) {
        if ((tmp >= 100000.0) && (tmp <= 900000.0)) {
	  *lat = tmp;
          return true;
	} else {
	  return false;
	}
      } else {
        if ((tmp >= 0.0) && (tmp <= 10000000.0)) {
          *lon = tmp;
	  return true;
        } else {
	  return false;
	}
      }
      break;
  } /* switch (gPrefs.units.pos_unit) */

  /* skip degree sign */
  strP = trailP + 1;
  
  /* convert string to float */
  tmp = cvt_atof(strP, &trailP, &frac);  

  /*
   * Handle minutes part
   *   The following input is valid:
   *   POS_DM:  <mm.m>'
   *   POS_DMS: <mm>'
   */
  if (*trailP != '\'') {
    /* invalid entry */
    return false;
  }

  if (gPrefs.units.pos_unit == POS_DMS && frac > 0) {
    /* invalid entry */
    return false;
  }
  if ((tmp >= 60.0) || (tmp < 0.0)) {
    /* range error */
    return false;
  }

  if (fieldID == LatitudeField) {
    /* o.k., update Latitude  */
    *lat += tmp/60.0;
  } else {
    /* o.k., update Longitude  */
    *lon += tmp/60.0;
  }
    
  if (gPrefs.units.pos_unit == POS_DMS) {

    /* skip minutes sign */
    strP = trailP + 1;

    /* convert string to float */
    tmp = cvt_atof(strP, &trailP, NULL);  

    /*
     * Handle seconds part
     *   The following input is valid:
     *   <ss.s>" or <ss>"
     */
    if (*trailP != '\"') {
      /* invalid entry */
      return false;
    }
    if ((tmp >= 60.0) || (tmp < 0.0)) {
      /* range error */
      return false;
    }
    if (fieldID == LatitudeField) {
      /* o.k., update Latitude  */
      *lat += tmp/3600.0;
    } else {
      /* o.k., update Longitude  */
      *lon += tmp/3600.0;
    }
  } /* if (gPrefs.units.pos_unit == POS_DMS) */

  if (fieldID == LatitudeField) {
    /* update sign of Latitude  */
    *lat = *lat * sign;
  } else {
    /* update sign of Longitude  */
    *lon = *lon * sign;
  }

  return true;     
}


/*****************************************************************************
 * FUNCTION:	scan_zone
 *
 * DESCRIPTION: Scan zone field to UTM zone and belt,
 *		update latitude and longitude
 *
 * PARAMETERS:	eventP		-- pointer to event structure
 *
 * RETURNED:    Status flag: valid entry
 ****************************************************************************/
static
Boolean scan_zone(UInt16 *utm_zone, Char *utm_belt)
{
  FormPtr	frmP;
  Char		*strP;
  UInt16	zone;
  Char		belt;
  
  frmP = FrmGetActiveForm();
  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
	    FrmGetObjectIndex(frmP, ZoneField)));

  if ((*strP >= '0') && (*strP <= '9')) {
    zone = *strP - '0';
    strP++;
  } else {
    return false;
  }
  
  if ((*strP >= '0') && (*strP <= '9')) {
    zone = zone * 10 + (*strP - '0');
    strP++;
  }
  
  if ((zone < 1) || (zone > 60)) {
    return false;
  }
  
  belt = toupper(*strP);
  
  if ((belt >= 'C') && (belt <= 'X') &&
      (belt != 'I') && (belt != 'O')) {
    *utm_zone = zone;
    *utm_belt = belt;
    return true;
  } else {
    return false;
  }
}


/*****************************************************************************
 * FUNCTION:	add_waypoint
 *
 * DESCRIPTION: Add new waypoint to database
 *
 * GLOBALS:	gWaypointDB	-- Waypoint Database Reference
 *
 * PARAMETERS:	lat		-- Latitude
 *		lon		-- Longitude
 *
 * RETURNED:    %
 ****************************************************************************/
static
void add_waypoint(double lat, double lon, UInt16 smbl)
{
  FormPtr	frmP;
  ListPtr	lstP;
  waypoint_t	waypoint;
  UInt16	len;
  Char		*strP;
  UInt16	index;
  MemHandle	h;
  float		proxmity;
  
  frmP = FrmGetActiveForm();
  
  /* Copy Identifier */
  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
    FrmGetObjectIndex(frmP, IdentField)));
  len = StrLen(strP);
  MemSet((Char *)waypoint.ident, 6, 0);
  StrNCopy(waypoint.ident, strP, (len > 6) ? 6 : len);
  
  /* Copy Latitude */
  waypoint.posn.lat = deg2semi(lat);
  
  /* Copy Longitude */
  waypoint.posn.lon = deg2semi(lon);
   
  /* Copy Proxmity Distance (no sanity check done!) */
  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
    FrmGetObjectIndex(frmP, ProxField)));
  proxmity = cvt_atof(strP, NULL, NULL);
  proxmity = (proxmity < 0) ? 0 : proxmity;
  waypoint.dst = proxmity;
  
  /* Copy Symbol */
  waypoint.smbl = smbl;
  
  /* Copy Display Mode */
  lstP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DisplayList));
  switch ( LstGetSelection(lstP) ) {
    case 0:
      /* none */
      waypoint.dspl = dspl_smbl_none;
      break;

    case 1:
      /* symbol only */
      waypoint.dspl = dspl_smbl_only;
      break;
      
    case 2:
      /* with name */
      waypoint.dspl = dspl_smbl_name;
      break;
    
    case 3:
      /* with comment */
      waypoint.dspl = dspl_smbl_cmnt;
      break;
  }  
  
  /* Copy Comment */
  strP = FldGetTextPtr(FrmGetObjectPtr(frmP,
    FrmGetObjectIndex(frmP, CommentField)));
  StrCopy(waypoint.cmnt, strP);
  
  /* Find sort position of new record in waypoint DB */  
  index = DmFindSortPosition(gWaypointDB, &waypoint, 0 /* newRecordInfo */,
    (DmComparF *)CompareWayptIdent /* DmComparF *compar */, 0 /* other */);
    
  h = DmNewRecord(gWaypointDB, &index, 
    1 /* size, will be set by PackWaypoint() */);

  if (h) {
    DmReleaseRecord(gWaypointDB, index, false /* dirty */);
  } else {
    Die("Cannot create waypoint record!");
  }
  
  /* Pack Waypoint and store it in Database */
  PackWaypoint(&waypoint, index);
  
}


/*****************************************************************************
 * FUNCTION:	insert_position
 *
 * DESCRIPTION: Insert position to fields
 *
 * GLOBALS:	gPrefs		-- Application Preferences
 *
 * PARAMETERS:	lat		-- Latitude
 *		lon		-- Longitude
 *		eastP		-- UTM Easting (ptr)
 *		northP		-- UTM Northing (ptr)
 *
 * RETURNED:    flag	-- Conversion to UTM successful. Always true, if any
 *		position format other than UTM is set in the preferences.
 *		If UTM position format is selected and the conversion is
 *		successful, easting and northing are updated by eastP/northP.
 ****************************************************************************/
static
Boolean insert_position(double lat, double lon, double *eastP, double *northP)
{
  Char		lat_str[20];	/* string buffer */
  Char		lon_str[20];	/* string buffer */
  Char		field_str[5];	/* string buffer */
  Int32		err;		/* error code */
  Int32		zone;		/* UTM Zone */
  char		hemisphere;	/* 'N' / 'S' */
  double	easting;	/* Easting in m */
  double	northing;	/* Northing in m */
  
  switch (gPrefs.units.pos_unit) {
    case POS_D:
      /* Degrees */
      format_number(lat, 7, lat_str);
      StrCat(lat_str, "");
      SetFieldText(LatitudeField, lat_str, false, true);
      
      format_number(lon, 7, lon_str);
      StrCat(lon_str, "");
      SetFieldText(LongitudeField, lon_str, false, true);
      break;
    
    case POS_DM:
      deg_to_dm_str(lat, lon, lat_str, lon_str);
      SetFieldText(LatitudeField, lat_str, false, true);
      SetFieldText(LongitudeField, lon_str, false, true);
      break;

    case POS_DMS:
      deg_to_dms_str(lat, lon, lat_str, lon_str);
      SetFieldText(LatitudeField, lat_str, false, true);
      SetFieldText(LongitudeField, lon_str, false, true);
      break;
      
   case POS_UTM:   
      /* Universal Transverse Mercator - convert from lat/lon double values */ 
      err = Convert_Geodetic_To_UTM (
      				deg2rad(lat) /* Latitude (rad) */,
                                deg2rad(lon) /* Longitude (rad) */,
                                &zone /* long  *Zone */,
                                &hemisphere /* Hemisphere */,
                                &easting /* *Easting */,
                                &northing /* *Northing */);
      
      if (err == UTM_NO_ERROR) {
        /* print zone and belt to string */
	StrPrintF(field_str, "%ld%c ", zone, Lat_To_UTM_Belt(lat));
	SetFieldText(ZoneField, field_str, false, true);
	
	/* print easting to string */
	format_number(easting, 1, lat_str);
        SetFieldText(LatitudeField, lat_str, false, true);
		
	/* print northing to string */
	format_number(northing, 1, lon_str);
        SetFieldText(LongitudeField, lon_str, false, true);
        
	/* return easting and northing */
	*eastP = easting;
	*northP = northing;
      } else {
        return false;
      }
      break;   
  } /* switch (gPrefs.units.pos_unit) */           
  
  return true;
}


/*****************************************************************************
 * FUNCTION:	select_symbol
 *
 * DESCRIPTION: Select waypoint symbol, show bitmap and string.
 *
 * PARAMETERS:	sym		-- Garmin symbol ID (ptr) 
 *		dir		-- change direction (-1/0/+1)
 *
 * RETURNED:    %
 ****************************************************************************/
static
void select_symbol(UInt16 *sym, Int16 dir)
{
  FormPtr		frmP;			/* form ptr */
  MemHandle		resH;			/* resource handle */
  BitmapType		*bmpP;			/* bitmap ptr */
  RectangleType		rect;			/* rectangle */
  UInt16		wpt_smbl;		/* waypt. symbol ID (Garmin)*/
  UInt16		smbl;			/* waypt. symbol ID (OS) */
	
  frmP = FrmGetActiveForm();
  wpt_smbl = *sym;
  wpt_smbl += dir;
	
  /* skip 'holes' */
  switch (wpt_smbl) {
    case sym_mob+1:
      wpt_smbl = sym_boat_ramp;
      break;

    case sym_boat_ramp-1:
      wpt_smbl = sym_mob;
      break;

    case sym_circle_x+1:
      wpt_smbl = sym_is_hwy;
      break;

    case sym_is_hwy-1:
      wpt_smbl = sym_circle_x;
      break;

    case sym_border+1:
      wpt_smbl = sym_airport;
      break;

    case sym_airport-1:
      wpt_smbl = sym_border;
      break;
  }
  
  /* update symbol ID */
  *sym = wpt_smbl;
  
  /* adjust PalmOS ID to Garmin ID */
  smbl = (wpt_smbl >= 16383) ? wpt_smbl-16000 : wpt_smbl;
  
  /* get bitmap resource */
  resH = DmGetResource('Tbmp', smbl);

  /* use default if symbol unavailable */
  if (!resH) {
    resH = DmGetResource('Tbmp', WayptDefBmp);
  }

  /* draw bitmap */
  if (resH) { 
    RctSetRectangle(&rect, WPT_X-1, WPT_Y-1, 16 /* w */, 14 /* h */);
    WinEraseRectangle(&rect, 0 /* cornerDiam */ );
    bmpP = MemHandleLock(resH);
    WinDrawBitmap(bmpP, WPT_X, WPT_Y);
    MemHandleUnlock(resH);
    DmReleaseResource(resH);
  }	

  /* get string resource */
  resH = DmGetResource('tSTR', smbl);

  /* use default if symbol string unavailable */
  if (!resH) {
    resH = DmGetResource('tSTR', WayptDefStr);
  }

  /* update field */
  if (resH) { 
    FldSetTextHandle(FrmGetObjectPtr(frmP,
      FrmGetObjectIndex(frmP, WayptEditSymField)), resH);
    FldDrawField(FrmGetObjectPtr(frmP,
      FrmGetObjectIndex(frmP, WayptEditSymField)));
    /* Note: resH still in use by field! */
    /* DmReleaseResource(resH); */
  }
}


/****************************************************************************/
/**
 * \brief	Convert GeoDB location strings to Latitude/Longitude
 *
 * \note	This function has been adapted from scan_position()
 *
 * \param	loc		GeoDB location structure
 * \param	lat		latitude returned by reference
 * \param	lon		longitude returned by reference
 *
 ****************************************************************************/
static
Boolean geodb_str_to_deg(Location *loc, double *lat, double *lon)
{
  Char		*trailP;	/* trailing string ptr */
  Char		*strP;		/* string ptr */
  double	tmp;		/* temp. double */
  short		frac;		/* number of fractional digits */
  
  /* Convert Latitude */
  
  /* first value is angle in degrees */
  tmp = cvt_atof(loc->latStr, &trailP, &frac);
  
  /*
   * Field's contents must start with an integer number, 
   * which must end with a colon.
   */
  if (*trailP != ':' || frac > 0) {
    /* invalid entry */
    return false;
  } else {

    if ((tmp > 90.0) || (tmp < 0.0)) {
      /* range error */
      return false;
    }
    /* o.k., update position */
    *lat = tmp;

  }

  /* Handle minutes part */
  /* skip degree sign/colon */
  strP = trailP + 1;
  
  /* convert string to float */
  tmp = cvt_atof(strP, &trailP, &frac);

  if ((tmp >= 60.0) || (tmp < 0.0)) {
    /* range error */
    return false;
  }

  *lat += tmp/60.0;
  
    
  /* Convert Longitude */
  
  /* first value is angle in degrees */
  tmp = cvt_atof(loc->longStr, &trailP, &frac);
  
  
  /*
   * Field's contents must start with an integer number, 
   * which must end with a colon.
   */
  if (*trailP != ':' || frac > 0) {
    /* invalid entry */
    return false;
  } else {  
    
    if ((tmp > 180.0) || (tmp < 0.0)) {
      /* range error */
      return false;
    } 
    /* o.k., update position */
    *lon = tmp;
  } 

  /* Handle minutes part */
  /* skip degree sign/colon */
  strP = trailP + 1;
  
  /* convert string to float */
  tmp = cvt_atof(strP, &trailP, &frac);

  if ((tmp >= 60.0) || (tmp < 0.0)) {
    /* range error */
    return false;
  }

  *lon += tmp/60.0;

  /* adjust signs */
  *lat = (loc->latNS) ? *lat : *lat * -1.0;
  *lon = (loc->longEW) ? *lon : *lon * -1.0;
  
  /* success */
  return true;
}


/****************************************************************************/
/**
 * \brief	Waypoint Edit Form Event Handler
 *
 * GLOBALS:	gPrefs		Application Preferences
 *
 * \param	eventP		pointer to event structure
 *
 * \return	Status flag: event handled
 ****************************************************************************/
Boolean WayptEditFormHandleEvent(EventPtr eventP)
{
  Boolean		handled;		/* flag: event handled */
  FormPtr		frmP;			/* form ptr */
  ListPtr		lstP;			/* list ptr */
  FieldPtr		fldP;			/* field ptr */
  UInt16		sel;			/* list selection */
  UInt16		rec;			/* record index */
  packed_waypoint_t	*packed_waypoint;	/* packed waypoint ptr */
  waypoint_t		waypoint;		/* unpacked waypoint */
  MemHandle		waypointDBEntry;	/* waypoint record handle */
  static double		lat;			/* latitude */
  static double		lon;			/* longitude */
  static UInt16		utm_zone = 32;		/* UTM zone */
  static char		utm_belt = 'U';		/* UTM belt */
  static Char		tmp[20];		/* temp. string buffer */
  static Boolean	lat_valid;		/* latitude valid */
  static Boolean	lon_valid;		/* longitude valid */
  Boolean		valid;			/* input valid */
  UInt16		fld_focus;		/* ID of active field */
  static UInt16		fld_focus_old;		/* ID of previously act field */
  static UInt16		wpt_smbl;		/* waypoint symbol ID (Gar.) */
  Int16			dir;			/* waypoint change direction */
  Err			err;			/* UTM error */

  frmP = FrmGetActiveForm();
  handled = false;
  
  /*
   * The fields' contents have to be checked whenever a field looses focus. 
   * Unfortunately, there is no appropriate event (ctlExitEvent is not triggered
   * by fields). So we have to track the active field manually.
   */
  if ( field_active(LatitudeField) ) {
    fld_focus = LatitudeField;
  } else if ( field_active(LongitudeField) ) {
    fld_focus = LongitudeField;
  } else if ( field_active(ZoneField) ) {
    fld_focus = ZoneField;
  } else {
    /* none */
    fld_focus = 0;
  }
  
  if ((fld_focus_old != 0) && (fld_focus != fld_focus_old)) {
    /* Pseudo-fldExitEvent */
    switch (fld_focus_old) {
      case LatitudeField:
	valid = scan_position(LatitudeField, &lat, &lon);

	if (valid) {
	  lat_valid = true; 
	} else {
	  /* restore old text */
	  SetFieldText(LatitudeField, tmp, false /* append */,
	    true /* redraw */);	  
	}
	break;

      case LongitudeField:
	valid = scan_position(LongitudeField, &lat, &lon);

	if (valid) {
	  lon_valid = true;
	} else {
	  /* restore old text */
	  SetFieldText(LongitudeField, tmp, false /* append */,
	    true /* redraw */);
	}
	break;
	
      case ZoneField:
	valid = scan_zone(&utm_zone, &utm_belt);

	if (!valid) {
	  /* restore old text */
	  SetFieldText(ZoneField, tmp, false /* append */,
	    true /* redraw */);
	}
	break;
       
    }
  }
  
  switch (eventP->eType) {
    case frmOpenEvent:
      WayptEditFormInit();
      FrmDrawForm(frmP);

      lat_valid = lon_valid = false;

      /* show objects depending on selected position units */
      if (gPrefs.units.pos_unit == POS_UTM) {
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, ZoneLabel));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, ZoneField));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, EastLabel));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, NorthLabel));
      } else {
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, LatLabel));
	FrmShowObject(frmP, FrmGetObjectIndex(frmP, LonLabel));       
      }

      if (gSelWaypoint.valid) {
        /* write selected waypoint data to fields */
	/* Note: Waypoint Symbol will be lost */
	err = DmFindRecordByID(gWaypointDB, gSelWaypoint.waypointID,
	  &rec);
	
	if (!err) {
	  /* active waypoint found */
	  waypointDBEntry = DmQueryRecord(gWaypointDB, rec);
          packed_waypoint =
	    (packed_waypoint_t *)MemHandleLock(waypointDBEntry);
          UnpackWaypoint(&waypoint, packed_waypoint);
          
	  /* copy waypoint ID */
	  MemSet((Char *)tmp, 7, 0);
	  StrNCopy(tmp, (Char *)&waypoint.ident, 6);
	  SetFieldText(IdentField, tmp, false, true);
	  
	  /* copy comment */
	  SetFieldText(CommentField, waypoint.cmnt, false, true);
	  format_number(waypoint.dst, 0, tmp);
	  
	  /* copy proximity distance */
	  SetFieldText(ProxField, tmp, false, true);
	  
	  /* set display mode pop-up selection */
	  switch (waypoint.dspl) {
	    case dspl_smbl_none:
	      /* none */
	      sel = 0;
	      break;

	    case dspl_smbl_only:
	      /* symbol only */
	      sel = 1;
	      break;

	    case dspl_smbl_name:
	      /* with name */
	      sel = 2;
	      break;

	    case dspl_smbl_cmnt:
	      /* with comment */
	      sel = 3;
	      break;
	    
	    default:
	      /* symbol only */
	      sel = 1;
	      break;
	  }
	  lstP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DisplayList));
	  LstSetSelection (lstP, sel);
	  CtlSetLabel(FrmGetObjectPtr(frmP,
	    FrmGetObjectIndex(frmP, DisplayPopup)),
            LstGetSelectionText(lstP, sel));

          /* draw waypoint symbol */
	  wpt_smbl = waypoint.smbl;
          select_symbol(&wpt_smbl, 0 /* dir */);
	  
	  /* Insert current position to fields */
	  if (insert_position(semi2deg(waypoint.posn.lat),
	                      semi2deg(waypoint.posn.lon), &lat, &lon)) {
	    if (gPrefs.units.pos_unit == POS_UTM) {
	      /* easting/northing returned in lat/lon ! */
	      /* get UTM field */
	      scan_zone(&utm_zone, &utm_belt);
	    } else {
	      lat = semi2deg(waypoint.posn.lat);
	      lon = semi2deg(waypoint.posn.lon);	    
	    }  
	    lat_valid = true;
	    lon_valid = true;
	  }
	  MemHandleUnlock(waypointDBEntry);
        }
        
      } else { 
	/* write templates to fields */
        MemHandle	strH;		/* string handle */
	MemPtr		strP;		/* string ptr */
	
	/* get string from resource */
	strH = DmGetResource(strRsc, WptIdentStr);

	if (strH != NULL) {
	  /* string handle valid, lock it */
	  strP = MemHandleLock(strH);

          SetFieldText(IdentField, strP, false, true);
	  
	  /* unlock string handle */
	  MemHandleUnlock(strH);

	  /* release string resource */
	  DmReleaseResource(strH);
	}

	strH = DmGetResource(strRsc, WptCmntStr);

	if (strH != NULL) {
	  /* string handle valid, lock it */
	  strP = MemHandleLock(strH);

          SetFieldText(CommentField, strP, false, true);
	  
	  /* unlock string handle */
	  MemHandleUnlock(strH);

	  /* release string resource */
	  DmReleaseResource(strH);
	}

	SetFieldText(ProxField, "1000.0", false, true);

	switch (gPrefs.units.pos_unit) {
          case POS_D:
            SetFieldText(LatitudeField, "DD.DDDDDDD", false, true);
	    SetFieldText(LongitudeField, "DDD.DDDDDDD", false, true);
	    break;

          case POS_DM:
            SetFieldText(LatitudeField, "DDMM.MMMM\'", false, true);
	    SetFieldText(LongitudeField, "DDDMM.MMMM\'", false, true);
	    break;

          case POS_DMS:
	    //            SetFieldText(LatitudeField, "DDMM\'SS.SSS\"", false, true);
	    //	    SetFieldText(LongitudeField, "DDDMM\'SS.SSS\"", false, true);
	    // ABMH
            SetFieldText(LatitudeField, "\'.\"", false, true);
	    SetFieldText(LongitudeField, "\'.\"", false, true);

	    break;

          case POS_UTM:
	    SetFieldText(ZoneField, "32U", false, true);
            SetFieldText(LatitudeField, "EEEEEE.E", false, true);
	    SetFieldText(LongitudeField, "NNNNNNN.N", false, true);
	    break;
	}

	/* set default pop-up selection */
	lstP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DisplayList));
	LstSetSelection (lstP, 1 /* sel */);
	CtlSetLabel(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, DisplayPopup)),
          LstGetSelectionText(lstP, 1 /* sel */));
	
	/* draw waypoint symbol */
	wpt_smbl = WayptDefBmp;
        select_symbol(&wpt_smbl, 0 /* dir */);
	
      }		         
      handled = true;
      break;
    
    case menuEvent:
      /* check if any field has focus */      
      if ( field_active(IdentField)     ||
           field_active(CommentField)   ||
	   field_active(LatitudeField)  ||
	   field_active(LongitudeField) ||
	   field_active(ProxField)) {   
        fldP = (FieldPtr)FrmGetObjectPtr(frmP, FrmGetFocus(frmP));
	handled = true;
      } else {
        fldP = NULL;
      }
      
      switch (eventP->data.menu.itemID) {
 	case UndoMenu:
	  if (fldP)
	    FldUndo(fldP);
	  break;
 	
	case CutMenu:
	  if (fldP)
	    FldCut(fldP);
	  break;
 	
	case CopyMenu:
	  if (fldP)
	    FldCopy(fldP);
	  break;
 	
	case PasteMenu:
	  if (fldP)
	    FldPaste(fldP);
	  break;	  
	  
	case KeyboardMenu:
	  if ( field_active(IdentField) || field_active(CommentField) ) {
	    SysKeyboardDialog(kbdAlpha);
	  } else if ( field_active(LatitudeField)   ||
	              field_active(LongitudeField) ||
		      field_active(ProxField) ) {
	    SysKeyboardDialog(kbdNumbersAndPunc);
	  }
	  handled = true;
	  break;
	  
	case GraffitiMenu:
	  SysGraffitiReferenceDialog(referenceDefault);
	  handled = true;
	  break;
       
      } /* switch (eventP->data.menu.itemID) */
      break;
      
    case fldEnterEvent:
      switch (eventP->data.ctlSelect.controlID) {
        case LatitudeField:
	  /*
	   * save contents, we might have to restore it
	   * if the new input is invalid
	   */
	  if (fld_focus_old != LatitudeField) {
	    StrCopy(tmp, FldGetTextPtr(FrmGetObjectPtr(frmP,
	      FrmGetObjectIndex(frmP, LatitudeField))));
	  }
	  break;

        case LongitudeField:
	  /*
	   * save contents, we might have to restore it
	   * if the new input is invalid
	   */
	  if (fld_focus_old != LongitudeField) {
	    StrCopy(tmp, FldGetTextPtr(FrmGetObjectPtr(frmP,
	      FrmGetObjectIndex(frmP, LongitudeField))));
	  }
	  break;

        case ZoneField:
	  /*
	   * save contents, we might have to restore it
	   * if the new input is invalid
	   */
	  if (fld_focus_old != ZoneField) {
	    StrCopy(tmp, FldGetTextPtr(FrmGetObjectPtr(frmP,
	      FrmGetObjectIndex(frmP, ZoneField))));
	  }
	  break;
	  
      }
      break;

    case keyDownEvent:
      switch (eventP->data.keyDown.chr) {

      case pageUpChr:
	dir = 0;
	/* increment */
	if (wpt_smbl == sym_seaplane) {
	  wpt_smbl = sym_anchor;
	} else {
	  dir = 1;
	}
        select_symbol(&wpt_smbl, dir);
	handled = true;
	break;
	
      case pageDownChr:
        dir = 0;
	/* decrement */
	if (wpt_smbl == 0) {
	  wpt_smbl = sym_seaplane;	/* wrap around */
	} else {
	  dir = -1;
	}
        select_symbol(&wpt_smbl, dir);
	handled = true;
	break;
      }
      break;
    
    case ctlSelectEvent:
      switch (eventP->data.ctlSelect.controlID) {

	case WayptEditOkButton:
	  switch (fld_focus) {
	    case LatitudeField:
	      /* Latitude field is still active, get contents */
	      valid = scan_position(LatitudeField, &lat, &lon);

	      if (valid) {
	        lat_valid = true;
	      } else {
		/* restore old text */
		SetFieldText(LatitudeField, tmp, false /* append */,
		  true /* redraw */);
	      }
              break;
	    
	    case LongitudeField:
	      valid = scan_position(LongitudeField, &lat, &lon);

	      if (valid) {
	        lon_valid = true;
	      } else {
		/* restore old text */
		SetFieldText(LongitudeField, tmp, false /* append */,
		  true /* redraw */);
	      }
	      break;

	    case ZoneField:
	      valid = scan_zone(&utm_zone, &utm_belt);

	      if (!valid) {
		/* restore old text */
		SetFieldText(ZoneField, tmp, false /* append */,
		  true /* redraw */);
	      }
	      break;

            default:
	      valid = true;
	      break;
	  } /* switch (fld_focus) */
          
	  if (gPrefs.units.pos_unit == POS_UTM) {
	    /* lat and lon actually contain easting and northing! */
	    double	easting	 = lat;
	    double	northing = lon;

	    err = Convert_UTM_To_Geodetic(utm_zone,
                        	 (utm_belt >= 'N') ? 'N' : 'S' /* Hemisphere */,
                        	 easting,
                        	 northing,
                        	 &lat,
                        	 &lon);	    
	    	    
	    if (err != UTM_NO_ERROR) {
	      break;
	    }
	    
	    lat = rad2deg(lat);
	    lon = rad2deg(lon);
	  }
	  
	  if (lat_valid & lon_valid) {
	    /* add waypoint and close form only if fields are valid */
	    add_waypoint(lat, lon, wpt_smbl);
	    FrmGotoForm(WaypointForm);
	  }
	  handled = true;
	  break;
	  
	case WayptEditCancelButton:
	  /* close form without changing database */
	  FrmGotoForm(WaypointForm);
	  handled = true;
	  break;
      
        case WayptEditHereButton:
	  /* insert current position */

	  /* throw away anything in the buffer-- we want fresh data */
	  DoReceiveFlush(gPortID, 1);
          
	  if (!gGPSData.scroll) {
	    /* get new data */
	    ReadFromGPS();
	  }

	  if (!gGPSData.valid)
	    break;
	  
	  if (fld_focus != 0) {
	    /* release focus from active field */
	    FldReleaseFocus(FrmGetObjectPtr(frmP,
	      FrmGetObjectIndex(frmP, fld_focus)));
	    fld_focus = 0;
	  }
	  
	  /* Insert current position to fields */
	  if (insert_position(gGPSData.lat, gGPSData.lon, &lat, &lon)) {
	    if (gPrefs.units.pos_unit == POS_UTM) {
	      /* easting/northing returned in lat/lon ! */
	      /* get UTM field */
	      scan_zone(&utm_zone, &utm_belt);
	    } else {
	      lat = gGPSData.lat;
	      lon = gGPSData.lon;	    
	    }  
	    lat_valid = true;
	    lon_valid = true;
	  }
	  
	  handled = true;
	  break;
	
        case WayptEditImportButton:
          /* get GeoDB data */
	  if ( importGeoDBLocation() ) {
	      Char      tempStr[80];	/* string buffer */
	      Int	dls;		/* daylight saving time array index */
              CharPtr   dlsList[] = {
	                              "No", "Yes", "Australia", "Europe",
				      "N.America", "S.America", "User Defined"
				    };
            
	    /* Convert GeoDB strings to double Latitude and Longitude */
	    if ( geodb_str_to_deg(&location, &lat, &lon) ) {
	      
	      /* Insert current position to fields */
	      if (insert_position(lat, lon, &lat, &lon)) {
	      
		if (gPrefs.units.pos_unit == POS_UTM) {
		  /* easting/northing returned in lat/lon ! */
		  /* get UTM field */
		  scan_zone(&utm_zone, &utm_belt);
		}  
		lat_valid = true;
		lon_valid = true;
	      }	      
	    }
	    
	    /* add Name */
	    if (StrLen(location.name)) {
	      StrPrintF(tempStr, "%s;",location.name);
	    }
	    
	    /* add Altitude */
	    if (StrLen(location.altStr)) {
	      StrPrintF(&tempStr[StrLen(tempStr)],
	        "Alt: %s;", location.altStr);
	    }
	    
	    /* add GMT Offset */
	    if (StrLen(location.gmtStr)) {
	      StrPrintF(&tempStr[StrLen(tempStr)], "GMT Offset: %s %s;",
	        location.gmtStr, (CharPtr)(location.gmtEW?"E":"W")); 
	    } 
	    
            /* add Daylight Saving Time */
	    
	    /* high bit set when daylight time */
	    StrPrintF(&tempStr[StrLen(tempStr)], "DST: %s ",
	      location.dayLight & 0x80?"ST":"DT");

	    dls = location.dayLight & 0x7f;					// display DayLight rule

	    if (dls <8) {
	      StrCopy(&tempStr[StrLen(tempStr)], dlsList[dls]);
	    } else {
	      StrCopy(&tempStr[StrLen(tempStr)], "DLS Error");
	    }
	    
	    /* truncate to Garmin Comment Field width */ 
	    tempStr[39] = '\0';
	    
	    /* set Comment Field*/
	    SetFieldText(CommentField, tempStr, false, true);
	  
	  } else {
	    FrmAlert(GeodbImportAlert);
          }
	  handled = true;
	  break;

	case WayptEditPrevSymButton:
	  dir = 0;
	  /* increment */
	  if (wpt_smbl == sym_seaplane) {
	    wpt_smbl = sym_anchor;
	  } else {
	    dir = 1;
	  }
          select_symbol(&wpt_smbl, dir);
	  handled = true;
	  break;
	
	case WayptEditNextSymButton:
          dir = 0;
	  /* decrement */
	  if (wpt_smbl == 0) {
	    wpt_smbl = sym_seaplane;	/* wrap around */
	  } else {
	    dir = -1;
	  }
          select_symbol(&wpt_smbl, dir);
	  handled = true;
	  break;
      }
      break;

    case frmCloseEvent:
      /* no active field */
      fld_focus = fld_focus_old = 0;
      
      /* Deallocate Fields' Text Memory */
      SetFieldText(IdentField, NULL, false, true);
      FldSetTextHandle(FrmGetObjectPtr(frmP,
        FrmGetObjectIndex(frmP, WayptEditSymField)), NULL);
      if (gPrefs.units.pos_unit == POS_UTM) {
	SetFieldText(ZoneField, NULL, false, true);
      }
      SetFieldText(LatitudeField, NULL, false, true);
      SetFieldText(LongitudeField, NULL, false, true);
      SetFieldText(ProxField, NULL, false, true);
      SetFieldText(CommentField, NULL, false, true);
      break;
  }

  /* save current focus */  
  fld_focus_old = fld_focus;
  
  return(handled);
} /* WayptEditFormHandleEvent() */
