/*****************************************************************************
 *
 * $RCSfile: file_io.c,v $
 *
 * File Stream API / VFS API Wrapper Functions
 *
 *
 * This program is Copyright (C) 03/2005 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-15 10:42:58 $
 *
 * $Revision: 1.3 $
 *
 * $Log: file_io.c,v $
 * Revision 1.3  2005-05-15 10:42:58  mp
 * modified fread() documentation
 *
 * Revision 1.2  2005/05/14 12:25:34  mp
 * added parameter fd to fread(), fclose(), feof(), rewind(), and fseek()
 * added return value to fopen()
 *
 * Revision 1.1  2005/05/13 18:36:12  mp
 * moved functions from MapForm, added fseek()
 *
 *
 ****************************************************************************/ 
#include <PalmOS.h>
#include <Extensions/ExpansionMgr/VFSMgr.h>
#include "file_io.h"
#include "getdir.h"


/****************************************************************************/
/**
 * \brief	File Stream API / VFS API wrapper function.
 *
 * \param	fd	file descriptor (FS File Handle/VFS File Reference)
 * \param	buf	storage location for read data
 * \param	size	number of bytes to read
 *
 * \return	number of items successfully read
 ****************************************************************************/ 
unsigned long fread(FileDescrType fd, void *buf, unsigned long size)
{
  unsigned long x;
  if (fd.fr)
    VFSFileRead(fd.fr, size, buf, &x);
  else
    x = FileRead(fd.fh, buf, 1, size, NULL);
  return x;
}


/****************************************************************************/
/**
 * \brief	File Stream API / VFS API wrapper function.
 ****************************************************************************/ 
FileDescrType fopen(UInt32 filecard, const char *filename, const char *dir, Err *err)
{
  MemPtr	pathName;
  UInt32	vfs_vol = filecard & ~VFS_FLAG;
  FileDescrType fd;
  
  if (!(filecard & VFS_FLAG)) {
    fd.fh = FileOpen(	  filecard, /* card nr */
			  filename,
			  'DATA',  /* type */
			  'BOXR',  /* creator */
                    	  fileModeReadOnly | fileModeAnyTypeCreator, /* mode */
			  err);   /* err */
    fd.fr = 0;
  } else {
    /* generate absolute file path */
    pathName = MemPtrNew(StrLen(dir) + StrLen(filename) + 2);
    StrCopy(pathName, dir);
    StrCat(pathName, "/");
    StrCat(pathName, filename);

    *err = VFSFileOpen(vfs_vol, pathName, vfsModeRead, &fd.fr);
    fd.fh = 0;
    MemPtrFree(pathName);

  }
  return fd;
}


/****************************************************************************/
/**
 * \brief	File Stream API / VFS API wrapper function.
 ****************************************************************************/ 
void fclose(FileDescrType fd)
{
  if (fd.fr) {
    VFSFileClose(fd.fr);
    fd.fr = 0;
  } else {
    FileClose(fd.fh);
    fd.fh = 0;
  }
}


/****************************************************************************/
/**
 * \brief	File Stream API / VFS API wrapper function.
 *
 * \return	non-zero if end-of-file is set
 ****************************************************************************/ 
int feof(FileDescrType fd)
{
  if (fd.fr)
    return VFSFileEOF(fd.fr);
  else
    return FileEOF(fd.fh);
}


/****************************************************************************/
/**
 * \brief	File Stream API / VFS API wrapper function.
 ****************************************************************************/ 
void rewind(FileDescrType fd)
{
  if (fd.fr)
    VFSFileSeek(fd.fr, vfsOriginBeginning, 0);
  else
    FileSeek(fd.fh, 0, fileOriginBeginning);
}


/****************************************************************************/
/**
 * \brief	File Stream API / VFS API wrapper function.
 ****************************************************************************/ 
int fseek(FileDescrType fd, long offset, int whence)
{
  Err err;
  
  if (fd.fr) {
    switch (whence) {
      case SEEK_SET:
        err = VFSFileSeek(fd.fr, vfsOriginBeginning, offset);
	break;
      
      case SEEK_CUR:
        err = VFSFileSeek(fd.fr, vfsOriginCurrent, offset);
	break;
      
      case SEEK_END:
        err = VFSFileSeek(fd.fr, vfsOriginEnd, offset);
	break;
      
      default:
        return -1;
    }
    
    return (err == errNone) ? 0 : -1;
    
  } else {
    switch (whence) {
      case SEEK_SET:
        err = FileSeek(fd.fh, offset, fileOriginBeginning);
	break;
      
      case SEEK_CUR:
        err = FileSeek(fd.fh, offset, fileOriginCurrent);
	break;
      
      case SEEK_END:
        err = FileSeek(fd.fh, offset, fileOriginEnd);
	break;
      
      default:
        return -1;
    }
    
    return (err == 0) ? 0 : -1;
    
  }

}
