/*****************************************************************************
 *
 * $RCSfile: fp.c,v $
 *
 * Double to String/String to Double Conversion Routines
 *
 * Written by Bill Ezell <wje@quackers.net>
 * http://www.mv.com/users/wje/pilot.html
 *
 * $Author: mp $
 *
 * $Date: 2004-11-23 17:46:11 $
 *
 * $Revision: 1.6 $
 *
 * $Log: fp.c,v $
 * Revision 1.6  2004-11-23 17:46:11  mp
 * removed unused variable
 *
 * Revision 1.5  2004/04/30 16:30:39  mp
 * modified comments for doxygen
 *
 * Revision 1.4  2004/03/20 11:17:41  mp
 * changed float to double
 *
 * Revision 1.3  2004/03/09 19:03:54  mp
 * Fixed search for decimal point in strings with multiple numerical values.
 *
 * Revision 1.2  2004/03/06 15:56:06  mp
 * added parameter trailP to cvt_atof()
 *
 * Revision 1.1  2003/12/31 12:32:17  mp
 * imported to CVS
 *
 *
 ****************************************************************************/
 // ABMHlogging
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <VFSMgr.h>
//


#include "fp.h"




#ifdef UNIX
#define StrChr(x,y)	strchr(x,y)
#define StrCopy(x,y)	strcpy(x,y)
#define StrLen(x)	strlen(x)

#include <stdlib.h>
#include <string.h>
#else
#define StrChr(x,y)	strchr(x,y)
#define StrCopy(x,y)	strcpy(x,y)
#define StrLen(x)	strlen(x)
#include <Unix/sys_types.h>
#include "stringil.h"
#endif

/* fp.c - simple float/ascii conversion routines
 *
 * These routines handle signed fixed-point strings only; no exponential
 * notation is supported.
 *
 * Contains:
 *
 * double cvt_atof(char *srcP, char **trailP, short *fracP)
 *    converts the passed string to a double. If fracP is not 0, the
 *    number of fractional digits found will be returned there. If
 *    trailP is not 0, the pointer to the remaining substring will be
 *    returned there. 
 *    An acceptable string is of the form [+-]digits[.digits]; leading
 *    whitespace is ignored.
 *
 * void format_number(double src, short digits, char *destP)
 *    converts the floating point src to a string; the value of digits
 *    determines the number of fractional digits to produce.
 *    The result will be rounded by adding 1/2 of the least-significant
 *    digit position's scaling value to the fractional part and then truncating
 *    the fractional part; e.g. 3.6 is rounded by adding .05 to the value.
 *    
 * WARNING: if you use these functions with Pilot gcc, you MUST include
 * the function prototypes (below) in your code! Otherwise, gcc will
 * generate incorrect calls and your code won't work.
 *
*/


/* Internal prototypes */

static unsigned long cvt_atoul(char *);
static void cvt_ultoa(unsigned long,char *);


/*****************************************************************************
 * FUNCTION:	cvt_ultoa
 *
 * DESCRIPTION:	Convert unsigned long value to string.
 *
 * PARAMETERS:	value	-- unsigned long value
 *		strP	-- ptr to string buffer for returning result
 *
 * RETURNED:	%
 *****************************************************************************/
static void cvt_ultoa(unsigned long value, char *strP)
{
  int i;
  char *cP;
  char tmpstr[16];		/* a 32 bit long can give up to 10 digits */

    if( value <= 9 )		/* a quick special case */
    {
	*strP++ = value + '0';
	*strP = '\0';
	return;
    }

    /* We convert the number backwards */
    for( cP = tmpstr+9, i = 10; i-- > 0; )
    {
	*cP-- = (value % 10) + '0';
	value /= 10;
    }

    tmpstr[10] = '\0';

    for( cP = tmpstr; *cP && (*cP == '0'); )
	++cP;
    
    StrCopy(strP,cP);
}


/*****************************************************************************
 * FUNCTION:	cvt_atoul
 *
 * DESCRIPTION:	Convert string to unsigned long.
 *
 * PARAMETERS:	strP	-- string to convert
 *
 * RETURNED:	unsigned long value.
 *****************************************************************************/
static unsigned long cvt_atoul(char *strP)
{
  unsigned long value;

    for( value = 0; (*strP >= '0') && (*strP <= '9'); )
    {
	value *= 10;
	value += *strP++ - '0';
    }

    return(value);
}


/*****************************************************************************
 * FUNCTION:	num_len
 *
 * DESCRIPTION:	Return length of numerical substring.
 *
 * PARAMETERS:	strP	-- string to search
 *
 * RETURNED:	Length of numerical substring.
 *****************************************************************************/
static unsigned num_len(char *strP)
{
  unsigned i = 0;
  
  while ((*strP >= '0') && (*strP <= '9')) {
    strP++;
    i++;
  }
  return i;
}


/*****************************************************************************
 * FUNCTION:	find_decimal
 *
 * DESCRIPTION:	Search for decimal point, but stop at any non-numerical
 *		character.
 *
 * PARAMETERS:	strP	-- string to search
 *
 * RETURNED:	Pointer to position of decimal point within string or NULL
 *		if no decimal point found.
 *****************************************************************************/
static char *find_decimal(char *strP)
{
  while ((*strP >= '0') && (*strP <= '9')) {
    strP++;
  }
  return ((*strP == '.') ? strP : NULL); 
}


/****************************************************************************/
/**
 * Convert string to a double precision floating point value.
 *
 * \param	strP	string to convert
 * \param	trailP	pointer to remaining, unconverted part of string
 * \param	fracP	if not null, return number of fractional digits
 *
 * \return	double precision floating point value
 *****************************************************************************/
double cvt_atof(char *strP, char **trailP, short *fracP)
{
  int i;
  int neg;
  char *cP;
  unsigned long ipart,fpart;
  double rslt;

    while( (*strP == ' ') || (*strP == '\t') )		/* trim whitespace */
	++strP;

    if( *strP == '-' )					/* check for neg. */
    {
	neg = 1;
	++strP;
    }
    else
    {
	neg = 0;

	if( *strP == '+' )				/* drop any sign */
	    ++strP;
    }

    ipart = cvt_atoul(strP);			/* get integer part */
    if( fracP )
	*fracP = 0;
    
    if ( trailP )
      *trailP = strP + num_len(strP);
    
    if( (cP = find_decimal(strP)) )
    {
	i = num_len(++cP);
	if ( trailP )
	  *trailP = cP + i;
	fpart = cvt_atoul(cP);			/* and frac part */
	rslt = (double)fpart;

	if( fracP )
	    *fracP = i;
	while( i-- > 0 )			/* scale as needed */
	    rslt /= 10.0;

	rslt += (double)ipart;
    }
    else
	rslt = (double)ipart;

    return( neg?-rslt:rslt );
}



abmh_debug(char *pos, unsigned long l, char *stri) {
  char str[16];

  Err err = errNone;
  UInt32 volumeIterator = 0;
  UInt16 firstVolume = 0;
  FileRef f;
  //    UInt8 *bufferP = NULL;
  UInt8 bufferP[255];
  UInt32	seconds;		/* current time */
  DateTimeType	dt;			/* date/time structure */
  UInt32 written = 0;
  static UInt32	last_seconds = 0;

  seconds = TimGetSeconds();

  TimSecondsToDateTime(seconds, &dt);

  // Get first card
  volumeIterator = vfsIteratorStart;
  err = VFSVolumeEnumerate(&firstVolume, &volumeIterator);
  if (!(err == expErrEnumerationEmpty)) {
    
    f = NULL;
    err = VFSFileOpen(firstVolume, "/debug.log", vfsModeWrite|vfsModeCreate, &f);
    if (!err) {
      /* write time to string */
      TimeToAscii(dt.hour, dt.minute,
		  tfColon24h,
		  bufferP );

      StrCat(bufferP, " ");	  
      StrCat(bufferP, pos);

      StrCat(bufferP, " ");
      cvt_ultoa(l,str);
      StrCat(bufferP, str);

      StrCat(bufferP, " ");
      StrCat(bufferP, stri);
      StrCat(bufferP, "\n");
      
      VFSFileSeek(f, vfsOriginEnd, 0);
      err = VFSFileWrite(f, StrLen(bufferP), bufferP, &written);
      VFSFileClose(f);      
    }
  }
}



/****************************************************************************/
/**
 * Format a double precision floating point value into a string.
 *
 * \param	value	value to format
 * \param	frac	number of fractional digits
 * \param	strP	pointer to destination string buffer
 *
 *****************************************************************************/
void format_number(double value, short frac, char *strP)
{
  char *strPP = strP;
  int size = sizeof(strP);
  int i;
  unsigned long ipart, fpart;
  double limit;
  char str[16];

  //abmh_debug("format_number().01", frac, strPP);

    if( frac > (sizeof(str) - 1) )		/* limit frac digits */
	frac = sizeof(str) - 1;


  //abmh_debug("format_number().02", frac, strPP);


    if( value < 0.0 )
    {
	*strP++ = '-';
	value = -value;
    }

  //abmh_debug("format_number().03", frac, strPP);

    ipart = (long)value;

  //abmh_debug("format_number().04", ipart, strPP);

    value -= (double)ipart;			/* recover frac part as int */
    for( limit = 1.0, i = frac; i-- > 0; )
    {
	value *= 10.0;
	limit *= 10.0;
    }

  //abmh_debug("format_number().05", value, strPP);
  //abmh_debug("format_number().06", limit, strPP);

    value += 0.5;				/* do some rounding */
    if( value >= limit )
    {
	fpart = 0;				/* overflowed */
	ipart++;
    }
    else
	fpart = (unsigned long)(value);

  //abmh_debug("format_number().07", fpart, strPP);

    cvt_ultoa(ipart,strP);

  //abmh_debug("format_number().08", ipart, strPP);

    if( frac )
    {

  //abmh_debug("format_number().09", frac, strPP);

	cvt_ultoa(fpart,str);
  //abmh_debug("format_number().10", frac, str);

	strP += StrLen(strP);
	*strP++ = '.';

  //abmh_debug("format_number().11", frac, strPP);

	for( i = StrLen(str); i++ < frac; )
	    *strP++ = '0';			/* need some padding */

  //abmh_debug("format_number().12", frac, str);

	StrCopy(strP,str);			/* and now the value */
    }

  //abmh_debug("format_number().13", size, strPP);

}



