/*****************************************************************************
 *
 * $RCSfile: geo.c,v $
 *
 * GPS4Palm Geodesic/Geometric Utility Functions
 *
 * History:
 * 2003-07-06:	first version
 *
 * This program is Copyright (C) 07/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-28 10:15:01 $
 *
 * $Revision: 1.15 $
 *
 * $Log: geo.c,v $
 * Revision 1.15  2005-05-28 10:15:01  mp
 * fixed StrPrintF format in deg_to_d_str()
 *
 * Revision 1.14  2005/04/02 07:23:37  mp
 * added deg_to_geodb_str()
 *
 * Revision 1.13  2005/03/25 13:31:59  mp
 * added deg_to_str()
 *
 * Revision 1.12  2004/11/26 19:57:34  mp
 * added deg_to_d_str(), fixed deg_to_dm_str() and deg_to_dms_str()
 *
 * Revision 1.11  2004/04/30 16:30:39  mp
 * modified comments for doxygen
 *
 * Revision 1.10  2004/04/21 15:33:00  mp
 * fixed deg2semi() and semi2deg()
 *
 * Revision 1.9  2004/04/12 10:33:05  mp
 * modified init_radius() and calcR()
 *
 * Revision 1.8  2004/03/11 18:50:43  mp
 * fixed gc_course_sphere()
 *
 * Revision 1.7  2004/03/10 17:07:07  mp
 * Added rad2deg(), deg2rad(), gc_dist_sphere(), and gc_course_sphere().
 *
 * Revision 1.6  2004/02/28 13:20:51  mp
 * added semi2deg(), deg2semi(), deg_to_dm_str(), and deg_to_dms_str()
 *
 * Revision 1.5  2004/01/18 19:53:32  mp
 * modified range check in Lat_To_UTM_Belt()
 *
 * Revision 1.4  2004/01/18 11:43:55  mp
 * added function Lat_To_UTM_Belt()
 *
 * Revision 1.3  2003/12/30 19:46:46  mp
 * replaced math.h by MathLib.h
 *
 * Revision 1.2  2003/10/20 19:12:11  mp
 * added map scale to calculation
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/

/*****************************************************************************
 * 1 Latitude : 111120m
 * 1' Latitude : 1852m = 1nmi
 *
 * 1 Longitude: 111120m * cos(lat)
 * 1' Longitude: 1852m * cos(lat)
 *
 * d_map : distance on map (in m)
 * d_act : distance in real world (actual distance) (in m)
 * r_map : map resolution (in dpi)
 *
 * d_map = d_act / scale
 * 
 * d_screen = d_map * r_map * (1/0.0254) in/m
 *
 ****************************************************************************/

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "MathLib.h"
#include "geo.h"
#include "common.h"		/* POS_D|POS_DM|POS_DMS|POS_UTM */
#include "utm.h"		/* Convert_Geodetic_To_UTM() */
#include "fp.h"

/** \brief square, i.e. x^2 */
#define SQR(x)	((x) * (x))

/** \brief epsilon, i.e. a really small number greater than zero */
#define EPS	0.00000000000000000000005

/** Earth's radius in meters at given latitude (in integer degrees) */
double Ra[101];

/*
 * PIXELFACT = r_map [dpi] * 1/0.0254 in/m
 *           = 72 dpi * 1/0.0254 in/m = 2834.65 pixel/m
 */


/****************************************************************************/
/**
 * \brief	Calculate Earth radius at given latitude 
 *		according to GRS80/WGS84 Ellipsoid.
 *
 *		see:
 * 		The Universal Grids: Universal Transverse Mercator (UTM) and
 *		Universal Polar Stereographic (UPS)
 *		Defense Mapping Agency
 *		Technical Manual 8358.2
 * 		http://www.nima.mil/GandG/pubs.html
 *		(Section 2-2.1 Ellipsoid Parameters)
 *
 * \param	lat	latitude in degrees 
 *
 * \return	Earth's radius at given latitude in meters
 ****************************************************************************/
double calcR(double lat)
{
  double a = SEMIMAJOR_AXIS/1000.0;
  double r, sc, x, y, z;
  double e2 = 0.081082 * 0.081082;

/*
the radius of curvature of an ellipsoidal Earth in the plane of the 
meridian is given by 

 R' = a * (1 - e^2) / (1 - e^2 * (sin(lat))^2)^(3/2) 

 where a is the equatorial radius, 
b is the polar radius, and 
e is the eccentricity of the ellipsoid = sqrt(1 - b^2/a^2) 

a = 6378 km (3963 mi) Equatorial radius (surface to center distance) 
b = 6356.752 km (3950 mi) Polar radius (surface to center distance) 
e = 0.081082 Eccentricity 
*/

  lat = lat * M_PI / 180.0;
  sc = sin (lat);
  x = a * (1.0 - e2);
  z = 1.0 - e2 * sc * sc;
  y = pow (z, 1.5);
  r = x / y;

  r = r * 1000.0;

  return r;
}


/****************************************************************************/
/**
 * Initialize look-up table with Earth's radii (in meters)
 * at given latitudes. To save memory and to reduce
 * time required for initialization, the symmetry is exploitet.
 * ( sin(90+x) = sin(90-x) ) 
 *
 ****************************************************************************/
void init_radius(void)
{
  Int16 i;
  
  /*  Build array for earth radii */
  for (i = 0; i <= 100; i++)
    Ra[i] = calcR(i);
}

/**********************************************************************/
/**
 * \brief	Calculate screen position from current position and
 *		map parameters.
 *
 * \param	posx		horizontal map position
 * \param	posy		vertical map position
 * \param	lon		current longitude
 * \param	lat		current latitude
 * \param	zero_lon	longitude of map's center
 * \param	zero_lat	latitude of map's center
 * \param	scale		scale factor
 *
 * Changes as compared to original version (GPSDrive):
 * - data types: gdouble to double, gint to UInt16
 * - removed screen specific calculations
 * - added zero_lon and zero_lat to parameters
 * - renamed zero_long to zero_lon for consistency
 * - changed type of posx/posy from double * to Int32 *
 *
 * TBD: Check geodetic background of formulae!
 **********************************************************************/
void calcxy(Int32 *posx, Int32 *posy, double lon, double lat,
		double zero_lon, double zero_lat, UInt32 scale)
{
  double dif;
  double x;
  double y;

  x = (Ra[(UInt32) fabs(lat)] * M_PI / 180.0) * cos (M_PI * lat / 180.0) *
        (lon - zero_lon);

  *posx = (Int32)(x * PIXELFACT / scale);

  y = (Ra[(UInt32) fabs(lat)] * M_PI / 180.0) * (lat - zero_lat);
  
  dif =
    Ra[(UInt32) fabs(lat)] * (1 - (cos ((M_PI * (lon - zero_lon)) / 180.0)));
  
  y = y + dif / 1.85;
  *posy = (Int32)(y * PIXELFACT / scale);

}


/****************************************************************************/
/**
 * Calculate UTM Belt from latitude.
 *
 * \param	lat	latitude in degrees
 *
 * \return	UTM Belt Designator
 ****************************************************************************/
char Lat_To_UTM_Belt(double lat)
{
  Int16 	belt;
  const char	utm_belt[] = "CDEFGHJKLMNPQRSTUVWX";

  /*
   * range check actually done by Convert_Geodetic_To_UTM(),
   *   so we just make sure to keep in the valid range of UTM belts
   */
  lat = (lat > 84.0) ? 84.0 : lat;
  lat = (lat < -79.0) ? -79.0 : lat; 
   
  /* correction for zone X which extends from 72 to 84 */
  if (lat >= 80.0 && lat < 84.5) {
    lat =  72.0;
  }
  
  belt = (Int16) floor(lat / 8.0);
  belt += 10;
  return utm_belt[belt];
}


/****************************************************************************/
/**
 * Convert semicircles to degrees.
 * degrees = semicircles * (180 / 2^31)
 *
 * \param	semi	angle in semicircles
 *
 * \return	angle in degrees
 ****************************************************************************/
double semi2deg(Int32 semi) {
  return ((double)(semi * (180.0/(1UL<<31))));
}


/****************************************************************************/
/**
 * Convert degrees to semicircles.
 * semicircles = degrees * (2^31 / 180)
 *
 * \param	deg	angle in degrees
 *
 * \return	angle in semicircles
 ****************************************************************************/
Int32 deg2semi(double deg)
{
  return ((Int32)(deg * ((1UL<<31)/180.0)));
}


/****************************************************************************/
/**
 * Convert radians to degrees.
 *
 * \param	rad	angle in radians
 *
 * \return	angle in degrees
 ****************************************************************************/
double rad2deg(double rad)
{
  return (rad / M_PI * 180.0);
}


/****************************************************************************/
/**
 * Convert degrees to radians.
 *
 * \param	deg	angle in degrees
 *
 * \return	angle in radians
 ****************************************************************************/
double deg2rad(double deg)
{
  return (deg / 180.0 * M_PI);
}


/****************************************************************************/
/**
 * Convert degrees to degrees string
 * for latitude and longitude.
 *
 * \param	lat		latitude in degrees
 * \param	lon		longitude in degrees
 * \param	lat_d_str	string buffer for returning latitude
 *				   in degrees
 * \param	lon_d_str	string buffer for returning longitude
 *				   in degrees
 *
 ****************************************************************************/
void deg_to_d_str(double lat, double lon, char *lat_d_str, char *lon_d_str)
{
  double	tmp;		/* temp */
  UInt16	ideg;		/* integer degrees */
  
  /* Latitude */
  if (lat < 0) {
    StrCopy(lat_d_str, "-");
    lat = -lat;
  } else {
    StrCopy(lat_d_str, "");
  }

  tmp  = lat;
  ideg = (UInt16)tmp;
  tmp  = (tmp-ideg) * 10000000.0 + 0.5;	/* 1/10000000 degrees of arc */
  
  if (tmp >= 9999999.9) {
    tmp = 0;
    ideg++;
  }
  
  StrPrintF(&lat_d_str[StrLen(lat_d_str)], "%02u.%07lu", 
    ideg, (UInt32)tmp);
  
  /* Longitude */
  if (lon < 0) {
    StrCopy(lon_d_str, "-");
    lon = -lon;
  } else {
    StrCopy(lon_d_str, "");
  }

  tmp  = lon;
  ideg = (UInt16)tmp;
  tmp  = (tmp-ideg) * 10000000.0 + 0.5;	/* 1/10000000 minutes of arc */
  
  if (tmp >= 9999999.9) {
    tmp = 0;
    ideg++;
  }
  
  StrPrintF(&lon_d_str[StrLen(lon_d_str)], "%03u.%07lu", 
    ideg, (UInt32)tmp);

}


/****************************************************************************/
/**
 * Convert degrees to degrees and minutes string
 * for latitude and longitude.
 *
 * \param	lat		latitude in degrees
 * \param	lon		longitude in degrees
 * \param	lat_dm_str	string buffer for returning latitude
 *				   in degrees and minutes
 * \param	lon_dm_str	string buffer for returning longitude
 *				   in degrees and minutes
 *
 ****************************************************************************/
void deg_to_dm_str(double lat, double lon, char *lat_dm_str, char *lon_dm_str)
{
  double	tmp;		/* temp */
  UInt16	ideg;		/* integer degrees */
  UInt16	imin;		/* integer minutes */
  
  /* Latitude */
  if (lat < 0) {
    StrCopy(lat_dm_str, "-");
    lat = -lat;
  } else {
    StrCopy(lat_dm_str, "");
  }

  tmp  = lat;
  ideg = (UInt16)tmp;
  tmp  = (tmp-ideg) * 60.0;		/* minutes of arc  */
  imin = (UInt16)tmp;
  tmp  = (tmp-imin) * 10000.0 + 0.5;	/* 1/10000 minutes of arc */
  
  if (tmp >= 9999.9) {
    tmp = 0;
    imin++;
  }
  if (imin > 59) {
    imin = 0;
    ideg++;
  }
  
  StrPrintF(&lat_dm_str[StrLen(lat_dm_str)], "%02u%02u.%04u'", 
    ideg, imin, (UInt16)tmp);
  
  /* Longitude */
  if (lon < 0) {
    StrCopy(lon_dm_str, "-");
    lon = -lon;
  } else {
    StrCopy(lon_dm_str, "");
  }

  tmp  = lon;
  ideg = (UInt16)tmp;
  tmp  = (tmp-ideg) * 60.0;		/* minutes of arc  */
  imin = (UInt16)tmp;
  tmp  = (tmp-imin) * 10000.0 + 0.5;	/* 1/10000 minutes of arc */
  
  if (tmp >= 9999.9) {
    tmp = 0;
    imin++;
  }
  if (imin > 59) {
    imin = 0;
    ideg++;
  }
  
  StrPrintF(&lon_dm_str[StrLen(lon_dm_str)], "%03u%02u.%04u'", 
    ideg, imin, (UInt16)tmp);

}


/****************************************************************************/
/**
 * Convert degrees to degrees, minutes and seconds string
 * for latitude and longitude.
 *
 * \param	lat		latitude in degrees
 * \param	lon		longitude in degrees
 * \param	lat_dms_str	string buffer for returning latitude in
 *				   degrees, minutes and seconds
 * \param	lon_dms_str	string buffer for returning longitude in
 *				   degrees, minutes and seconds
 *
 ****************************************************************************/
void deg_to_dms_str(double lat, double lon, char *lat_dms_str, char *lon_dms_str)
{
  double	tmp;		/* temp */
  UInt16	ideg;		/* integer degrees */
  UInt16	imin;		/* integer minutes */
  UInt16	isec;		/* integer seconds */
  
  /* Latitude */
  if (lat < 0) {
    StrCopy(lat_dms_str, "-");
    lat = -lat;
  } else {
    StrCopy(lat_dms_str, "");
  }

  tmp  = lat;
  ideg = (UInt16)tmp;
  tmp  = (tmp-ideg) * 60.0;		/* minutes of arc  */
  imin = (UInt16)tmp;
  tmp  = (tmp-imin) * 60.0;		/* seconds of arc */
  isec = (UInt16)tmp;
  tmp  = (tmp-isec)*1000.0 + 0.5;	/* 1/1000 seconds of arc */
  
  if (tmp >= 999.9) {
    tmp = 0;
    isec++;
  }
  if (isec > 59) {
    isec = 0;
    imin++;
  }
  if (imin > 59) {
    imin = 0;
    ideg++;
  }
  
  StrPrintF(&lat_dms_str[StrLen(lat_dms_str)], "%02u%02u'%02u.%03u\"", 
    ideg, imin, isec, (UInt16)tmp);
  
  /* Longitude */
  if (lon < 0) {
    StrCopy(lon_dms_str, "-");
    lon = -lon;
  } else {
    StrCopy(lon_dms_str, "");
  }
  
  tmp  = lon;
  ideg = (UInt16)tmp;
  tmp  = (tmp-ideg) * 60.0;		/* minutes of arc  */
  imin = (UInt16)tmp;
  tmp  = (tmp-imin) * 60.0;		/* seconds of arc */
  isec = (UInt16)tmp;
  tmp  = (tmp-isec)*1000.0 + 0.5;	/* 1/1000 seconds of arc */
  
  if (tmp >= 999.9) {
    tmp = 0;
    isec++;
  }
  if (isec > 59) {
    isec = 0;
    imin++;
  }
  if (imin > 59) {
    imin = 0;
    ideg++;
  }

  StrPrintF(&lon_dms_str[StrLen(lon_dms_str)], "%03u%02u'%02u.%03u\"", 
    ideg, imin, isec, (UInt16)tmp);

}


/****************************************************************************/
/**
 * Convert degrees to degrees and minutes string
 * for latitude and longitude as required by GeoDB.
 *
 * \param	lat		latitude in degrees
 * \param	lon		longitude in degrees
 * \param	lat_str		string buffer for returning latitude
 *				   in degrees, minutes and hemisphere
 * \param	lon_str		string buffer for returning longitude
 *				   in degrees, minutes and hemisphere
 *
 * \note	Mostly identical to deg_to_dm_str()
 ****************************************************************************/
void deg_to_geodb_str(double lat, double lon, char *lat_str, char *lon_str)
{
  double	tmp;		/* temp */
  UInt16	ideg;		/* integer degrees */
  UInt16	imin;		/* integer minutes */
  Char		ns;		/* N/S */
  Char		ew;		/* E/W */
  
  /* Latitude */
  if (lat < 0) {
    ns = 'S';
    lat = -lat;
  } else {
    ns = 'N';
  }

  tmp  = lat;
  ideg = (UInt16)tmp;
  tmp  = (tmp-ideg) * 60.0;		/* minutes of arc  */
  imin = (UInt16)tmp;
  tmp  = (tmp-imin) * 10000.0 + 0.5;	/* 1/10000 minutes of arc */
  
  if (tmp >= 9999.9) {
    tmp = 0;
    imin++;
  }
  if (imin > 59) {
    imin = 0;
    ideg++;
  }
  
  StrPrintF(lat_str, "%02u:%02u.%04u,%c", ideg, imin, (UInt16)tmp, ns);
  
  /* Longitude */
  if (lon < 0) {
    ew = 'W';
    lon = -lon;
  } else {
    ew = 'E';
  }

  tmp  = lon;
  ideg = (UInt16)tmp;
  tmp  = (tmp-ideg) * 60.0;		/* minutes of arc  */
  imin = (UInt16)tmp;
  tmp  = (tmp-imin) * 10000.0 + 0.5;	/* 1/10000 minutes of arc */
  
  if (tmp >= 9999.9) {
    tmp = 0;
    imin++;
  }
  if (imin > 59) {
    imin = 0;
    ideg++;
  }
  
  StrPrintF(lon_str, "%03u:%02u.%04u,%c", ideg, imin, (UInt16)tmp, ew);
}


/****************************************************************************/
/**
 * \brief	Convert position to string according to selected position unit
 *
 * \param	lat		Latitude
 * \param	lon		Longitude
 * \param	pos_unit	position unit <POS_D|POS_DM|POS_DMS|POS_UTM>
 * \param	lat_strP	ptr to Latitude/Easting String buffer
 *				(>=20 bytes)
 * \param	lon_strP	ptr to Longitude/Northing String buffer
 *				(>=20 bytes)
 * \param	field_strP	ptr to UTM field String buffer
 *				If pos_unit != POS_UTM, an empty string is
 *				returned. If conversion to UTM failed, "n.a."
 *				is returned. 
 *				(>=5 bytes)
 * \param	eastP		UTM Easting returned by reference
 *				(pass NULL if not needed)
 * \param	northP		UTM Northing returned by reference
 *				(pass NULL if not needed)
 *
 * \return	Conversion to UTM successful. Always true, if any
 *		position format other than UTM is set in the preferences.
 *
 * \note	eastP and northP are updated only if pos_unit == POS_UTM and
 *		conversion to UTM was successful
 ****************************************************************************/
Boolean deg_to_str(double lat, double lon, UInt8 pos_unit,
                   char *lat_strP, char *lon_strP, char *field_strP,
		   double *eastP, double *northP)
{
  Int32		err;		/* error code */
  Int32		zone;		/* UTM Zone */
  char		hemisphere;	/* 'N' / 'S' */
  double	easting;	/* Easting in m */
  double	northing;	/* Northing in m */
  
  field_strP[0] = '\0';
  
  switch (pos_unit) {
    case POS_D:
      /* Degrees */
      deg_to_d_str(lat, lon, lat_strP, lon_strP);
      break;
    
    case POS_DM:
      deg_to_dm_str(lat, lon, lat_strP, lon_strP);
      break;

    case POS_DMS:
      deg_to_dms_str(lat, lon, lat_strP, lon_strP);
      break;
      
   case POS_UTM:   
      /* Universal Transverse Mercator - convert from lat/lon double values */ 
      err = Convert_Geodetic_To_UTM (
      				deg2rad(lat) /* Latitude (rad) */,
                                deg2rad(lon) /* Longitude (rad) */,
                                &zone /* long  *Zone */,
                                &hemisphere /* Hemisphere */,
                                &easting /* *Easting */,
                                &northing /* *Northing */);
      
      if (err == UTM_NO_ERROR) {
        /* print zone and belt to string */
	StrPrintF(field_strP, "%ld%c", zone, Lat_To_UTM_Belt(lat));
	
	/* print easting to string */
	format_number(easting, 1, lat_strP);
		
	/* print northing to string */
	format_number(northing, 1, lon_strP);
        
	/* return easting and northing if pointers are not NULL */
	if (eastP && northP) {
	  *eastP = easting;
	  *northP = northing;
	}
      } else {
        StrCopy(field_strP, "n.a.");
	lat_strP[0] = '\0';
	lon_strP[0] = '\0';
        return false;
      }
      break;   
  } /* switch (pos_unit) */
  
  return true;
}


/****************************************************************************/
/**
 * Calculate Great Circle distance on spheroid of two points
 *   specified by latitude and longitude in degrees.
 *
 *		Source: Aviation Formulary V1.41
 *		        By Ed Williams
 *			http://williams.best.vwh.net/avform.htm
 *
 * \param	lat1	latitude  of point 1
 * \param	lon1	longitude of point 1
 * \param	lat2	latitude  of point 2
 * \param	lon2	longitude of point 2
 *
 * \return	distance in radians
 ****************************************************************************/
double gc_dist_sphere(double lat1, double lon1, double lat2, double lon2)
{
  double	dist;		/* distance in radians */
  
  lat1 = deg2rad(lat1);
  lon1 = deg2rad(lon1);
  lat2 = deg2rad(lat2);
  lon2 = deg2rad(lon2);
  
  
  dist = 2*asin(sqrt(SQR(sin((lat1-lat2)/2)) + 
                 cos(lat1)*cos(lat2)*SQR(sin((lon1-lon2)/2))));

  return dist;
}


/****************************************************************************/
/**
 * Calculate initial Great Circle course
 * between two points on spheroid
 * specified by latitude and longitude in degrees.
 *
 * As opposed to the source (see below), the following
 * convention is followed:
 * - East longitude: positive
 * - West longitude: negative
 *
 *		Source: Aviation Formulary V1.41
 *		        By Ed Williams
 *			http://williams.best.vwh.net/avform.htm
 *
 * \param	lat1	latitude  of point 1
 * \param	lon1	longitude of point 1
 * \param	lat2	latitude  of point 2
 * \param	lon2	longitude of point 2
 *
 * \return	initial course in radians
 ****************************************************************************/
double gc_course_sphere(double lat1, double lon1, double lat2, double lon2)
{
  double tc1;
  
  lat1 = deg2rad(lat1);
  lon1 = deg2rad(lon1);
  lat2 = deg2rad(lat2);
  lon2 = deg2rad(lon2);

  /*
   * We obtain the initial course, tc1, (at point 1)
   * from point 1 to point 2 by the following.
   */
   
  if (cos(lat1) < EPS) {  /* EPS a small number ~ machine precision */
    /*
     * The formula fails if the initial point is a pole.
     * We can special case this with:
     */
    if (lat1 > 0) {
      tc1 = M_PI;		/*  starting from N pole */
    } else {
      tc1 = 2*M_PI;		/*  starting from S pole */
    }
  } else {
    /* For starting points other than the poles: */
    tc1 = fmod(atan2(sin(lon2-lon1)*cos(lat2),
           cos(lat1)*sin(lat2)-sin(lat1)*cos(lat2)*cos(lon2-lon1)), 2*M_PI);
    
    tc1 = (tc1 < 0) ? tc1+2*M_PI : tc1;
  }
  return tc1;
}
