/*****************************************************************************
 *
 * $RCSfile: georef.c,v $
 *
 * GPS4Palm Map File Georeference Decoding
 *
 * History:
 * 2003-10-19:	first version
 *
 * Reads Georeference Parameters
 *   - currently the parameters SCALE, LAT and LON are encoded in the filename
 *   - the next step should be to get these parameters from the GIF87a
 *     Comment Extension or PNG tEXt chunks, respectively
 *   - the final implementation should use an appropriate file format like
 *     GeoTiff, which contains both image data and georeferencing meta data
 *
 *
 * This program is Copyright (C) 10/2003 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-05-15 12:36:04 $
 *
 * $Revision: 1.9 $
 *
 * $Log: georef.c,v $
 * Revision 1.9  2005-05-15 12:36:04  mp
 * added map size to map search conditions
 *
 * Revision 1.8  2005/05/06 13:49:26  mp
 * modified documentation
 *
 * Revision 1.7  2005/05/06 13:42:40  mp
 * removed call to WinGetDisplayExtent() and supporting code
 *
 * Revision 1.6  2005/04/17 18:32:58  mp
 * added High-Density Display support to search_map()
 *
 * Revision 1.5  2005/02/19 14:39:17  mp
 * modified VFS support to search_map(), modified comments for doxygen
 *
 * Revision 1.4  2003/12/30 14:47:52  mp
 * filename cleared at start of search_map, removed debugging code
 *
 * Revision 1.3  2003/12/28 18:08:38  mp
 * fixed order of arguments to calcxy()
 *
 * Revision 1.2  2003/11/23 20:00:44  mp
 * added search_map()
 *
 * Revision 1.1  2003/10/19 16:56:30  mp
 * initial version
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "stringil.h"
#include "MapForm.h"
#include "georef.h"
#include "geo.h"
#include "file_io.h"
#include "getdir.h"
#include "imginfo.h"
#include "common.h"

/* Preferences data structure */
extern PrefsType gPrefs;

/* VFS Manager Feature Set */
extern Boolean   gVfsManagerFtr;



/**********************************************************************/
/**
 * \brief	Get the map parameters SCALE, LAT and LON from the filename
 *
 * \param	filename	filename
 * \param	map_lat		Ptr to Latitude
 * \param	map_lon		Ptr to Longitude
 * \param	map_scale	Ptr to Scale
 *
 * \return	true if all parameters decoded successfully, otherwise false
 *		map_lat, map_lon, and map_scale
 *
 ****************************************************************************/
Boolean georef_filename(const Char *filename, double *map_lat, double *map_lon,
                 UInt32 *map_scale)
{
  UInt32	scale = 0;		/* Scale */
  double	lat;			/* Latitude */
  double	lon;			/* Longitude */
  FlpCompDouble	dval;
  Char		tmp[12];
  UInt16	i;			/* loop index */
  UInt16	j;			/* loop index */
  Boolean	valid = false;		/* decoding valid flag */
  
  /*
   * Decode scale information
   *   Scale information ends with 'k', 'M', '+' or '-'
   *   and can contain at most 3 digits.
   */
   
   for (i=0; i<5; i++) {
     if (filename[i] == 'k') {
       /* multiply by 10^3 */
       scale *= 1000;
       i++;
       valid = true;
       break;
     }
     if (filename[i] == 'M') {
       /* multiply by 10^6 */
       scale *= 1000000L;
       i++;
       valid = true;
       break;
     }
     if ((filename[i] == '+') || (filename[i] == '-')) {
       valid = true;
       break;
     }
     if ((filename[i] >= '0') && (filename[i] <= '9')) {
       /* next digit */
       scale *= 10;
       scale += filename[i] - '0';
     }
   }
   
   if (!valid) {
     /* Decoding scale info failed */
     return false;
   }
   

   /*
    * Decode Latitude
    *   Latitude information ends with '+' or '-' and consists of at most
    *   11 characters.
    *   Example: +52.1614999
    */
   if (filename[i] == '+') {
     lat = 1.0;
   } else if (filename[i] == '-') {
     lat = -1.0;
   } else {
     /* Lat. decoding failed - no sign found */ 
     return false;
   }

   /* sign done */
   i++;
   
   /* copy lat to temp. string */
   for (j=0; j < 10; j++) {
     if ((filename[i] == '+') || (filename[i] == '-'))
       break;
     tmp[j] = filename[i++];
   }
   
   if ((filename[i] == '+') || (filename[i] == '-')) {
     /* valid end (sign of lon.) found, terminate temp. string */
     tmp[j] = '\0';
     
     /* get floating point value from string */
     FlpBufferAToF(&dval.fd, tmp);
   
     /* store "double"-field of union FlpCompDouble in variable lat */
     lat = lat * dval.d;
   } else {
     /* decoding failed */
     return false;
   }
   

   /*
    * Decode Longitude
    *   Longitude information ends with '.' and consists of 5 to 11
    *   characters.
    *   Example: -110.3126601
    */
   if (filename[i] == '+') {
     lon = 1.0;
   } else if (filename[i] == '-') {
     lon = -1.0;
   } else {
     /* Lon. decoding failed - no sign found */ 
     return false;
   }
   
   /* sign done */
   i++;
   
   /* copy lon to temp. string */
   for (j=0; j < 11; j++) {
     if ((j > 4) && (filename[i] == '.'))
       break;
     tmp[j] = filename[i++];
   }
   
   if (filename[i] == '.') {
     /* valid end (filename extension) found, terminate temp. string */
     tmp[j] = '\0';
     
     /* get floating point value from string */
     FlpBufferAToF(&dval.fd, tmp);
   
     /* store "double"-field of union FlpCompDouble in variable lat */
     lon = lon * dval.d;
   } else {
     /* decoding failed */
     return false;
   }

   /* return internal values */ 
   *map_scale	= scale;
   *map_lat	= lat;
   *map_lon	= lon;
   
   /* decoding successful */
   return true;
}




static void copy_of_cvt_ultoa(unsigned long,char *);
/*****************************************************************************
 * FUNCTION:	cvt_ultoa
 *
 * DESCRIPTION:	Convert unsigned long value to string.
 *
 * PARAMETERS:	value	-- unsigned long value
 *		strP	-- ptr to string buffer for returning result
 *
 * RETURNED:	%
 *****************************************************************************/
static void copy_of_cvt_ultoa(unsigned long value, char *strP)
{
  int i;
  char *cP;
  char tmpstr[16];		/* a 32 bit long can give up to 10 digits */

    if( value <= 9 )		/* a quick special case */
    {
	*strP++ = value + '0';
	*strP = '\0';
	return;
    }

    /* We convert the number backwards */
    for( cP = tmpstr+9, i = 10; i-- > 0; )
    {
	*cP-- = (value % 10) + '0';
	value /= 10;
    }

    tmpstr[10] = '\0';

    for( cP = tmpstr; *cP && (*cP == '0'); )
	++cP;
    
    StrCopy(strP,cP);
}


Boolean search_map_all_dirs(char *filename, UInt32 *card, double lon, double lat,
			    UInt16 maxx, UInt16 maxy) {
  Boolean map_found = false;
  Int16 i;
  
  StrCopy(FILEDIR, FILEDIRpre);
  if (search_map(filename, card, lon, lat, maxx, maxy)) {
    map_found = true;
  } else {
    char scaleStr[16];
    Int16 numpos;
    
    copy_of_cvt_ultoa((unsigned long) gPrefs.mapprefs.scale, scaleStr);
    StrCopy(FILEDIR, FILEDIRpre);
    StrCat(FILEDIR, "/");
    StrCat(FILEDIR, scaleStr);
    StrCat(FILEDIR, ".x");
    numpos = StrLen(FILEDIR) - 1;
    
    for (i = 0; i < 10; i++) {
      FILEDIR[numpos] = i + 48;
      if (search_map(filename, card, lon, lat, maxx, maxy)) {
	map_found = true;
	break;
      }
    }
  }
  //  StrCopy(FILEDIR, FILEDIRpre);

  return map_found;
}



/**********************************************************************/
/**
 * \brief	Search for usable map
 *
 * \note	Currently search_map stops on the first map which fits, even
 *		if there is a better map (i.e. with the map center closer to
 *		the current position) available. 
 *
 * \param	filename	filename or NULL (returned by reference)
 * \param	card		card/volume number and VFS flag
 *                                (returned by reference)
 * \param	lat		Latitude of current position
 * \param	lon		Longitude of current position
 * \param	maxx		Max. x-extent of screen
 * \param	maxy		Max. y-extent of screen
 *
 * \return	true if usable map found, otherwise false
 ****************************************************************************/
Boolean search_map(char *filename, UInt32 *card, double lon, double lat,
                   UInt16 maxx, UInt16 maxy)
{
  static DmOpenRef	scratch;		/* scratchpad DB reference */
  static char		*filelist;		/* filelist DB record */
  static char		*filecard;		/* filecard DB record */
  static char		**filelp;		/* filelp DB record */
  Err			err;			/* error code */
  FileDescrType		filedescr;		/* file descriptor */
  UInt16		i;			/* loop index */
  UInt16		files = 0;		/* number of files found */
  Char			**f;			/* filename */
  double		map_lat;		/* map center latitude */
  double		map_lon;		/* map center longitude */
  UInt32		map_scale;		/* map scale */
  Int32			x;			/* map xoord. x */
  Int32			y;			/* map coord. y */
  Int32			w;			/* map width/2 */
  Int32			h;			/* map height/2 */
  UInt16		width;			/* map width */
  UInt16		height;			/* map height */
  UInt8			type;			/* map type */
  Boolean		found = false;		/* map found */

  *filename = '\0';
  *card = 0;

  /* get directory */
  CreateDatabase(&scratch, &filelist, &filecard, &filelp);
  files = getdir(filelist, filecard, filelp);

  if (gVfsManagerFtr) {
    files = getdir_vfs(filelist, filecard, filelp, files);
  }
  
  if (files == 0) {
    /* no map found */    
    /* free directory storage */
    DestroyDatabase(scratch, filelist, filecard, filelp);

    return false;
  }
  
  /* file list pointer */
  f = filelp;
  
  /* search for usable map */
  for (i = 0; i<files; i++, f++) {
    
    /* get map parameters from filename, skip invalid files */
    if (georef_filename(*f+SORTPREFIX, &map_lat, &map_lon, &map_scale)) {
      
      /* check if current map has reqired scale */
      if (map_scale == gPrefs.mapprefs.scale) {
      
        /* calculate offset between position and map center */
        calcxy(&x, &y, lon, lat, map_lon, map_lat, map_scale);
        
	/* copy card/volume number and VFS flag */
	memcpy(card, &filecard[i*4], 4);
	
	/* get map's dimensions */
        filedescr = fopen(*card, *f+SORTPREFIX, FILEDIR, &err);
    
	if (!err) {
	  if (imginfo(filedescr, &type, &width, &height)) {
            /* no valid info */
	    fclose(filedescr);
	    continue;
	  }
	  fclose(filedescr);
	} else {
	  width  = 0;
	  height = 0;
	}
	
	/* offset from map center must be smaller than half image */
	w = width/2;
	h = height/2;
	
        /* check if current position is on map */
        if ( (x > -w+CROSS_X) &&
	     (x <  w-CROSS_X) &&
	     (y > -h+CROSS_Y) &&
	     (y <  h-CROSS_Y) ) {
	  StrCopy(filename, *f+SORTPREFIX);
	  found = true;
	  break;
	}
      } /* if (scale == gPrefs.mapprefs.scale) */
    } /* if (georef_filename()) */
    
  } /* for (i = 0; i<files; i++) */
  
  /* free directory storage */
  DestroyDatabase(scratch, filelist, filecard, filelp);
  
  return found;
}
