/*****************************************************************************
 *
 * $RCSfile: getdir.c,v $
 *
 *
 *   Store "directory" of a set of databases / files
 *   in 3 different databases (see parameters).
 *
 * History:
 * 2002-12-23:	first version, taken from the ZBoxZ application suite
 *		  (Rev. 0.30) by Tom Zerucha
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2005-02-26 14:50:23 $
 *
 * $Revision: 1.5 $
 *
 * $Log: getdir.c,v $
 * Revision 1.5  2005-02-26 14:50:23  mp
 * removed debug output
 *
 * Revision 1.4  2005/02/19 14:34:33  mp
 * removed #include <Unix/sys_types.h> (now included in stringil.h)
 *
 * Revision 1.3  2005/02/19 11:15:11  mp
 * added getdir_vfs(), modified name/creator of scratchpad DB
 *
 * Revision 1.2  2003/11/23 19:57:30  mp
 * moved CreateDatabase and DestroyDatabase from FileselForm.c
 *
 * Revision 1.1  2003/10/18 16:25:21  mp
 * imported from piv
 *
 *
 *
 ****************************************************************************/

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <Extensions/ExpansionMgr/VFSMgr.h>
#include "stringil.h"
#include "getdir.h"
#include "common.h"


/* Preferences data structure */
extern PrefsType gPrefs;

//#define VERBOSE


/****************************************************************************/
/**
 * \brief	Store "directory" of a set of databases (i.e. files)
 *		in 3 different database records.
 * <pre>
 *		filelist	< size0 (4 bytes) >< filename0 >
 *				< size1 (4 bytes) >< filename1 >
 *				...
 *		filecard	< card0 > (card of 1st file)
 *				< card1 > (card of 2nd file)
 *				...
 *		filelp		< ptr0 > (pointer to 1st filelist entry)
 *				< ptr1 > (pointer to 2nd filelist entry)
 *				...
 * </pre>
 * \param	filelist	start of filelist record
 * \param	filecard	filecard record
 * \param	filelp		filelp record
 *
 * \return	Total number of files.
 ****************************************************************************/
UInt16 getdir(char *filelist, char *filecard, char **filelp)
{
  UInt16	cardno;	/* card no. */
  UInt32	card;	/* card no. and File Streaming Flag */
  UInt16	attr;	/* attributes */
  UInt32	type;	/* type ID */
  UInt32	crea;	/* creator ID */
  UInt32	ndb;	/* number of databases on current card */
  UInt32	size;	/* size (payload) */
  UInt32	i;	/* index (database, list) */
  UInt32	j;	/* offset within record */
  UInt32	tmp;	/* start position of file suffix */ 
  UInt32	*flpt;	/* file list pointer */
  UInt32	*fl;	/* file list pointer */
  UInt32	disp;	/* displacement in filelist */
  UInt16	maxrow;	/* total number of rows (files) */
  UInt16	n;	/* y-position of WinDrawChars() */
  Char		tp[44];	/* temporary filelist buffer */
  LocalID	lid;	/* local database ID */

  maxrow = 0;
  disp = 0;
  n = 15;

#ifdef VERBOSE  
  WinDrawChars("Getting Directory...", 20, 0, 0);
#endif
  
  /* search files on all cards */
  for (cardno = 0; cardno < MemNumCards(); cardno++) {
    ndb = DmNumDatabases(cardno);    
    j = maxrow;
    fl = flpt = MemPtrNew(ndb * 4);
    
    /* search all files on current card */
    for (i = 0; i < ndb; i++) {
      memset(tp, 0, SORTPREFIX); /* for robustness, if sortprefix is >4, and the sort opt is smaller */
      lid = DmGetDatabase(cardno, i);
      DmDatabaseInfo(cardno, lid, &tp[SORTPREFIX], &attr, NULL, NULL, NULL,
                     NULL, NULL, NULL, NULL, &type, &crea);
      /* check type and attributes: only file streams of type data are valid */
      if (!(attr & dmHdrAttrStream) || type != 'DATA')
        continue;
	
      /* check creator ID: only 'BOXR' is valid */
      if (crea != 'BOXR')
        continue;
	
      /* check filename suffix: only GIF and PNG valid */
      tmp = StrLen(&tp[SORTPREFIX]) - 4;
      if ( (StrCaselessCompare(&tp[SORTPREFIX + tmp], ".gif") != 0) &&
           (StrCaselessCompare(&tp[SORTPREFIX + tmp], ".png") != 0) )
	continue;

#ifdef VERBOSE
      WinDrawChars(&tp[SORTPREFIX], StrLen(&tp[SORTPREFIX]), 0, n);
      n += 12;
#endif 
     
      /*
       * Copy type, creator, size, i, backup/creation/modification date,
       * or nrecs into tp (4 bytes) for alternate sorting
       * (only using payload size here)
       */
      DmDatabaseSize(cardno, lid, NULL, NULL, &size);
      memcpy(tp, &size, 4);
#ifdef VERBOSE
      {
        char t[10];
	StrIToA(t, *tp);
        WinDrawChars(t, StrLen(t), 0, n);
        n += 12;
      }
#endif 

      /*
       * Write file list DB 
       *   Each entry consists of a <size><filename> pair.
       *   While the <size> entry always has 4 bytes, the
       *   filename has a variable length.
       */
      DmWrite(	filelist,	/* recordP: file list */
      		disp,		/* offset: see below */
		tp,		/* srcP: pointer to <size><name> entry */
		strlen(&tp[SORTPREFIX]) + SORTPREFIX + 1 /* bytes */
	     );

      card = cardno | FS_FLAG;
      
      /* write card number to filecard */     
      DmWrite(filecard,		/* recordP: filecard */
              maxrow * 4,	/* offset */
	      &card,		/* srcP */ 
	      4			/* bytes */
	     );

      /* fl: array of pointers to file list entries */
      *fl++ = (UInt32) filelist + disp;
      
      /* calculate offset (displacement) for next file list entry */
      disp += strlen(&tp[SORTPREFIX]) + SORTPREFIX + 1;
      
      /* increment total number of rows (= files) on all cards */ 
      maxrow++;
    } /* for (i = 0; i < ndb; i++) */
    
    /*
     * Write file list pointer DB
     *  Each entry consists of a pointer to a file list entry (4 bytes).
     *  All entries for the current card are written as a block.
     */
    DmWrite(	filelp,	/* recordP: file list pointers */
    		j * 4,	/* offset: number of files on prev. cards x (4 bytes) */
		flpt,	/* srcP: array of file pointers fo current card */
		(maxrow - j) * 4); /* bytes: number of files on current card
		                               x (4 bytes) */
    
    /* Deallocate array of file pointers */
    MemPtrFree(flpt);
  } /* for (cardno = 0; cardno < MemNumCards(); cardno++) */

  return maxrow;
}


/****************************************************************************/
/**
 * \brief	Store directory on external memory cards (VFS)
 *		in 3 different database records.
 * <pre>
 *		filelist	< size0 (4 bytes) >< filename0 >
 *				< size1 (4 bytes) >< filename1 >
 *				...
 *		filecard	< vol0 > (volume of 1st file)
 *				< vol1 > (volume of 2nd file)
 *				...
 *		filelp		< ptr0 > (pointer to 1st filelist entry)
 *				< ptr1 > (pointer to 2nd filelist entry)
 *				...
 * </pre>
 * \param	filelist	start of filelist record
 * \param	filecard	filecard record
 * \param	filelp		filelp record
 * \param	files		number of files already found
 *                                (File Streaming API)
 *
 * \return	Total number of files.
 ****************************************************************************/
 UInt16 getdir_vfs(char *filelist, char *filecard, char **filelp, UInt16 files)
{
  UInt16	volRefNum;	/* VFS volume reference number */
  UInt32	volIterator;	/* VFS volume iterator */
  FileRef	dirRef;		/* directory reference */
  FileRef	fileRef;	/* file reference */
  UInt32	card;		/* card/volume number */
  UInt32	size;		/* file size */
  UInt16	file_cnt;	/* file counter */
  UInt32	flp;		/* ptr to filelist entry (typecasted) */
  UInt32	tmp;		/* start position of file suffix */ 
  UInt32	disp;		/* displacement in filelist */
  UInt16	n;		/* y-position of WinDrawChars() */
  Err		err;		/* error code */
  
  n = 15;
  file_cnt = files;

  /*
  if (gPrefs.mapprefs.scale == 8500) {
    StrCopy(FILEDIR, FILEDIRpre);
    StrCat(FILEDIR, "/8500");
  } else if (gPrefs.mapprefs.scale == 18000) {
    StrCopy(FILEDIR, FILEDIRpre);
    StrCat(FILEDIR, "/18k");
  } else {
    StrCopy(FILEDIR, FILEDIRpre);
  }
  */

  
  /* find end of records */
  if (file_cnt > 0) {
    disp = (UInt32)(filelp[files-1]) - (UInt32)*filelp ;
    disp += StrLen(filelp[files-1] + SORTPREFIX) + SORTPREFIX + 1;
    flp = (UInt32)filelist + disp;
  } else {
    disp     = 0;
    flp = (UInt32)filelist;
  }

#ifdef VERBOSE
  {
    Char tmp[40];
    
    StrCopy(tmp, "Getting VFS Directory...");
    WinDrawChars(tmp, StrLen(tmp), 0, 0);
  }
#endif
  
  volIterator = vfsIteratorStart;

  /* Iterate through all VFS volumes */
  while (volIterator != vfsIteratorStop) {
    err = VFSVolumeEnumerate(&volRefNum, &volIterator);
    
    if (err == errNone) {

      /* Open the directory and iterate through the files in it. */
      err = VFSFileOpen(volRefNum, FILEDIR, vfsModeRead, &dirRef);
  
      if (err == errNone) {
	UInt32		fileIterator;
	FileInfoType	fileInfo;
	Char		fileName[FILENAMEMAXLEN];
        MemPtr		pathName;

	fileInfo.nameP = fileName;
	fileInfo.nameBufLen = sizeof(fileName);
	fileIterator = expIteratorStart;
    
	/* Iterate through all the files in the open directory */
	while (fileIterator != expIteratorStop) {
	  /* Get the next file */
	  err = VFSDirEntryEnumerate(dirRef, &fileIterator, &fileInfo);
	  
	  if (err == errNone) {
            /* Process the file */
	    
	    /* filter filename suffix */
	    tmp = StrLen(fileInfo.nameP) - 4;
            if ( (StrCaselessCompare(&fileInfo.nameP[tmp], ".gif") != 0) &&
                 (StrCaselessCompare(&fileInfo.nameP[tmp], ".png") != 0) )
	      continue;

#ifdef VERBOSE
	    WinDrawChars(fileInfo.nameP, StrLen(fileInfo.nameP), 0, n);
            n += 12;
#endif

            /* generate absolute file path */
	    //            pathName = MemPtrNew(sizeof(FILEDIR) + StrLen(fileInfo.nameP) + 2);
            pathName = MemPtrNew(StrLen(FILEDIR) + StrLen(fileInfo.nameP) + 2);
	    StrCopy(pathName, FILEDIR);
	    StrCat(pathName, "/");
	    StrCat(pathName, fileInfo.nameP);
	    
	    /* open file to get size */
            err = VFSFileOpen(volRefNum, pathName, vfsModeRead, &fileRef);
	    
	    if (!err) {
	      err = VFSFileSize(fileRef, &size);
	      
	      if (err) {
	        size = 0;
	      }
	      VFSFileClose(fileRef);
	    }
	    MemPtrFree(pathName);
	    
	    /* write file size to filelist */
            DmWrite(filelist,			/* recordP: file list */
      		    disp,			/* offset */
		    &size,			/* srcP: pointer to size */
		    sizeof(size)		/* bytes (4) */
	    );
            disp += sizeof(size);

	    /* write file name to filelist */
            DmWrite(filelist,			/* recordP: filelist */
      		    disp,			/* offset */
		    fileInfo.nameP,		/* srcP: pointer to file name */
		    StrLen(fileInfo.nameP) + 1 	/* bytes */
	    );
            disp += StrLen(fileInfo.nameP) + 1;
	    
	    /* volume/card reference number */
	    card = volRefNum | VFS_FLAG;

	    /* write volume reference number to filecard */
	    DmWrite(filecard,			/* recordP: filecard */
                    file_cnt * 4,		/* offset */ 
		    &card,			/* srcP */
		    4				/* bytes */
            );
	    
	    /* write pointer to filelist entry to filelp */
	    DmWrite(filelp,			/* recordP: filelist pointers */
    		    file_cnt * 4,		/* offset */
		    &flp,			/* srcP */
                    4				/* bytes */
	    );

            /* pointer to next file list entry */
            flp = (UInt32)filelist + disp;
	    
	    /* increment file count */
	    file_cnt++;
	    
	  } else {
            /* handle directory open error */
#ifdef VERBOSE
	    {
	      Char tmp[40];

	      StrCopy(tmp, "Cannot open file: ");
	      StrIToH(&tmp[StrLen(tmp)], err);
	      WinDrawChars(tmp, StrLen(tmp), 0, n);
              n += 12;
            }
#endif
	    break;
	  }
	  
	} /* while (fileIterator != expIteratorStop) */
    
        err = VFSFileClose(dirRef);


      } else {
        /* handle error... */

#ifdef VERBOSE
	{
	  Char tmp[40];
	
          StrCopy(tmp, "Cannot open directory: ");
	  StrIToH(&tmp[StrLen(tmp)], err);
	  WinDrawChars(tmp, StrLen(tmp), 0, n);
          n += 12;
        }
#endif
	break;
      }
    } 
  } /* while (volIterator != vfsIteratorStop) */
  
  return file_cnt;
}


/****************************************************************************/
/**
 * \brief	Create scratchpad database with 3 records.
 *
 *
 * \note	All parameters are returned 'by reference', because the memory
 *		locations of the database records are determined by this
 *		function.
 *
 * \param	ref		pointer to database reference
 * \param	filelist	pointer to start of filelist record
 * \param	filecard	pointer to filecard record
 * \param	filelp		pointer to filelp record
 *
 ****************************************************************************/
void CreateDatabase(DmOpenRef *ref,
			     char **filelist, char **filecard, char ***filelp)
{
  LocalID	lid;		/* local database ID */
  DmOpenRef	scratch;	/* scratchpad DB reference */
  MemHandle	hand;		/* DB record memory handle */
  UInt16	dbrec = 0;	/* index of new record */
  
  /* open/create scratchpad database */
  lid = DmFindDatabase(0, "GPS4PALMSCRATCHPAD");
  if (!lid) {
    DmCreateDatabase(0, "GPS4PALMSCRATCHPAD", kCreatorId, 'DATA', false);
    lid = DmFindDatabase(0, "GPS4PALMSCRATCHPAD");
  }
  scratch = DmOpenDatabase(0, lid, dmModeReadWrite);
  
  if (!scratch)
    ErrDisplay("Could not access scratchpad!");
  
  hand = DmNewRecord(scratch, &dbrec, FILELISTRECORDSIZE);
  *filelist = MemHandleLock(hand);
  hand = DmNewRecord(scratch, &dbrec, FILELPRECORDSIZE);
  *filelp = MemHandleLock(hand);
  hand = DmNewRecord(scratch, &dbrec, FILECARDRECORDSIZE);
  *filecard = MemHandleLock(hand);
 
  *ref = scratch;
}


/****************************************************************************/
/**
 * \brief	Remove all records and close scratchpad database
 *
 * \param	ref		database reference
 * \param	filelist	start of filelist record
 * \param	filecard	filecard record
 * \param	filelp		filelp record
 *
 ****************************************************************************/
void DestroyDatabase(DmOpenRef ref,
                              char *filelist, char *filecard, char **filelp)
{
  MemPtrUnlock(filelp);
  MemPtrUnlock(filecard);
  MemPtrUnlock(filelist);
  
  /* remove all records from scratchpad DB */
  while (DmNumRecords(ref))
    DmRemoveRecord(ref, 0);

  /* close scratchpad DB */
  DmCloseDatabase(ref);
}
