/*****************************************************************************
 *
 * $RCSfile: gifread.c,v $
 *
 * GIF Image Decoder
 *
 * created: 2002-12-15
 *
 * History:
 * 2007-07-11:	testing feature NOPROGRESSBAR
 * 2002-12-15:	imported from the program PiNGer which is
 *		  part of the ZBoxZ application suite (Rev. 0.30)
 *		  by Tom Zerucha.
 * 2002-12-29:  fixed cleanGIF() to work when called after early abort,
 *                added skipping of Extension Blocks (optional in GIF89a),
 *                added abort decoding (pen down event)
 * 2002-12-30:  added progress bar, modified status messages
 *
 * This program is Copyright (C) 12/2002 Matthias Prinke
 * <matthias.prinke@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * Changes for Palm TX are Copyright (C) 07/2007 Frank Saurbier
 * <frank.saurbier@surfeu.de> and covered by GNU's GPL.
 * In particular, this program is free software and comes WITHOUT
 * ANY WARRANTY.
 *
 * $Author: mp $
 *
 * $Date: 2007-10-05 22:06:17 $
 *
 * $Log: gifread.c,v $
 * Revision 1.10.2.1  2007-10-05 22:06:17  mp
 * added Frank's changes (NOPROGRESSBAR)
 *
 * $Author: mp $
 *
 * $Date: 2007-10-05 22:06:17 $
 *
 * $Revision: 1.10.2.1 $
 *
 * $Log: gifread.c,v $
 * Revision 1.10.2.1  2007-10-05 22:06:17  mp
 * added Frank's changes (NOPROGRESSBAR)
 *
 * Revision 1.11  2007/07/11 16:00:00  fs
 * added ifndef NOPROGRESSBAR for testing
 *
 * Revision 1.10  2005/05/15 10:46:02  mp
 * added passing of file descriptor (fd) to functions
 *
 * Revision 1.9  2005/05/14 12:28:09  mp
 * added parameter fd to fread()
 *
 * Revision 1.8  2005/05/13 18:29:40  mp
 * modified decoding abortion
 *
 * Revision 1.7  2004/11/27 10:16:26  mp
 * replaced ErrFatalDisplay() by Die() for memory allocation failures
 *
 * Revision 1.6  2004/11/24 21:07:25  mp
 * moved static function declarations from header to implementation file
 *
 * Revision 1.5  2004/11/23 17:45:36  mp
 * moved unused variable to debug specific code block
 *
 * Revision 1.4  2004/03/10 17:18:39  mp
 * added progressbar.h
 *
 * Revision 1.3  2003/12/27 12:08:42  mp
 * repeated calling of cleanGIF() now handled properly
 *
 * Revision 1.2  2003/10/18 16:11:51  mp
 * renamed ImgInfo to gPrefs
 *
 * Revision 1.1.1.1  2003/07/14 18:59:29  mp
 * Imported GPS4Palm to CVS revision control.
 *
 *
 ****************************************************************************/
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "stringil.h"
#include "gifread.h"
#include "file_io.h"		/* fread() */
#include "common.h"
#include "Utils.h"		/* Die() */
#ifndef NOPROGRESSBAR
#include "progressbar.h"
#endif

#define malloc MemPtrNew
#define free MemPtrFree

extern PrefsType	gPrefs;

Boolean color;
unsigned short w,h,bandh;
unsigned int mapmax;
RGBColorType map[256];

unsigned char tcmap[256];

DmOpenRef scratch;
char **lines, *lcur, *editln;

/***************************************************************************/

extern short int *lwzstack, lwzsp, *table;

int nextLWZ(FileDescrType fd);
void initLWZ(short int input_code_size);
#define readLWZ(fd) (lwzsp ? lwzstack[--lwzsp] : nextLWZ(fd))

#define USAT(a,b)                 (((b)<<8)|(a))

int ycnt;


/* Static Functions */
static void putline(FileDescrType fd, int ypos)			GIFREAD_SECTION;
static Boolean stop(void)					GIFREAD_SECTION;


static void putline(FileDescrType fd, int ypos)
{
  unsigned char *cp;
  int xpos;

  
#ifdef VERBOSE
  {
    char buf[10];
  
    StrIToA(buf,ycnt);
    strcat(buf,"    ");
    WinDrawChars(buf,strlen(buf),0,40);
  }
#endif
  
#ifndef NOPROGRESSBAR
  progressbar(80 /* ypos */, ycnt++ /* val */, h /* total */);
#endif
  
  cp = editln;
  xpos = w;
  while( xpos-- )
    *cp++ = tcmap[readLWZ(fd)];
  DmWrite(lines[ypos / bandh], w * (ypos % bandh), editln, w);
}

static Boolean stop(void)
{
  EventType event;
  
  EvtGetEvent(&event, 0);
  if (event.eType == penDownEvent && event.screenY < 160) {
    MemPtrFree(table);
    MemPtrFree(lwzstack);
    return 1;
  } else {
    return 0;
  }
}

/***************************************************************************/

int doGIF(FileDescrType	fd)
{
  unsigned char gbuf[24], c, *cp;
  int i, ypos, lace;
  int Bkgnd;
  LocalID lid;
  unsigned short dbrec;
  VoidHand hand;
  unsigned char cmap[3 * 256];

  lines = NULL;
  editln = NULL;
  scratch = 0;
  
  fread(fd, gbuf, 13);

  if (!strncmp(gbuf, "GIF87a", 6)) {
    strcpy(gPrefs.imgtype, "GIF87a");
  } else if (!strncmp(gbuf, "GIF89a", 6)) {
    strcpy(gPrefs.imgtype, "GIF89a");
  } else {
    return 1;
  }

  
  //  w = USAT(gbuf[6], gbuf[7]);
  //  h = USAT(gbuf[8], gbuf[9]);
  // [12] is aspect ratio
  mapmax = 2 << (gbuf[10] & 0x07);
  
  if (gbuf[10] & 0x80) {         /* Global CMap */
    fread(fd, cmap, mapmax * 3);
    Bkgnd = gbuf[11];
  } else {
    /* if no color maps are present default should have black and white as the first two colors. */
    static int colors[] = { 0, 7, 1, 2, 4, 3, 5, 6 };
    for (i = 0; i < 256 - 8; i++) {
      cmap[i * 3] = (colors[i & 7] & 1) ? ((255 - i) & 0xf8) : 0;
      cmap[i * 3 + 1] = (colors[i & 7] & 2) ? ((255 - i) & 0xf8) : 0;
      cmap[i * 3 + 2] = (colors[i & 7] & 4) ? ((255 - i) & 0xf8) : 0;
    }
    for (; i < 256; i++) {
      cmap[i * 3] = (colors[i & 7] & 1) ? 4 : 0;
      cmap[i * 3 + 1] = (colors[i & 7] & 2) ? 4 : 0;
      cmap[i * 3 + 2] = (colors[i & 7] & 4) ? 4 : 0;
    }
    Bkgnd = -1;                 /* background unspecified */
  }
    
  /* skip optional extension blocks (GIF89a) */
  while (true) {
    if (fread(fd, gbuf, 1) == 0) {
      /* error: end of file */
      return 2;
    }
    if (gbuf[0] == ',') {
      /* Image Separator found */
      break;
    } else if (gbuf[0] == '!') {
      /* Extension Introducer found */
      fread(fd, gbuf, 1);
      if (gbuf[0] == 0xFE) {
        /* Plain Text Extension */
	/* read until Block Terminator found */
	do {
	  fread(fd, gbuf, 1);
	} while (gbuf[0] != '\0');
      } else {
        /* any other extension, next byte: Block Size */
	fread(fd, gbuf, 1);
	for (i = gbuf[0] + 1; i > 0; i--) {
	  /* read complete block and Block Terminator */
	  fread(fd, gbuf, 1);
	}
      }
    } else {
      /* Invalid Data */
      return 2;
    }
  }

  /* read remaining bytes of Image Descriptor */
  fread(fd, &gbuf[1], 9);

  w = USAT(gbuf[5], gbuf[6]);
  h = USAT(gbuf[7], gbuf[8]);
  lace = gbuf[9] & 0x40;
  if (gbuf[9] & 0x80) {
    mapmax = 1 << ((gbuf[9] & 0x07) + 1);
    fread(fd, cmap, mapmax * 3);
  }

  for( i = 0,cp = cmap; i < mapmax; i++ ) {
    map[i].r = *cp++;
    map[i].g = *cp++;
    map[i].b = *cp++;
    map[i].index = i;
    if( !color )
      tcmap[i] = (map[i].r * 54 + map[i].g * 183 + map[i].b * 18) >> 8;
    else
      tcmap[i] = i;
  }

  {
    char buf[40];

    StrPrintF(buf, "%d x %d", w, h);
    WinDrawChars(buf,strlen(buf),0 ,24);
  }

  lwzstack = MemPtrNew(8192);
  table = MemPtrNew(16384);
  if (!lwzstack || !table )
    Die("Out of Memory (gifread1)!");

  fread(fd, &c, 1);
  initLWZ(c);

  lid = DmFindDatabase(0, "ZBOXZSCRATCHPAD");
  if (!lid) {
    DmCreateDatabase(0, "ZBOXZSCRATCHPAD", 'BOXR', 'DATA', false);
    lid = DmFindDatabase(0, "ZBOXZSCRATCHPAD");
  }
  scratch = DmOpenDatabase(0, lid, dmModeReadWrite);
  if (!scratch)
    ErrDisplay("Could not access scratchpad");

  while (DmNumRecords(scratch))
    DmRemoveRecord(scratch, 0);

  editln = malloc(w);
  bandh = 49152 / w;
  lines = malloc((h + bandh - 1) / bandh * 4);

  for (i = 0; i < h; i += bandh) {
    dbrec = 65535;
    hand = DmNewRecord(scratch, &dbrec, w * bandh);
    if (!hand)
      Die("Out of Memory (gifread2)!");
    lines[i / bandh] = MemHandleLock(hand);
  }

  ycnt = 0;
    
  if (lace) {
    for (ypos = 0; ypos < h; ypos += 8) {
      putline(fd, ypos);
      if ( stop() ) return 1;
    }
    for (ypos = 4; ypos < h; ypos += 8) { 
      putline(fd, ypos);
      if ( stop() ) return 1;
    }
    for (ypos = 2; ypos < h; ypos += 4) {
       putline(fd, ypos);
       if ( stop() ) return 1;
    }
    for (ypos = 1; ypos < h; ypos += 2) { 
       putline(fd, ypos);
       if ( stop() ) return 1;
    }
  } else {
    for (ypos = 0; ypos < h; ypos++) {
      putline(fd, ypos);
      if ( stop() ) return 1;
    }
  }

  MemPtrFree(table);
  MemPtrFree(lwzstack);

  return 0;
}

/**************************************************/
void cleanGIF()
{
  int i;
  
  if (lines) {
    for (i = 0; i < h; i += bandh)
      MemPtrUnlock(lines[i / bandh]);
  }

  if (scratch) {
    while (DmNumRecords(scratch))
      DmRemoveRecord(scratch, 0);
    DmCloseDatabase(scratch);
    scratch = 0;
  }

  if (lines) {
    free(lines);
    lines = 0;
  }
  
  if (editln) {
    free(editln);
    editln = 0;
  }
}
